
Imports System.Collections.Generic
Imports System.Collections.Specialized

Public Class SyntaxRichTextBox
    Inherits System.Windows.Forms.RichTextBox

    ' ===============================================================================================
    '   CREATED BY DOT-NET
    ' ===============================================================================================
    Private Sub InitializeComponent()
        Me.SuspendLayout()
        Me.ResumeLayout(False)
    End Sub

    ' ===============================================================================================
    '  FLAGS
    ' ===============================================================================================
    Friend TextModified As Boolean = False
    Friend TextMustBeSaved As Boolean = False
    Friend TextMustBeIndented As Boolean = False

    Friend LABELS_WITHOUT_GOSUB As Boolean = True


    ' ===============================================================================================
    '  AUTOMATION LANGUAGE COLORS
    ' ===============================================================================================
    Private Color_Text As Color = Color.FromArgb(0, 0, 220)
    Private Color_KeywordText As Color = Color.FromArgb(120, 0, 0) ' 80 0 0
    Private Color_CommentText As Color = Color.FromArgb(0, 50, 50)
    Private Color_CommentBack As Color = Color.FromArgb(230, 250, 200) ' 240, 250, 220
    Private Color_BreakpointText As Color = Color.FromArgb(50, 0, 0)
    Private Color_BreakpointBack As Color = Color.FromArgb(255, 240, 100)

    Friend Color_ActiveBack As Color = Color.FromArgb(255, 255, 190)
    Friend Color_ErrorBack As Color = Color.FromArgb(255, 180, 150)
    Friend Color_WarningBack As Color = Color.FromArgb(255, 230, 140)
    Friend Color_ButtonWarningBack As Color = Color.FromArgb(255, 240, 180)
    Friend Color_NotFoundBack As Color = Color.FromArgb(255, 200, 170)

    Friend Function IsErrorColor(ByVal c As Color) As Boolean
        If c = Color_ErrorBack Then
            Return True
        Else
            Return False
        End If
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  TOKENIZER
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Enum TokenizerStatus As Int32
        Normal = 0
        InsideToken = 1
        InsideString = 2
    End Enum
    Friend Function Tokenize(ByVal line As String) As String()
        Dim status As TokenizerStatus
        Dim sa(-1) As String
        For i As Int32 = 0 To line.Length - 1
            Select Case status
                Case TokenizerStatus.Normal
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                        Case "'"c
                            Exit For ' comments are completely skipped 
                        Case """"c
                            status = TokenizerStatus.InsideString
                            ReDim Preserve sa(sa.Length)
                            ' this removes the exception if there is a line with a single "
                            sa(sa.Length - 1) = ""
                        Case Else
                            status = TokenizerStatus.InsideToken
                            ReDim Preserve sa(sa.Length)
                            sa(sa.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideToken
                    Select Case line(i)
                        Case " "c, Chr(9), "("c, ")"c
                            ' ------------------------------------------ end of token - eval right side ?  
                            status = TokenizerStatus.Normal
                            Select Case sa.Length
                                Case 1
                                    Select Case sa(0).ToLower
                                        Case "load", "save", "print", "presskeys", "sendkeys"
                                            ' -------------------------- isolate right side formula (trimmed) and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1)
                                            RemoveComments(sa(1))
                                            sa(1) = sa(1).Trim
                                            ' ------------------------------------------ stringtofile
                                            If sa(0).ToLower = "save" And sa(1).ToLower.StartsWith("stringtofile") Then
                                                ReDim Preserve sa(3)
                                                Dim p2 As Int32 = InStrRev(sa(1), " ")
                                                If p2 > 0 Then
                                                    sa(3) = sa(1).Substring(p2)
                                                    sa(2) = sa(1).Substring(12, p2 - 12)
                                                    sa(1) = "stringtofile"
                                                End If
                                            End If
                                            ' ------------------------------------------- deletefile
                                            If sa(0).ToLower = "save" And sa(1).ToLower.StartsWith("deletefile") Then
                                                ReDim Preserve sa(2)
                                                sa(2) = sa(1).Substring(10)
                                                sa(1) = "deletefile"
                                            End If
                                            ' -------------------------------------------- exit for
                                            Exit For
                                        Case "label"
                                            ' -------------------------- isolate right side formula (without "") and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim.Replace("""", "")
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "goto", "gosub"
                                            ' -------------------------- isolate right side formula and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "if", "select", "case"
                                            ' -------------------------- isolate right side formula and exit
                                            ReDim Preserve sa(1)
                                            sa(1) = line.Substring(i + 1).Trim
                                            RemoveComments(sa(1))
                                            Exit For
                                        Case "slot", "slottext"
                                            ' -------------------------- isolate left side formula and continue
                                            Dim eq As Int32 = InStr(line, "=")
                                            If eq > i Then
                                                ReDim Preserve sa(1)
                                                sa(1) = line.Substring(i, eq - i - 2)
                                                i = eq - 2
                                            End If
                                        Case "for"
                                            RemoveComments(line)
                                            line = ReplaceMultipleSpacesAndTabAndTrim(line)
                                            Dim sa2() As String = line.Split(" "c)
                                            Dim sa3() As String = Split(line, " = ")
                                            If sa3.Length > 1 Then
                                                Dim sa4() As String = Split(sa3(1), " to ", -1, CompareMethod.Text)
                                                If sa4.Length > 1 Then
                                                    Dim sa5() As String = Split(sa4(1), " step ", -1, CompareMethod.Text)
                                                    If sa5.Length > 1 Then
                                                        ReDim sa(7)
                                                        sa(0) = "for"
                                                        sa(1) = sa2(1)
                                                        sa(2) = "="
                                                        sa(3) = sa4(0)
                                                        sa(4) = "to"
                                                        sa(5) = sa5(0)
                                                        sa(6) = "step"
                                                        sa(7) = sa5(1)
                                                    Else
                                                        ReDim sa(5)
                                                        sa(0) = "for"
                                                        sa(1) = sa2(1)
                                                        sa(2) = "="
                                                        sa(3) = sa4(0)
                                                        sa(4) = "to"
                                                        sa(5) = sa4(1)
                                                    End If
                                                End If
                                            End If
                                            Exit For

                                        Case Else
                                            ' --------------------------------- GOSUB WITH PARAMETERS - WITHOUT GOSUB
                                            If GetLabelLine(sa(0)) >= 0 Then
                                                ReDim sa(0)
                                                sa(0) = line
                                                RemoveComments(sa(0))
                                                sa(0) = sa(0).Trim
                                                Exit For
                                            End If
                                    End Select
                                Case 2
                                    Select Case sa(0).ToLower
                                        Case "wait"
                                            If sa(1).ToLower = "button" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = line.Substring(12).Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            ElseIf sa(1).ToLower <> "seconds" And _
                                                   sa(1).ToLower <> "application" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(1)
                                                sa(1) = line.Substring(5).Trim
                                                RemoveComments(sa(1))
                                                Exit For
                                            End If
                                        Case "com"
                                            If sa(1).ToLower = "open" Then
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, sa(1).ToLower).Trim
                                                RemoveComments(sa(2))
                                                Dim n() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(2)).Split(" "c)
                                                If n.Length = 2 Then
                                                    ReDim Preserve sa(3)
                                                    sa(2) = n(0)
                                                    sa(3) = n(1)
                                                    Exit For
                                                End If
                                            ElseIf sa(1).ToLower <> "writestring" Or sa(1).ToLower <> "writeline" Or _
                                                   sa(1).ToLower <> "writestringhex" Or sa(1).ToLower <> "writestringdec" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, sa(1).ToLower).Trim
                                                RemoveComments(sa(2))
                                                If sa(2).StartsWith("(") Or sa(2).EndsWith(")") Then ReDim sa(1)
                                                Exit For
                                            End If
                                        Case "tts"
                                            ' ----------------------- isolate right side formula and exit
                                            ReDim Preserve sa(2)
                                            sa(2) = GetRightPartAfterThisLowerCaseWord(line, sa(1).ToLower).Trim
                                            RemoveComments(sa(2))
                                            Exit For
                                        Case Else
                                            If IsSomeVariable(sa(0)) Then
                                                If sa(1) = "=" Then
                                                    Dim pos As Int32 = line.IndexOf("=") + 1
                                                    If line.Substring(pos).Trim.ToLower.StartsWith("input") Then
                                                        ' -------------------------- isolate input prompt
                                                        ReDim Preserve sa(3)
                                                        sa(2) = "input"
                                                        sa(3) = GetRightPartAfterThisLowerCaseWord(line, "input")
                                                        RemoveComments(sa(3))
                                                        Exit For
                                                    ElseIf line.Substring(pos).Trim.ToLower.StartsWith("message") Then
                                                        ' -------------------------- isolate message prompt
                                                        ReDim Preserve sa(3)
                                                        sa(2) = "message"
                                                        sa(3) = GetRightPartAfterThisLowerCaseWord(line, "message")
                                                        RemoveComments(sa(3))
                                                        Exit For
                                                    Else
                                                        ' -------------------------- isolate right side formula
                                                        ReDim Preserve sa(2)
                                                        sa(2) = line.Substring(pos).Trim
                                                        RemoveComments(sa(2))
                                                    End If
                                                    ' ----------------------------- and exit
                                                    Exit For
                                                End If
                                            End If
                                    End Select
                                Case 3
                                    Select Case sa(0).ToLower
                                        Case "slot", "slottext"
                                            If sa(2) = "=" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = line.Substring(i + 1)
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                        Case "wait"
                                            If sa(1).ToLower = "seconds" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, "seconds").Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            End If
                                            If sa(1).ToLower = "application" Then
                                                ' -------------------------- isolate right side formula and exit
                                                ReDim Preserve sa(2)
                                                sa(2) = GetRightPartAfterThisLowerCaseWord(line, "application").Trim
                                                RemoveComments(sa(2))
                                                Exit For
                                            End If
                                        Case "button"
                                            If sa(2).ToLower = "label" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, "label").Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                            If sa(2).ToLower = "text" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, "text").Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                            If sa(2).ToLower = "color" Then
                                                ' ----------------------- isolate right side formula and exit
                                                ReDim Preserve sa(3)
                                                sa(3) = GetRightPartAfterThisLowerCaseWord(line, "color").Trim
                                                RemoveComments(sa(3))
                                                Exit For
                                            End If
                                        Case "key"
                                                If sa(2).ToLower = "goto" OrElse sa(2).ToLower = "gosub" Then
                                                    ' ----------------------- isolate right side formula and exit
                                                    ReDim Preserve sa(3)
                                                    sa(3) = GetRightPartAfterThisLowerCaseWord(line, sa(2).ToLower).Trim
                                                    RemoveComments(sa(3))
                                                    Exit For
                                                End If
                                    End Select
                            End Select
                        Case "'"c
                            Exit For ' comments are completely skipped 
                        Case Else
                            sa(sa.Length - 1) += line(i)
                    End Select
                Case TokenizerStatus.InsideString
                    Select Case line(i)
                        Case """"c
                            If i >= line.Length - 2 OrElse line(i + 1) <> """"c Then
                                status = TokenizerStatus.Normal
                            Else
                                sa(sa.Length - 1) += line(i)
                                i += 1
                            End If
                        Case Else
                            sa(sa.Length - 1) += line(i)
                    End Select
            End Select
        Next
        ' -------------------------------------------- Immediate vars tokenizer
        If sa.Length > 5 Then
            If sa(0).ToLower = "variable" Then
                ' ----------------------- isolate right side formula and exit
                ReDim Preserve sa(4)
                sa(4) = GetRightPartAfterThisLowerCaseWord(line, "=").Trim
                RemoveComments(sa(4))
            End If
        End If
        '
        Return sa
    End Function

    ' ===============================================================================================
    '  Convert "Button n Text" to "Button n Label"
    ' ===============================================================================================
    Friend Sub ConvertButtonTextToLabel()
        RTB_Lines = Lines
        Dim line As String
        Dim sa(-1) As String
        For i As Int32 = 0 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("button ") Then
                line = ReplaceMultipleSpacesAndTabAndTrim(line)
                RemoveComments(line)
                sa = line.Split(" "c)
                'If sa.Length = 4 Then
                If sa(2) = "text" Then
                    If IsValidButtonNumber(sa(1)) Then
                        RTB_Lines(i) = RTB_Lines(i).Replace("Text", "Label")
                    End If
                End If
                'End If
            End If
        Next
        Lines = RTB_Lines
    End Sub

    ' ===============================================================================================
    '  FIND LABEL
    ' ===============================================================================================
    Friend Function FindLabel(ByVal lines() As String, _
                              ByVal TextToFind As String, _
                              Optional ByVal AllowPartialName As Boolean = False, _
                              Optional ByVal ExcludeLine As Int32 = -1) As Int32
        '
        ' -----------------------------------------------------------------------
        If TextToFind Is Nothing Then Return -1
        If TextToFind = "" Then Return -1
        ' ----------------------------------------------------------------------- prepare text to find
        TextToFind = TextToFind.ToLower
        ' ----------------------------------------------------------------------- test complete names
        Dim name As String
        Dim TextToFind0 As String = TextToFind.Trim
        For i As Int32 = 0 To LabelNames.Length - 1
            name = LabelNames(i).Trim.ToLower
            RemoveComments(name)
            If name = TextToFind0 Then
                If ExcludeLine < 0 OrElse ExcludeLine <> LabelLineNumbers(i) Then
                    Return LabelLineNumbers(i)
                End If
            End If
        Next
        ' ----------------------------------------------------------------------- test partial names
        Dim TextToFind1 As String
        Dim TextToFind2 As String
        Dim TextToFind3 As String
        If AllowPartialName Then
            TextToFind1 = TextToFind
            TextToFind2 = ExtractFirstWord(TextToFind)
            TextToFind3 = EvalString_AutoQuoting(TextToFind)
            For i As Int32 = 0 To LabelNames.Length - 1
                name = LabelNames(i).Trim.ToLower
                RemoveComments(name)
                If name.StartsWith(TextToFind1) Or _
                   name.StartsWith(TextToFind2) Or _
                   name.StartsWith(TextToFind3) Then
                    If ExcludeLine < 0 OrElse ExcludeLine <> LabelLineNumbers(i) Then
                        Return LabelLineNumbers(i)
                    End If
                End If
            Next
        End If
        ' ------------------------------------------------------------------------ if not found then try to evaluate the string
        Dim line As String
        TextToFind1 = " " & EvalString(TextToFind).ToLower
        For i As Int32 = 0 To lines.Length - 1
            If ExcludeLine < 0 OrElse ExcludeLine <> i Then
                line = lines(i).Trim.ToLower
                If line.StartsWith("label") Then
                    RemoveComments(line)
                    If line.EndsWith(TextToFind1) Then
                        Return i
                    End If
                End If
            End If
        Next
        ' ------------------------------------------------------------------------ not found
        Return -1
    End Function


    ' ===============================================================================================================
    '   RTB - SAVE and RESTORE selection and scroll ( scroll aligned with the first text line )
    ' ===============================================================================================================
    Friend Structure ScrollData
        Dim OldSelectionStart As Int32
        Dim OldSelectionLength As Int32
        Dim OldSelectionFirstChar As Int32
        Dim OldSelectionLastChar As Int32
        Dim OldSelectionFirstLine As Int32
        Dim OldSelectionLastLine As Int32
    End Structure

    Friend Sub SaveSelectionAndScroll(ByRef sd As ScrollData)
        sd.OldSelectionStart = Me.SelectionStart
        sd.OldSelectionLength = Me.SelectionLength
        sd.OldSelectionFirstChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        sd.OldSelectionLastChar = Me.GetCharIndexFromPosition(New Point(1, Me.ClientSize.Height))
        sd.OldSelectionFirstLine = Me.GetFirstSelectedLine
        sd.OldSelectionLastLine = Me.GetLastSelectedLine
    End Sub

    Friend Sub RestoreSelectionAndScroll(ByRef sd As ScrollData)
        '
        ' ---------------------------------------------------------- restore scroll 
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        '
        ' ---------------------------------------------------------- restore selection point (if ...)
        If sd.OldSelectionStart >= sd.OldSelectionFirstChar Then
            Me.Select(sd.OldSelectionStart, sd.OldSelectionLength)
        End If
    End Sub

    Friend Sub RestoreScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
    End Sub

    Friend Sub RestoreSelectedLinesAndScroll(ByRef sd As ScrollData)
        Me.Select(sd.OldSelectionFirstChar, 0)
        Me.ScrollToCaret()
        ' --------------------------------------------------------------
        If sd.OldSelectionLastLine >= sd.OldSelectionFirstLine Then
            Dim firstch As Int32 = GetFirstCharIndexFromLine(sd.OldSelectionFirstLine)
            Dim lastch As Int32 = GetLastCharIndexFromLine(sd.OldSelectionLastLine)
            Me.Select(firstch, lastch - firstch)
        End If
    End Sub



    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  COLOR LINES
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Friend ColorAllLinesMilliseconds As Int64

    Friend Sub ColorAllLines()
        ' ---------------------------------------------------------- necessary when called from replace or load
        InitButtonsIdentifiers()
        ' ---------------------------------------------------------- init load time
        Dim sw As Stopwatch = New Stopwatch
        sw.Start()
        ' ---------------------------------------------------------- init
        FirstErrorLine = -1
        ErrorsCounter = 0
        WarningsCounter = 0
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTB_Lines.Length - 1
            ' ------------------------------------------------------ AutomationLanguageColors
            ColorSingleLine_AndReplaceKeywords(i)
            ' ------------------------------------------------------ Progress indicator
            If i Mod 100 = 0 Then
                ' --------------------------------------------------
                Form1.StatusLabel1.Text = " Line: " + i.ToString
                Form1.StatusLabel1.Refresh()
                ' -------------------------------------------------- 
                If sw.ElapsedMilliseconds > 2000 Then
                    Form1.Opacity = 1
                    Form1.Show()
                End If
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        Me.Focus()
        ' ---------------------------------------------------------- load time
        ColorAllLinesMilliseconds = sw.ElapsedMilliseconds
    End Sub

    Private Sub ColorRangeOfLines(ByVal first As Int32, ByVal last As Int32)
        Static oldline As Int32 = -1
        Dim line As Int32 = GetLineFromCharIndex(Me.SelectionStart)
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ----------------------------------------------------------
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        For i As Int32 = first To last
            ColorSingleLine(i)
        Next
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        ' ----------------------------------------------------------
        Me.Focus()
        oldline = line
    End Sub

    Private Sub ColorAllLines_StartingWithVariable()
        ' ----------------------------------------------------------
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Int32 = 0 To RTB_Lines.Length - 1
            If RTB_Lines(i).ToLower.Trim.StartsWith("variable") Then
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        Me.Focus()
    End Sub

    Friend Sub ColorAllLines_StartingWith(ByVal str As StringCollection)
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTB_Lines.Length - 1
            Dim ok As Boolean = False
            For j As Int32 = 0 To str.Count - 1
                If RTB_Lines(i).TrimStart.ToLower.StartsWith(str.Item(j)) Then
                    ok = True
                End If
            Next
            If ok Then
                SelectLine(i)
                Me.SelectionColor = Color_Text
                Me.SelectionBackColor = Color_ActiveBack
                ' -------------------------------------------------- AutomationLanguageColors
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        Me.Focus()
    End Sub

    Friend Sub ColorAllLines_Containing(ByVal str As StringCollection)
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------- save position 
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        For i As Integer = 0 To RTB_Lines.Length - 1
            Dim ok As Boolean = False
            For j As Int32 = 0 To str.Count - 1
                If RTB_Lines(i).TrimStart.ToLower.Contains(str.Item(j)) Then
                    ok = True
                End If
            Next
            If ok Then
                SelectLine(i)
                Me.SelectionColor = Color_Text
                Me.SelectionBackColor = Color_ActiveBack
                ' -------------------------------------------------- AutomationLanguageColors
                ColorSingleLine_AndReplaceKeywords(i)
            End If
        Next
        ' ---------------------------------------------------------- restore
        RestoreSelectionAndScroll(sd)
        ' ---------------------------------------------------------- 
        Me.Enabled = oldEnabled
        PaintEventEnabled = oldPaintEvent
        Me.Focus()
    End Sub

    Friend Sub ColorLines(ByVal startline As Int32, ByVal endline As Int32)
        RTB_Lines = Me.Lines
        For i As Int32 = startline To endline
            ColorSingleLine(i)
        Next
    End Sub

    Friend Sub ColorSingleLine(ByVal i As Int32)
        Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- save position 
        Dim OldCursorPosition As Integer = SelectionStart
        ' ---------------------------------------------------------- set the whole line to default colors
        SelectLine(i)
        Me.SelectionColor = Color_Text
        Me.SelectionBackColor = Color_ActiveBack
        ' ---------------------------------------------------------- AutomationLanguageColors
        ColorSingleLine_AndReplaceKeywords(i)
        ' ---------------------------------------------------------- restore
        SelectionStart = OldCursorPosition
        SelectionLength = 0
        ' ---------------------------------------------------------- 
        EventsAreEnabled = OldEventsAreEnabled
    End Sub

    Friend Sub ColorSelectedLine()
        RTB_Lines = Me.Lines
        ColorSingleLine(GetLineFromCharIndex(SelectionStart))
    End Sub

    Friend Sub ColorVisibleLines()
        ColorRangeOfLines(GetFirstVisibleLine(), GetLastVisibleLine())
    End Sub


    ' ===============================================================================================
    '  TEST KEYWORDS AND EVENTUALLY RECOLOR LINES 
    ' ===============================================================================================
    Private TestLines_StartingWords As StringCollection = New StringCollection
    Private TestLines_ContainingWords As StringCollection = New StringCollection

    Private Sub TestKeywords_PrepareKeywordsToRecolor()
        ' ---------------------------------------------------------------
        '  This function is called from KEYDOWN AND KEYUP
        ' ---------------------------------------------------------------
        Dim startline As Int32 = GetFirstSelectedLine()
        If startline < 0 Then Return
        If startline >= RTB_Lines.Length Then Return
        Dim endline As Int32 = GetLastSelectedLine()
        For i As Int32 = startline To endline
            TestKeywords_PrepareKeywordsToRecolor(i)
        Next
    End Sub

    Private Sub TestKeywords_PrepareKeywordsToRecolor(ByVal nline As Int32)
        Dim s As String = RTB_Lines(nline).Trim.ToLower
        If s.StartsWith("for") Then
            TestLines_StartingWords.Add("variable")
        End If
        If s.StartsWith("next") Then
            TestLines_StartingWords.Add("for")
            TestLines_StartingWords.Add("exit")
        End If
        If s.StartsWith("end") Or s.StartsWith("return") Then
            TestLines_StartingWords.Add("label")
        End If
        If s.StartsWith("label") Then
            AddString_NewStringOnly(TestLines_StartingWords, "button")
            AddString_NewStringOnly(TestLines_StartingWords, "key")
            AddString_NewStringOnly(TestLines_StartingWords, "gosub")
            AddString_NewStringOnly(TestLines_StartingWords, "goto")
            AddString_NewStringOnly(TestLines_StartingWords, "label")
            If s.Length > 5 Then
                Dim word As String = s.Substring(6)
                RemoveComments(word)
                AddString_NewStringOnly(TestLines_ContainingWords, word.Trim.ToLower.Replace("""", ""))
            End If
        End If
        If s.StartsWith("else") Then
            AddString_NewStringOnly(TestLines_StartingWords, "if")
        End If
        If s.StartsWith("endif") Then
            AddString_NewStringOnly(TestLines_StartingWords, "if")
            AddString_NewStringOnly(TestLines_StartingWords, "else")
        End If
        If s.StartsWith("endselect") Then
            AddString_NewStringOnly(TestLines_StartingWords, "select")
            AddString_NewStringOnly(TestLines_StartingWords, "case")
        End If
        If s.StartsWith("variable") Then
            Dim sa() As String = ReplaceMultipleSpacesAndTabAndTrim(s).Split(" "c)
            If sa.Length >= 3 Then
                If sa(2).Length >= UserVariablesMinLength Then
                    AddString_NewStringOnly(TestLines_ContainingWords, sa(2))
                End If
            End If
            TestLines_StartingWords.Add("label") ' < to recolor the Event_ExternalCommands
        End If
        If s.StartsWith("return") Then
            TestLines_StartingWords.Add("label") ' < to recolor the Event_ExternalCommands
        End If
    End Sub

    Private Sub AddString_NewStringOnly(ByRef sc As StringCollection, ByVal s As String)
        If Not sc.Contains(s) Then
            sc.Add(s)
        End If
    End Sub

    Private Sub TestKeywords_RecolorLines()
        ' ----------------------------------------------------------------
        Cursor.Current = Cursors.WaitCursor
        ' ----------------------------------------------------------------
        RTB_Lines = Me.Lines()  ' <--- 900 uS
        ' ----------------------------------------------------------------
        If TestLines_StartingWords.Count > 0 Then
            ColorAllLines_StartingWith(TestLines_StartingWords)
            TestLines_StartingWords.Clear()
        End If
        If TestLines_ContainingWords.Count > 0 Then
            ColorAllLines_Containing(TestLines_ContainingWords)
            TestLines_ContainingWords.Clear()
        End If
        Cursor.Current = Cursors.Default
    End Sub

    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  COLOR SINGLE LINE AND REPLACE KEYWORDS
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    Private LineWithErrors As Boolean

    Private Sub ColorSingleLine_AndReplaceKeywords(ByVal nline As Int32)
        If nline < 0 Or nline > RTB_Lines.Length - 1 Then Return
        ' --------------------------------------------------------------- 
        Dim line As String = RTB_Lines(nline).Trim
        ' --------------------------------------------------------------- 
        If line <> "" Then
            ' ----------------------------------------------------------- color active back
            SelectLine(nline)
            Me.SelectionColor = Color_Text
            Me.SelectionBackColor = Color_ActiveBack
            ' ----------------------------------------------------------- comment position
            Dim CommentPosition As Int32 = RemoveComments(line)
            If CommentPosition = 0 Then
                MarkCommentLine(nline, 0)
                Return
            End If
            '
            LineWithErrors = False
            ' ----------------------------------------------------------- tokenize to array of keywords
            Dim sa() As String = Tokenize(line.Trim.ToLowerInvariant)
            ' ----------------------------------------------------------- 
            Select Case sa.Length
                Case 0
                Case 1
                    Select Case sa(0)
                        Case "beep", "controls", "next", "case", "caseelse", "endif", "endselect", "return", "select"
                        Case "sewabort"
                            If Not SewFunctionsEnabled Then MarkErrorLine(nline)
                        Case "stop", "end"
                            Me.SelectionBackColor = Color.FromArgb(120, 255, 120)
                        Case "exit"
                            If Not TestNext(nline, 1) Then MarkNotFoundLine(nline)
                        Case "else"
                            If Not TestEndif(nline) Then MarkNotFoundLine(nline)
                        Case Else
                            If LABELS_WITHOUT_GOSUB Then
                                Dim n As Int32 = FindLabel(RTB_Lines, sa(0))
                                ' ------------------------------------------------------ GOSUB WITH PARAMETERS - WITHOUT GOSUB
                                If n < 0 Then
                                    ' -------------------------------------------------- extract params and remove them from the name (if they are valid)
                                    Dim par() As String = ExtractFunctionParams(sa(0))
                                    ' -------------------------------------------------- Test the label-name without parameters
                                    n = GetLabelLine(sa(0))
                                End If
                                If n < 0 Then
                                    MarkErrorLine(nline)
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                    End Select
                Case 2
                    Select Case sa(0)
                        Case "beep"
                            If Not IsValidString_AutoQuoting(sa(1)) Then MarkNotFoundLine(nline)
                        Case "case"
                            If Not TestEndselect(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "com"
                            If sa(1) <> "close" AndAlso _
                               sa(1) <> "discardinbuffer" AndAlso _
                               sa(1) <> "discardoutbuffer" AndAlso _
                               sa(1) <> "enabledtr" AndAlso _
                               sa(1) <> "enablerts" AndAlso _
                               sa(1) <> "disabledtr" AndAlso _
                               sa(1) <> "disablerts" Then
                                MarkErrorLine(nline)
                            End If
                        Case "controls"
                            If sa(1) <> "opentextbox" AndAlso _
                               sa(1) <> "closetextbox" AndAlso _
                               sa(1) <> "cleartextbox" Then
                                MarkErrorLine(nline)
                            End If
                        Case "if"
                            If Not TestElseOrEndif(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBoolean(sa(1)) Then MarkErrorLine(nline)
                        Case "wait"
                            If Not IsValidBoolean(sa(1)) Then MarkErrorLine(nline)
                        Case "label"
                            If FindLabel(RTB_Lines, sa(1), False, nline) > -1 Or line.Contains("""") Then
                                MarkErrorLine(nline)
                            Else
                                If sa(1) = "eventstop" Then
                                    If Not TestTheFinalEndForTheEventStopLabel(nline) Then
                                        MarkErrorLine(nline)
                                    End If
                                ElseIf sa(1) = "eventtimer" Then
                                    Form1.EventTimer_Init()
                                    If Form1.EventTimer_StartLine < 0 Then
                                        MarkErrorLine(nline)
                                    End If
                                ElseIf sa(1) = "event_externalcommands" Then
                                    If GetNumericUserVariable("Slot_ExternalCommands") <= 0 Or _
                                                       Form1.GetReturnLineNumber(nline) <= 0 Then
                                        MarkErrorLine(nline)
                                    End If
                                End If
                            End If
                        Case "exit"
                            If CInt(Val(sa(1))) <= 0 Then MarkNotFoundLine(nline)
                            If Not TestNext(nline, CInt(Val(sa(1)))) Then MarkNotFoundLine(nline)
                        Case "gosub"
                            ' ------------------------------------------------------------------------- GOSUB WITH PARAMETERS
                            If GetLabelLine(GetFunctionNameWithoutParams(line.Substring(6))) >= 0 Then
                                ' -------------------------------------------------- test if params are valid
                                ExtractFunctionParams(sa(1))
                                ' -------------------------------------------------- Test the label-name without parameters 
                                If Not FunctionParamsAreValid(sa(1)) Then
                                    MarkErrorLine(nline)
                                End If
                            ElseIf line.Contains("""") And Not line.Contains("+") Then
                                MarkErrorLine(nline)
                            ElseIf FindLabel(RTB_Lines, sa(1)) < 0 Then
                                If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                    If Not IsValidString(sa(1)) Then
                                        MarkWarningLine(nline)
                                    End If
                                Else
                                    MarkNotFoundLine(nline)
                                End If
                            End If
                        Case "goto"
                            If line.Contains("""") And Not line.Contains("+") Then
                                MarkErrorLine(nline)
                            ElseIf FindLabel(RTB_Lines, sa(1)) < 0 Then
                                If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                    If Not IsValidString(sa(1)) Then
                                        MarkWarningLine(nline)
                                    End If
                                Else
                                    MarkNotFoundLine(nline)
                                End If
                            End If
                        Case "load"
                            If sa(1) <> "slots" AndAlso _
                               sa(1) <> "vars" AndAlso _
                               sa(1) <> "stop" AndAlso _
                               sa(1) <> "pause" AndAlso _
                               sa(1) <> "play" AndAlso _
                               sa(1) <> "hide" AndAlso _
                               sa(1) <> "openapps" AndAlso _
                               sa(1) <> "closeapps" AndAlso _
                               sa(1) <> "closeall" Then
                                If sa(1).ToLower.Trim.StartsWith("slots") Or _
                                   sa(1).ToLower.Trim.StartsWith("vars") Or _
                                   sa(1).ToLower.Trim.StartsWith("openapps") Then
                                    Dim n() As Double = ExtractNumericParams(sa(1), 3)
                                    If n.Length <> 2 Then MarkErrorLine(nline)
                                Else
                                    '
                                    If sa(1).StartsWith("position") Then
                                        Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c)
                                        If sa2.Length <> 2 Then
                                            MarkErrorLine(nline)
                                        End If
                                    End If
                                    '
                                    If Not DoubleQuotesAreBalanced(sa(1)) Then
                                        MarkErrorLine(nline)
                                    End If
                                    '
                                    Dim str As String = EvalString_AutoQuoting(sa(1))
                                    Select Case Strings.Right(str, 4).ToLower
                                        Case ".txt"
                                            '
                                            ' try to load any string (if exists) 
                                            ' 
                                            ' --------------------------------------------- test load file
                                            'Dim sa2() As String = sa(1).Split(" "c)
                                            'If sa2.Length > 1 Then
                                            '    If Not IsStringVariable(sa2(0)) Then MarkErrorLine(nline)
                                            'End If

                                            If Not IsValidString_AutoQuoting(str) Then MarkErrorLine(nline)
                                    End Select
                                End If
                            End If
                        Case "presskeys"
                            If Not IsValidString_AutoQuoting(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "sendkeys"
                            If Not IsValidString_AutoQuoting(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "print"
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "save"
                            If sa(1) <> "slots" AndAlso _
                               sa(1) <> "vars" Then
                                MarkErrorLine(nline)
                            End If
                        Case "select"
                            If Not TestCase(nline) Then MarkNotFoundLine(nline)
                            If Not TestEndselect(nline) Then MarkNotFoundLine(nline)
                            If Not IsValidBooleanStringOrDouble(sa(1)) Then MarkErrorLine(nline)
                        Case "tts"
                            Select Case sa(1)
                                Case "stop", "pause", "resume"
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "varsfromfile"
                            If Not IsValidString_AutoQuoting(sa(1)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "window"
                            If sa(1) <> "setfocus" AndAlso _
                               sa(1) <> "minsize" AndAlso _
                               sa(1) <> "sizable" AndAlso _
                               sa(1) <> "maximized" AndAlso _
                               sa(1) <> "fullscreen" Then
                                MarkErrorLine(nline)
                            End If
                        Case Else : MarkErrorLine(nline)
                    End Select
                Case 3
                    Select Case sa(0)
                        Case "beep"
                            If Not IsValidDouble(sa(1)) Or Not IsValidDouble(sa(2)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "button"
                            If IsValidButtonIdentifier(sa(1)) Then
                                Select Case sa(2)
                                    Case "enabled", "disabled"
                                    Case "color" : MarkWarningLine(nline)
                                    Case Else : MarkErrorLine(nline)
                                End Select
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "com"
                            Select Case sa(1)
                                Case "writestring", "writestringhex", "writestringdec", "writeline"
                                    If Not IsValidString_AutoQuoting(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "key"
                            If IsValidKey(sa(1)) Then
                                If line.Contains("""") And Not line.Contains("+") Then
                                    MarkErrorLine(nline)
                                ElseIf FindLabel(RTB_Lines, sa(2)) < 0 Then
                                    If sa(2).Contains(" ") Or IsSomeVariable(sa(2)) Then
                                        If Not IsValidString(sa(1)) Then
                                            MarkWarningLine(nline)
                                        End If
                                    Else
                                        MarkNotFoundLine(nline)
                                    End If
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "option"
                            Select Case sa(1)
                                Case "globalkeys"
                                    Select Case sa(2)
                                        Case "enabled", "disabled"
                                        Case Else
                                            MarkErrorLine(nline)
                                    End Select
                                Case "speed"
                                    If Not IsValidDouble(sa(2)) Then MarkErrorLine(nline)
                                Case "transparency"
                                    If Not IsValidDouble(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "save"
                            Select Case sa(1)
                                Case "deletefile"
                                    If Not IsValidString(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "tts"
                            Select Case sa(1)
                                Case "selectvoice"
                                    If Not IsValidString(sa(2)) Then MarkErrorLine(nline)
                                Case "setvolume", "setspeed"
                                    If Not IsValidDouble(sa(2)) Then MarkErrorLine(nline)
                                Case "speak", "speakwait"
                                    If Not IsValidString_AutoQuoting(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "variable"
                            Select Case sa(1)
                                Case "numeric"
                                    If Not IsNumericUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case "string"
                                    If Not IsStringUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "wait"
                            Select Case sa(1)
                                Case "button"
                                    '
                                Case "seconds"
                                    If Not IsValidDouble(sa(2)) Then MarkErrorLine(nline)
                                Case "application"
                                    If Not IsValidString_AutoQuoting(sa(2)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case Else
                            If sa(1) = "=" Then
                                If IsNumericVariable(sa(0)) Then
                                    If IsValidDouble(sa(2)) OrElse IsValidBoolean(sa(2)) Then
                                    Else
                                        MarkErrorLine(nline)
                                    End If
                                ElseIf IsStringVariable(sa(0)) Then
                                    If IsValidString(sa(2)) Then
                                    Else
                                        MarkErrorLine(nline)
                                    End If
                                Else
                                    MarkErrorLine(nline)
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                    End Select
                Case 4
                    Select Case sa(0)
                        Case "button"
                            If IsValidButtonIdentifier(sa(1)) Then
                                Select Case sa(2)
                                    Case "label"
                                        If DuplicateButtonIdentifier(sa(3)) Then MarkErrorLine(nline)
                                    Case "text"
                                        If IsNumeric(sa(1)) Then
                                            If FindLabel(RTB_Lines, sa(3).Replace("""", "")) < 0 Then
                                                If sa(3).Contains(" ") Or _
                                                   (sa(3).StartsWith("""") And sa(3).EndsWith("""")) Or _
                                                   sa(3).Contains("+") Or IsSomeVariable(sa(3)) Then
                                                Else
                                                    If sa(3).Replace("""", "").Length > 0 Then
                                                        MarkWarningLine(nline)
                                                    End If
                                                End If
                                            End If
                                        End If
                                        If DuplicateButtonIdentifier(sa(3)) Then MarkErrorLine(nline)
                                    Case "slot"
                                        Dim slot As Int32 = CInt(EvalDouble(sa(3)))
                                        If slot <= 0 Or slot > 999 Then
                                            If Not IsSomeVariable(sa(3)) Then
                                                MarkErrorLine(nline)
                                            End If
                                        End If
                                    Case "color"
                                        Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(3)).Split(" "c)
                                        If sa(3).StartsWith("flashing") Then
                                            If sa2.Length < 3 Then
                                                MarkWarningLine(nline)
                                            ElseIf sa2.Length = 3 Then
                                                If ColorFromColorName(sa2(1)) = Color.Empty Or _
                                                   ColorFromColorName(sa2(2)) = Color.Empty Then
                                                    MarkWarningLine(nline)
                                                End If
                                            Else
                                                MarkErrorLine(nline)
                                            End If
                                        Else
                                            If sa2.Length < 2 Then
                                                If ColorFromColorName(sa(3)) = Color.Empty Then
                                                    MarkWarningLine(nline)
                                                End If
                                            Else
                                                MarkErrorLine(nline)
                                            End If
                                        End If
                                    Case Else
                                        MarkErrorLine(nline)
                                End Select
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "com"
                            If sa(1) = "open" Then
                                If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                Else
                                    MarkErrorLine(nline)
                                End If
                            End If
                        Case "controls"
                            If sa(1) = "setcursorpos" Then
                                If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                Else
                                    MarkErrorLine(nline)
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "key"
                            If IsValidKey(sa(1)) AndAlso (sa(2) = "goto" Or sa(2) = "gosub") Then
                                If line.Contains("""") And Not line.Contains("+") Then
                                    MarkErrorLine(nline)
                                ElseIf FindLabel(RTB_Lines, sa(3)) < 0 Then
                                    If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                        If Not IsValidString(sa(1)) Then
                                            MarkWarningLine(nline)
                                        End If
                                    Else
                                        MarkNotFoundLine(nline)
                                    End If
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "save"
                            Select Case sa(1)
                                Case "stringtofile"
                                    If Not IsValidString(sa(2)) Or Not IsValidString(sa(3)) Then MarkErrorLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case "slot"
                            If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            Else
                                Select Case sa(3).Trim
                                    Case "reset", "sleep", "recognize", "calibrate"
                                        If IsValidDouble(sa(1)) And sa(2) = "=" Then
                                        Else
                                            MarkErrorLine(nline)
                                        End If
                                    Case Else : MarkErrorLine(nline)
                                End Select
                            End If
                        Case "slottext"
                            If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidString_AutoQuoting(sa(3)) Then
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case Else
                            Select Case sa(2)
                                Case "input", "message"
                                Case Else
                                    MarkErrorLine(nline)
                            End Select
                    End Select
                Case 5
                    ' -------------------------------------------- Immediate vars colors
                    Select Case sa(0)
                        Case "variable"
                            Select Case sa(1)
                                Case "numeric"
                                    If Not IsNumericUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If sa(3) <> "=" Then MarkErrorLine(nline) : Exit Select
                                    If Not IsValidDouble(sa(4)) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case "string"
                                    If Not IsStringUserVariable(sa(2), nline) Then MarkErrorLine(nline) : Exit Select
                                    If sa(3) <> "=" Then MarkErrorLine(nline) : Exit Select
                                    If Not IsValidString_AutoQuoting(sa(4)) Then MarkErrorLine(nline) : Exit Select
                                    If UnusedVariable(sa(2), nline) Then MarkWarningLine(nline)
                                Case Else : MarkErrorLine(nline)
                            End Select
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 6
                    Select Case sa(0)
                        Case "for"
                            If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                               IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                                If Not TestForNext(nline) Then
                                    MarkErrorLine(nline)
                                End If
                            Else
                                MarkErrorLine(nline)
                            End If
                        Case "sewinit"
                            If Not SewFunctionsEnabled Then MarkErrorLine(nline)
                            For i As Int32 = 1 To 5
                                If Not IsValidDouble(sa(i)) Then
                                    MarkErrorLine(nline)
                                End If
                            Next
                        Case "sewto"
                            If Not SewFunctionsEnabled Then MarkErrorLine(nline)
                            For i As Int32 = 1 To 4
                                If Not IsValidDouble(sa(i)) Then
                                    MarkErrorLine(nline)
                                End If
                            Next
                            If Not IsValidString_AutoQuoting(sa(5)) Then
                                MarkErrorLine(nline)
                            End If
                        Case "sewlimits"
                            If Not SewFunctionsEnabled Then MarkErrorLine(nline)
                            For i As Int32 = 1 To 5
                                If Not IsValidDouble(sa(i)) Then
                                    MarkErrorLine(nline)
                                End If
                            Next
                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 7
                    Select Case sa(0)

                        Case Else
                            MarkErrorLine(nline)
                    End Select
                Case 8
                    If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                       IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) And _
                       sa(6) = "step" And IsValidDouble(sa(7)) Then
                        If Not TestForNext(nline) Then
                            MarkErrorLine(nline)
                        End If
                    Else
                        MarkErrorLine(nline)
                    End If
                Case Else
                    MarkErrorLine(nline)
            End Select
            ' ----------------------------------------------------------- color the comment section
            If CommentPosition > -1 Then
                MarkCommentLine(nline, CommentPosition)
            End If
            ' ----------------------------------------------------------- replace keywords
            If Not LineWithErrors Then
                ReplaceKeywords(nline, RTB_Lines(nline), CommentPosition)
            End If
        End If
    End Sub

    Private Sub MarkErrorLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_ErrorBack
        LineWithErrors = True
        ErrorsCounter += 1
        Open_NextError_Button()
        If FirstErrorLine < 0 Then FirstErrorLine = i
    End Sub
    Private Sub MarkNotFoundLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_NotFoundBack
        WarningsCounter += 1
        Open_NextError_Button()
        If FirstErrorLine < 0 Then FirstErrorLine = i
    End Sub
    Private Sub MarkWarningLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_WarningBack
        WarningsCounter += 1
        If FirstErrorLine < 0 Then FirstErrorLine = i
    End Sub
    Private Sub MarkButtonWarningLine(ByVal i As Int32)
        SelectLine(i)
        Me.SelectionBackColor = Color_ButtonWarningBack
        WarningsCounter += 1
        Open_NextError_Button()
        If FirstErrorLine < 0 Then FirstErrorLine = i
    End Sub
    Private Sub MarkCommentLine(ByVal i As Int32, Optional ByVal FirstCommentedChar As Int32 = -1)
        SelectLine(i)
        If FirstCommentedChar > 0 Then
            ' -------------------------------------------------------------------- Workaround to recalculate the FirstCommentedChar
            'FirstCommentedChar = RemoveComments(CType(Lines(i).Clone, String))
            FirstCommentedChar = RemoveComments(CType(RTB_Lines(i).Clone, String))
            ' --------------------------------------------------------------------
            SelectionLength -= FirstCommentedChar ' lenght must be changed before start
            SelectionStart += FirstCommentedChar  ' otherwise at the end of the file the lenght is shortened 
        End If
        Me.SelectionColor = Color_CommentText
        Me.SelectionBackColor = Color_CommentBack
        ' ------------------------------------------------------------------
        If RTB_Lines(i).EndsWith("'Breakpoint") Then
            SelectLine(i)
            SelectionLength = 11
            SelectionStart += RTB_Lines(i).Length - 11
            Me.SelectionColor = Color_BreakpointText
            Me.SelectionBackColor = Color_BreakpointBack
        End If
    End Sub

    Private Function DoubleQuotesAreBalanced(ByVal s As String) As Boolean
        DoubleQuotesAreBalanced = True
        For i As Int32 = 0 To s.Length - 1
            If s(i) = """"c Then
                DoubleQuotesAreBalanced = Not DoubleQuotesAreBalanced
            End If
        Next
    End Function

    Private Sub Open_NextError_Button()
        If Not Form1.chk_Suggestions.Visible Then
            SuggestionsButtonType = "Errors"
            OpenCloseSuggestionButton()
        End If
    End Sub

    Friend Sub SelectNextErrorWarningLine(Optional ByVal AllowBeep As Boolean = True)
        '
        If RTB_Lines.Length = 0 Then Return
        '
        Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ----------------------------------------------------------
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        ' --------------------------------------------------------
        ' TODO - COLOR ALL LINES - ADDED 2024-02-23
        ColorAllLines()
        ' --------------------------------------------------------
        If Not FindNextErrorWarningLine(AllowBeep) Then
            SuggestionsButtonType = ""
            OpenCloseSuggestionButton()
            RestoreSelectionAndScroll(sd)
        End If
        ' --------------------------------------------------------
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        EventsAreEnabled = OldEventsAreEnabled
        Me.Focus()
        ' -------------------------------------------------------- set cursor to the end of the line and show suggestions
        Dim line As Int32 = GetLineFromCharIndex(SelectionStart)
        SelectionStart = GetLastCharIndexFromLine(line)
        ShowSuggestions()

        ' -------------------------------------------------------- set the line selected
        SelectionStart = GetFirstCharIndexFromLine(line)
        SelectionLength = 0 ' RTB_Lines(line).Length
        ' -------------------------------------------------------- set the last character selected
        'SelectionLength = 1
    End Sub

    Private Function FindNextErrorWarningLine(ByVal AllowBeep As Boolean) As Boolean
        '
        ReinitLabelNames()
        ReinitUserVariables()
        '
        Dim StartLine As Int32 = GetLineFromCharIndex(SelectionStart)
        Dim i As Int32 = StartLine
        Do
            i += 1
            If i > RTB_Lines.Length - 1 Then
                i = 0
                If AllowBeep Then Beep()
            End If
            ' ----------------------------------------------
            Me.Select(GetFirstCharIndexFromLine(i), 0)
            Dim c As Color = Me.SelectionBackColor
            ' ----------------------------------------------
            If IsErrorColor(c) Then
                ' ------------------------------------------ FIND NEXT ERROR LINE
                TestKeywords_PrepareKeywordsToRecolor()
                If RTB_Lines(i).Trim.ToLower.StartsWith("button") Then
                    TestLines_StartingWords.Add("button")
                End If
                TestKeywords_RecolorLines()
                Me.Select(GetFirstCharIndexFromLine(i), 0)
                c = Me.SelectionBackColor
                If IsErrorColor(c) Then
                    If RTB_Lines(i).Length > 0 Then
                        ScrollToCenterLine(i)
                        Me.Select(GetFirstCharIndexFromLine(i), 0) '1)
                        Me.Focus()
                        Return True
                    End If
                End If
            End If
            ' ----------------------------------------------
            If i = StartLine Then Return False
        Loop
    End Function


    ' ===============================================================================================
    '  IDENTIFIERS
    ' ===============================================================================================
    Private IdentifiersString As String

    Private Sub InitButtonsIdentifiers()
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        IdentifiersString = " "
        Dim line As String
        For i As Int32 = 0 To RTB_Lines.Length - 1
            If RTB_Lines(i).TrimStart.StartsWith("Button") Then
                line = RTB_Lines(i).Trim.ToLower
                line = ReplaceMultipleSpacesAndTabAndTrim(line)
                Dim sa() As String = line.Split(" "c)
                If sa.Length >= 4 Then
                    If IsNumeric(sa(1)) Then
                        If IsValidButtonNumber(sa(1)) Then
                            IdentifiersString += sa(3).Replace("""", "") + " "
                        End If
                    End If
                End If
            End If
        Next
        'Form1.Text = sw.ElapsedMilliseconds.ToString("0")
    End Sub

    Private Function IsValidButtonIdentifier(ByVal IdentifierText As String) As Boolean
        ' -----------------------------------------------------------------------
        If IdentifierText Is Nothing Then Return False
        If IdentifierText = "" Then Return False
        ' ----------------------------------------------------------------------- prepare text to find
        IdentifierText = IdentifierText.ToLower.Trim
        ' ----------------------------------------------------------------------- test complete names
        If IdentifiersString.Contains(" " + IdentifierText + " ") Then
            Return True
        Else
            If IsSomeVariable(IdentifierText) Then
                Return True
            Else
                If IsValidButtonNumber(IdentifierText) Then
                    Return True
                Else
                    ' -------------------------------------- Button Identifiers can use formula (without spaces)
                    If IsValidDouble(IdentifierText) Then
                        Return True
                    Else
                        Return False
                    End If
                End If
            End If
        End If
    End Function

    ' ===============================================================================================
    '  DUPLICATE BUTTON IDENTIFIER
    ' ===============================================================================================
    Private Function DuplicateButtonIdentifier(ByVal IdentifierText As String) As Boolean
        ' -----------------------------------------------------------------------
        If IdentifierText Is Nothing Then Return False
        If IdentifierText = "" Then Return False
        ' ----------------------------------------------------------------------- prepare text to find
        IdentifierText = IdentifierText.ToLower.Trim
        ' ----------------------------------------------------------------------- test complete names
        Dim n As Int32 = InStr(IdentifiersString, " " + IdentifierText + " ")
        If n >= 1 Then
            n = InStr(n + 1, IdentifiersString, " " + IdentifierText + " ")
            If n >= 1 Then
                Return True
            End If
        End If
        Return False
    End Function



    ' ===============================================================================================
    '  MIXED FUNCTIONS
    ' ===============================================================================================
    Private Function IsValidButtonNumber(ByVal str As String) As Boolean
        If Not IsValidDouble(str) Then Return False
        Dim n As Int32 = StringToInt32(str)
        If n >= 1 And n <= Form1.NBUTTONS Then Return True
        Return False
    End Function

    Friend Function IsValidKey(ByVal KeyString As String) As Boolean
        If KeyString = Nothing Then Return False
        Select Case KeyString
            Case "num0", "num1", "num2", "num3", "num4", "num5", "num6", "num7", "num8", "num9", _
                 "left", "right", "up", "down", "pageup", "pagedown", "space", "esc", "home", "end"
                Return True
            Case Else
                If KeyString.Length = 1 Then
                    If "0123456789abcdefghijklmnopqrstuvwxyz".Contains(KeyString) Then Return True
                ElseIf KeyString.Length = 2 Then
                    If "f1f2f3f4f5f6f7f8f9".Contains(KeyString) Then Return True
                ElseIf KeyString.Length = 3 Then
                    If "f10f11f12".Contains(KeyString) Then Return True
                End If
        End Select
        Return False
    End Function

    Friend Function ExtractFirstWord(ByVal s As String) As String
        If s = Nothing Then Return ""
        s = s.Replace(vbTab, " ").Replace(vbCrLf, " ").Trim
        Dim i As Int32 = s.IndexOf(" ")
        If i < 0 Then Return s
        Return s.Substring(0, i)
    End Function

    Friend Sub SelectLine(ByVal line As Int32)
        If line >= 0 AndAlso line < RTB_Lines.Length AndAlso RTB_Lines.Length > 0 Then
            Dim lineStart As Int32 = GetFirstCharIndexFromLine(line)
            Dim lineLength As Int32 = RTB_Lines(line).Length
            Me.Select(lineStart, lineLength)
        Else
            Me.Select(0, 0)
        End If
    End Sub
    Friend Sub SelectLines(ByVal FirstLine As Int32, ByVal LastLine As Int32)
        If RTB_Lines.Length > 0 Then
            Dim FirstChar As Int32 = Me.GetFirstCharIndexFromLine(FirstLine)
            If LastLine > RTB_Lines.Length - 1 Then LastLine = RTB_Lines.Length - 1
            Dim LastChar As Int32 = Me.GetFirstCharIndexFromLine(LastLine) + RTB_Lines(LastLine).Length
            Me.Select(FirstChar, LastChar - FirstChar)
        End If
    End Sub
    Friend Function GetFirstSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart)
    End Function
    Private Function GetLastSelectedLine() As Int32
        Return Me.GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
    End Function
    Private Function GetLastSelectedLine_WithoutLineFeed() As Int32
        Dim LastCharIndex As Int32 = Me.SelectionStart + Me.SelectionLength
        If LastCharIndex > 0 And LastCharIndex < Me.Text.Length Then
            If Me.Text(LastCharIndex - 1) = vbLf Then
                LastCharIndex -= 1
            End If
        End If
        Return Me.GetLineFromCharIndex(LastCharIndex)
    End Function

    Friend Function GetFirstVisibleLine() As Int32
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    End Function
    Friend Function GetLastVisibleLine() As Int32
        Dim p As Point = New Point(1, Me.ClientSize.Height)
        Return Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(p))
    End Function
    Friend Function GetNumLineFromPoint(ByVal p As Point) As Int32
        Dim charindex As Int32 = Me.GetCharIndexFromPosition(p)
        Dim rightpos As Point = Me.GetPositionFromCharIndex(charindex)
        Return GetLineFromCharIndex(charindex)
    End Function
    Friend Sub ShowLinesCount()
        RTB_Lines = Me.Lines
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        Dim nCol As Int32 = GetColumnFromCharIndex(SelectionStart)
        Form1.Label_Lines.Text = "Ln: " + (nline + 1).ToString + _
                                 "  Col: " + (nCol + 1).ToString + _
                                 "   Lines: " + RTB_Lines.Length.ToString
    End Sub

    Friend Function GetLastCharIndexFromLine(ByVal nline As Int32) As Int32
        If RTB_Lines.Length = 0 Then Return 0
        Return Me.GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length
    End Function

    Private Function GetLineLengthAsEquivalentSpaces(ByVal nline As Int32) As Int32
        RTB_Lines = Me.Lines
        If RTB_Lines.Length = 0 Then Return 0
        Return GetColumnFromCharIndex(GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length)
    End Function

    ' -------------------------------- 15 uS (independent from n-lines) -------
    Private Function GetColumnFromCharIndex(ByVal charindex As Int32) As Int32
        Dim n As Int32 = 0
        Dim nline As Int32 = Me.GetLineFromCharIndex(charindex)
        Dim nchars As Int32 = charindex - GetFirstCharIndexFromLine(nline)
        For i As Int32 = 0 To nchars - 1
            If RTB_Lines(nline)(i) = CChar(vbTab) Then
                n += 4 - (n Mod 4)
            Else
                n += 1
            End If
        Next
        Return n
    End Function

    Friend Sub ScrollToCenterLine(ByVal line As Int32)
        Dim VisibleLinesCount As Int32 = GetLastVisibleLine() - GetFirstVisibleLine()
        Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
        EventsAreEnabled = False
        ' ---------------------------------------------------------- 
        Dim oldPaintEvent As Boolean = PaintEventEnabled
        Dim oldEnabled As Boolean = Me.Enabled
        PaintEventEnabled = False
        Me.Enabled = False
        ' ----------------------------------------------------------
        Dim delta As Int32 = VisibleLinesCount \ 2 - 1
        ScrollToLine(line + delta)
        ScrollToLine(line - delta)
        ' ---------------------------------------------------------- 
        PaintEventEnabled = oldPaintEvent
        Me.Enabled = oldEnabled
        EventsAreEnabled = OldEventsAreEnabled
    End Sub

    Friend Sub ScrollToLine(ByVal nline As Int32)
        If RTB_Lines.Length = 0 Then Return
        If nline < 0 Then nline = 0
        If nline > RTB_Lines.Length - 1 Then nline = RTB_Lines.Length - 1
        Me.Select(GetFirstCharIndexFromLine(nline), 1)
        Me.ScrollToCaret()
    End Sub

    ' ===============================================================================================
    '  SCROLL TO LEFT
    ' ===============================================================================================
    Private Declare Auto Function SendMessage_Scroll _
                Lib "user32.dll" Alias "SendMessage" _
                (ByVal hWnd As IntPtr, _
                 ByVal Msg As Integer, _
                 ByVal wParam As Int32, _
                 ByRef lParam As Int32) As Integer

    Private Const WM_HSCROLL As Integer = &H114

    Private Const SB_LEFT As Integer = 6

    ' -------------------------------------------------------  15 uS
    Friend Sub ScrollToLeft()
        SendMessage_Scroll(Me.Handle, _
                           WM_HSCROLL, _
                           SB_LEFT, _
                           0)
    End Sub

    Private Function TestEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("endif") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function
    Private Function TestElseOrEndif(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("else") Then
                If counter <= 0 Then
                    Return True
                End If
            ElseIf line.StartsWith("endif") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("if") Then
                counter += 1
            End If
        Next
        Return False
    End Function
    Private Function TestCase(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("case") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("select") Then
                counter += 1
            ElseIf line.StartsWith("endselect") Then
                counter -= 1
            End If
        Next
        Return False
    End Function
    Private Function TestEndselect(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("endselect") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("select") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Private Function TestForNext(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("next") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("for") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Private Function TestTheFinalEndForTheEventStopLabel(ByVal lnum As Int32) As Boolean
        Dim counter As Int32 = 0
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("end") And Not _
               line.StartsWith("endif") And Not _
               line.StartsWith("endselect") Then
                Return True
            ElseIf line.StartsWith("label") Or _
                    line.StartsWith("return") Then
                Return False
            End If
        Next
        Return False
    End Function

    Private Function TestNext(ByVal lnum As Int32, ByVal order As Int32) As Boolean
        If order < 1 Then order = 1
        Dim counter As Int32 = order - 1
        Dim line As String
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("next") Then
                If counter <= 0 Then
                    Return True
                End If
                counter -= 1
            ElseIf line.StartsWith("for") Then
                counter += 1
            End If
        Next
        Return False
    End Function

    Private Function CountNestedNext(ByVal lnum As Int32) As Int32
        Dim line As String
        Dim counter As Int32 = 0
        For i As Int32 = lnum + 1 To RTB_Lines.Length - 1
            line = RTB_Lines(i).Trim.ToLower
            If line.StartsWith("next") Then
                counter += 1
            ElseIf line.StartsWith("for") Then
                counter -= 1
            End If
        Next
        Return counter
    End Function

    Friend Sub AddFinalSpace_ToRow()
        Dim s As String = RTB_Lines(GetFirstSelectedLine())
        If Not s.EndsWith(" ") And Not s.EndsWith(vbTab) Then
            SelectionLength = 0
            SelectedText = " "
        End If
    End Sub

    Friend Sub SelectCompleteWord_Short()
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub

    Friend Sub SelectCompleteWord_Long()
        Dim start, length As Int32
        FindWordPointers(Me.Text, Me.SelectionStart, start, length)
        Me.SelectionStart = start
        Me.SelectionLength = length
    End Sub


    ' ===============================================================================================
    '  REPLACE KEYWORDS
    ' ===============================================================================================
    Friend Sub ReplaceKeywords(ByVal lineNumber As Int32, _
                               ByRef line As String, _
                               ByVal commentPosition As Int32)
        Dim LineStart As Int32
        If lineNumber >= 0 Then LineStart = GetFirstCharIndexFromLine(lineNumber)
        ' --------------------------------------------------------------- SPECIAL LINES
        Dim line2 As String = line.Trim.ToLower
        '
        If line2.StartsWith("beep") Then
            ReplaceKeywords(New String() {"Beep"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True)

        ElseIf line2.StartsWith("com") Then
            ReplaceKeywords(New String() {"COM", "CRLF", "Open", "Close", "WriteString", "WriteStringHEX", "WriteStringDEC", "WriteLine", _
                                          "DiscardOutBuffer", "DiscardInBuffer", _
                                          "EnableDTR", "EnableRTS", "DisableDTR", "DisableRTS"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("controls") Then
            ReplaceKeywords(New String() {"Controls", "OpenTextBox", "CloseTextBox", "ClearTextBox", "SetCursorPos"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("button") Then
            ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)
            If line2.Contains("label") Or line2.Contains("text") Then
                ReplaceKeywords(New String() {"Button", "Label", "Text"}, lineNumber, LineStart, line, commentPosition)
                ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)
                ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True)
            ElseIf line2.Contains("color") Then
                ReplaceKeywords(New String() {"Button", "Color"}, lineNumber, LineStart, line, commentPosition)
            Else
                ReplaceKeywords(New String() {"Button", "Slot", "Enabled", "Disabled"}, lineNumber, LineStart, line, commentPosition)
            End If
            If line2.Contains("color") Then
                ReplaceKeywords(GetColorNames.Split(" "c), lineNumber, LineStart, line, commentPosition, True)
                ReplaceKeywords(New String() {"Flashing", "Flashing1", "Flashing2", "Flashing3", "Flashing4", _
                                              "Flashing5", "Flashing6", "Flashing7", "Flashing8", "Flashing9"}, lineNumber, LineStart, line, commentPosition, True)
            End If
        ElseIf line2.StartsWith("for") Then
            ReplaceKeywords(New String() {"For", "To", "Step"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)                  ' REPLACE ALL KEYWORDS
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("next") Then
            ReplaceKeywords(New String() {"Next"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("exit") Then
            ReplaceKeywords(New String() {"Exit"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("gosub") Then
            ReplaceKeywords(New String() {"Gosub"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)                  ' REPLACE ALL KEYWORDS
            ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)

        ElseIf line2.StartsWith("goto") Then
            ReplaceKeywords(New String() {"Goto"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)

        ElseIf line2.StartsWith("option") Then
            ReplaceKeywords(New String() {"Option", "GlobalKeys", "Enabled", "Disabled", "Speed", "Transparency"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("return") Then
            ReplaceKeywords(New String() {"Return"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("else") Then
            ReplaceKeywords(New String() {"Else"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("endif") Then
            ReplaceKeywords(New String() {"EndIf"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("key") Then
            ReplaceKeywords(New String() {"Key", "Goto", "Gosub", "Left", "Right", "Up", "Down", "Pageup", "Pagedown", "Space", "Esc", "Home", "End", _
                                          "Num0", "Num1", "Num2", "Num3", "Num4", "Num5", "Num6", "Num7", "Num8", "Num9"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)

        ElseIf line2.StartsWith("label") Then
            ReplaceKeywords(New String() {"Label", "EventStop", "EventTimer", "Event_ExternalCommands"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("load") Then
            ReplaceKeywords(New String() {"Load", "Play", "Pause", "Position", "Stop", "Hide", "Slots", "Vars", "OpenApps", "CloseApps", "CloseAll", "Windows"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("presskeys") Then
            ReplaceKeywords(New String() {"PressKeys"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetPressKeysNames_NoLetters().Split(" "c), lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("sendkeys") Then
            ReplaceKeywords(New String() {"SendKeys"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetSendKeysNames_NoLetters().Split(" "c), lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("save") Then
            ReplaceKeywords(New String() {"Save", "Slots", "Vars", "StringToFile", "DeleteFile"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS

        ElseIf line2.StartsWith("caseelse") Then ' <-- must be before the "case"
            ReplaceKeywords(New String() {"CaseElse"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("endselect") Then
            ReplaceKeywords(New String() {"EndSelect"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("variable") Then
            ReplaceKeywords(New String() {"Variable", "Numeric", "String", "Infinity"}, lineNumber, LineStart, line, commentPosition)

        ElseIf line2.StartsWith("wait") Then
            ReplaceKeywords(New String() {"Wait", "Seconds", "Button", "Application"}, lineNumber, LineStart, line, commentPosition)
            ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)
            Dim l2 As String = ReplaceMultipleSpacesAndTabAndTrim(line2)
            If Not l2.StartsWith("wait button") And _
               Not l2.StartsWith("wait application") Then
                ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)
            End If

        ElseIf line2.StartsWith("window") Then
            ReplaceKeywords(New String() {"Window", "MinSize", "SetFocus", "Sizable", "Maximized", "FullScreen"}, lineNumber, LineStart, line, commentPosition)

        Else
            ReplaceKeywords(Keywords, lineNumber, LineStart, line, commentPosition)                  ' REPLACE ALL KEYWORDS
            ReplaceKeywords(GetAllVariableNames, lineNumber, LineStart, line, commentPosition, True) ' REPLACE VARS
            If LABELS_WITHOUT_GOSUB Then
                ReplaceKeywords(LabelNames, lineNumber, LineStart, line, commentPosition, True)      ' REPLACE LABEL NAMES
            End If
        End If
    End Sub

    Private Sub ReplaceKeywords(ByVal Keywords As String(), _
                                ByVal lineNumber As Int32, _
                                ByVal lineStart As Int32, _
                                ByRef line As String, _
                                ByVal commentPosition As Int32, _
                                Optional ByVal UseTextColor As Boolean = False)

        If commentPosition < 0 Then commentPosition = 9999
        Dim position As Int32
        For Each kw As String In Keywords
            If kw = "" Then Continue For
            position = 0
            Do
                position = line.IndexOf(kw, position, StringComparison.InvariantCultureIgnoreCase)
                If position >= 0 Then
                    If IsCompleteWord(line, position, kw.Length) And position < commentPosition Then
                        If Not InsideQuotes(position, line) Then
                            If lineNumber >= 0 Then
                                Me.Select(lineStart + position, kw.Length)
                                Me.SelectionColor = If(UseTextColor, Color_Text, Color_KeywordText)
                                Me.Select(lineStart + position, kw.Length) ' <-- redo selection after SelectionColor 
                                If kw <> Me.SelectedText Then
                                    Me.SelectedText = kw
                                End If
                            Else
                                Mid(line, position + 1, kw.Length) = kw
                            End If
                        End If
                    End If
                    position += kw.Length
                End If
            Loop Until position < 0
        Next
    End Sub

    Private Function IsCompleteWord(ByVal l As String, ByVal position As Int32, ByVal length As Int32) As Boolean
        ' ------------------------------------------------------------ test left side
        If position > 0 AndAlso _
           IsLetterOrSpaceOrNumber(l(position - 1)) Then Return False
        ' ------------------------------------------------------------- test right side
        If position + length < l.Length AndAlso _
           IsLetterOrSpaceOrNumber(l(position + length)) Then Return False
        ' -------------------------------------------------------------
        Return True
    End Function

    Private Function IsLetterOrSpaceOrNumber(ByVal c As Char) As Boolean
        Return (c >= "a" And c <= "z") OrElse (c >= "A" And c <= "Z") OrElse (c >= "0" And c <= "9") OrElse c = "_"
    End Function

    Private Function InsideQuotes(ByVal index As Int32, ByVal text As String) As Boolean
        InsideQuotes = False
        If index >= text.Length Then Return False
        For Counter As Int32 = 0 To index - 1
            If text(Counter) = """"c Then
                InsideQuotes = Not InsideQuotes
            End If
        Next
    End Function


    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  AUTO INDENT
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    Friend Sub AutoIndent()
        Dim oldEnabled As Boolean = Me.Enabled
        Dim OldPaintEventEnabled As Boolean = PaintEventEnabled
        ' ----------------------------------------------------------- save scroll position
        Dim sd As ScrollData
        SaveSelectionAndScroll(sd)
        Dim SelectionModified As Boolean = False
        ' ----------------------------------------------------------- 
        RTB_Lines = Me.Lines
        Dim indent As Int32 = 0
        For i As Integer = 0 To RTB_Lines.Length - 1
            ' -------------------------------------------------------
            Dim l() As String
            l = RTB_Lines(i).Replace(vbTab, " ").Trim.ToLower.Split(" "c)
            If l(0).StartsWith("'") Then Continue For
            ' ------------------------------------------------------- 
            Select Case l(0)
                Case "endif", "case", "caseelse", "else", "next" : indent -= 1
                Case "endselect" : indent -= 2
                Case "label" : If LineIsEmptyOrCommentOrTerminator(i - 1) Then indent = 0
                Case "goto", "stop", "end"
                    If LineIsEmptyOrCommentOrTerminator(i + 1) Then indent -= 1
                Case "return"
                    If LineIsEmptyOrCommentOrTerminatorOrLabel(i + 1) Then indent -= 1
            End Select
            ' ------------------------------------------------------- 
            Dim ntabs As Int32 = indent
            If ntabs < 0 Then ntabs = 0
            ' ------------------------------------------------------- 
            Dim lineText As String = RTB_Lines(i)
            Dim InitialTabs As Int32 = CountInitialTabs_NotAllowingSpaces(lineText)
            '
            ' ------------------------------------------------------- if changed indentation level
            If ntabs <> InitialTabs Then
                ' --------------------------------------------------- DISABLE PaintEvent AND Me.Enable
                If PaintEventEnabled Then
                    ' -----------------------------------------------------------
                    PaintEventEnabled = False
                    ' -----------------------------------------------------------
                    '  This Enabled = False is critical
                    '  changes the cursor column if called from SelectionChanged
                    '  and could prevent mouse selection in some cases
                    ' -----------------------------------------------------------
                    If Control.MouseButtons = MouseButtons.None Then
                        Me.Enabled = False
                    End If

                End If
                ' --------------------------------------------------- save values
                Dim cursorpos As Int32 = Me.SelectionStart
                Dim InitialChars As Int32 = CountInitialChars(lineText)
                Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                ' --------------------------------------------------- insert or remove TABS
                Me.Select(start, InitialChars)
                Me.SelectedText = StrDup(ntabs, vbTab)
                SelectionModified = True
                ' --------------------------------------------------- correct SelectionStart for Return
                If cursorpos >= start Then
                    Dim delta As Int32 = ntabs - InitialChars
                    If cursorpos + delta < 0 Then delta = -cursorpos
                    Me.Select(cursorpos + delta, 0)
                Else
                    Me.Select(cursorpos, 0)
                End If
            End If
            ' -------------------------------------------------------
            Select Case l(0)
                Case "if", "label", "case", "caseelse", "else", "for" : indent += 1
                Case "select" : indent += 2
            End Select
            ' ------------------------------------------------------- 
            If indent < 0 Then indent = 0
        Next
        ' ----------------------------------------------------------- restore scroll position
        If SelectionModified Then
            Dim sel As Int32 = Me.SelectionStart
            RestoreScroll(sd)
            Me.Select(sel, 0)
            TextModified = True
            TextMustBeSaved = True
        End If
        ' ----------------------------------------------------------- 
        Me.Enabled = oldEnabled
        PaintEventEnabled = OldPaintEventEnabled
        Me.Focus()
        TextMustBeIndented = False
    End Sub


    'Private Function LineIsEmptyOrCommentOrInitiator(ByVal index As Int32) As Boolean
    '    If index < 0 Then Return True
    '    If index >= RTB_Lines.Length Then Return True
    '    Dim l As String = RTB_Lines(index).Trim.ToLower
    '    If l = "" OrElse l.StartsWith("'") _
    '        OrElse l.StartsWith("label") Then Return True
    '    Return False
    'End Function

    Private Function LineIsEmptyOrCommentOrTerminator(ByVal index As Int32) As Boolean
        If index < 0 Then Return True
        If index >= RTB_Lines.Length Then Return True
        Dim l As String = RTB_Lines(index).Trim.ToLower
        If l = "" OrElse l.StartsWith("'") _
            OrElse l.StartsWith("goto") _
            OrElse l.StartsWith("return") _
            OrElse l.StartsWith("stop") _
            OrElse (l.StartsWith("end") AndAlso Not l.StartsWith("endif") AndAlso Not l.StartsWith("endselect")) Then Return True
        Return False
    End Function

    Private Function LineIsEmptyOrCommentOrTerminatorOrLabel(ByVal index As Int32) As Boolean
        If index < 0 Then Return True
        If index >= RTB_Lines.Length Then Return True
        Dim l As String = RTB_Lines(index).Trim.ToLower
        If l = "" OrElse l.StartsWith("'") _
            OrElse l.StartsWith("label") _
            OrElse l.StartsWith("goto") _
            OrElse l.StartsWith("return") _
            OrElse l.StartsWith("stop") _
            OrElse (l.StartsWith("end") AndAlso Not l.StartsWith("endif") AndAlso Not l.StartsWith("endselect")) Then Return True
        Return False
    End Function



    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  SUGGESTIONS
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Friend SuggestionsButtonType As String = ""

    Friend Sub ShowSuggestions()
        Dim nline As Int32 = GetLineFromCharIndex(SelectionStart)
        ShowLinesCount()
        If RTB_Lines.Length = 0 Then
            Form1.StatusLabel1.BackColor = Color.FromArgb(220, 255, 200)
            Form1.SetStatusLabelFontProportional()
            Form1.StatusLabel1.Text = "Keywords:  " & AllKeywords
            Return
        End If
        ' ----------------------------------------------------------- 
        ShowSuggestions(RTB_Lines(nline), nline, GetCompleteWord_Short(Me.Text, Me.SelectionStart))
    End Sub

    Friend Sub ShowSuggestions(ByVal _line As String, ByVal nline As Int32, ByVal SelectedWord As String)
        LastError = ""
        Dim line As String = _line.Replace(vbTab, " ").Trim
        RemoveComments(line)
        ' ----------------------------------------------------------- Tokenize to array of keywords 
        Dim sa() As String = Tokenize(line)
        ' ----------------------------------------------------------- sa2 contains the original case strings 
        Dim saCase() As String = CType(sa.Clone, String())
        ' ----------------------------------------------------------- set sa string array case to lower
        For i As Int32 = 0 To sa.Length - 1
            If sa(i) IsNot Nothing Then sa(i) = sa(i).ToLower
        Next
        ' ----------------------------------------------------------- 
        Dim s As String = ""
        Select Case sa.Length
            Case 0
                s = "Keywords: " + AllKeywords
            Case 1
                Select Case sa(0)
                    Case "beep" : s = "OK: A system Beep will be emitted here (repetition is limited to 10 per second)"
                    Case "button" : s = "NoKeywords: Please write a BUTTON NUMBER or a BUTTON IDENTIFIER"
                    Case "com" : s = "Keywords: Open / Close / WriteString / WriteStringHEX / WriteStringDEC / WriteLine / DiscardOutBuffer / DiscardInBuffer / EnableDTR / EnableRTS / DisableDTR / DisableRTS"
                    Case "controls" : s = "Keywords: OpenTextBox / CloseTextBox / ClearTextBox / SetCursorPos"
                    Case "for" : s = "NoKeywords: Please write For V = N to N" + vbCrLf + " V is a numeric variable (v1 to v99 or a user defined numeric variable) and N is a number or a variable."
                    Case "next" : s = "OK: The execution returns to the preceding ""For"" (if exists)."
                    Case "exit"
                        If TestNext(nline, 1) Then
                            s = "OK: The execution of the current FOR continues after the first NEXT."
                        Else
                            s = "WARNING: This line has no effect without a corresponding Next"
                        End If
                    Case "gosub" : s = "NoKeywords: Please write a Label name"
                    Case "goto" : s = "NoKeywords: Please write a Label name"
                    Case "if" : s = "NoKeywords: EXPRESSION"
                    Case "else"
                        If TestEndif(nline) Then
                            s = "OK: Following instructions are executed when the preceding IF is FALSE"
                        Else
                            s = "WARNING: This line has no effect without a corresponding EndIf"
                        End If
                    Case "endif" : s = "OK: This is the end of the conditional execution."
                    Case "endselect" : s = "OK: This is the end of the Select/Case conditional execution."
                    Case "key" : s = "Keywords: 0..9 / a..z / F1..F12 / Num0..Num9 / Left / Right / Up / Down / Pageup / Pagedown / Space / Esc / Home / End"
                    Case "label" : s = "NoKeywords: LABEL-NAME or ""EventStop"" or ""EventTimer"" or ""Event_ExternalCommands""" + vbCrLf + _
                                       "  ( the EventStop label will be called when the program stops / the EventTimer 10 times per second / the Event_ExternalCommands when there is some text on the Slot_ExternalCommands)"
                    Case "load" : s = "Keywords: " + LoadKeywords
                    Case "option" : s = "Keywords: GlobalKeys / Speed / Transparency"
                    Case "presskeys" : s = "Keywords: " + GetPressKeysNames()
                    Case "sendkeys" : s = "Keywords: " + GetSendKeysNames()
                    Case "print" : s = "NoKeywords: ""TEXT TO PRINT"" (also complex expressions - use "" to include spaces in the text)"
                    Case "return" : s = "OK: The program will return to the line after ""GoSub"", ""Button"" or ""Key"" events"
                    Case "save" : s = "Keywords: Slots / Vars / StringToFile / DeleteFile"
                    Case "select" : s = "NoKeywords: EXPRESSION"
                    Case "case" : s = "NoKeywords: Please write ""CaseElse"" or an EXPRESSION"
                    Case "caseelse" : s = "OK: Following instructions are executed if no precedent ""Case"" has matched the ""Select"" value"
                    Case "slot" : s = "NUMERIC EXPRESSION"
                    Case "slottext" : s = "NUMERIC EXPRESSION"
                    Case "stop" : s = "OK: The program will stop here (execution paused)"
                    Case "end" : s = "OK: The program will stop here (execution terminated)"
                    Case "variable" : s = "Keywords: Numeric / String"
                    Case "varsfromfile" : s = "NoKeywords: Please write a valid filename. The filename starts in the folder ""Files"" and you can use ""\"" for subfolders, or "".."" for up folders"
                    Case "wait" : s = "Keywords: Seconds / Button / Application / Condition"
                    Case "window" : s = "Keywords: SetFocus / MinSize / Sizable / Maximized / FullScreen"
                    Case "sewabort"
                        If SewFunctionsEnabled Then
                            s = "OK: The sewing thread will be aborted (if it is running)"
                        Else
                            s = "ERROR: Sew functions are enabled in the industrial version only."
                        End If
                    Case "tts" : s = "Keywords: SelectVoice / SetVolume / SetSpeed / Speak / SpeakWait / Stop / Pause / Resume"
                    Case Else
                        If IsNumericVariable(sa(0)) OrElse IsStringVariable(sa(0)) Then
                            s = "Keywords: ="
                        Else
                            If LABELS_WITHOUT_GOSUB Then
                                Dim n As Int32
                                n = FindLabel(RTB_Lines, sa(0))
                                ' ------------------------------------------------------ GOSUB WITH PARAMETERS - WITHOUT GOSUB
                                If n < 0 Then
                                    ' -------------------------------------------------- extract params and remove them from the name
                                    ExtractFunctionParams(sa(0))
                                    ' -------------------------------------------------- Test the label-name without parameters 
                                    n = GetLabelLine(sa(0))
                                    If FunctionParamsAreValid(sa(0)) Then
                                        If n > 0 Then
                                            s = "OK: The program will call the function """ + saCase(0) + """ with params."
                                            Exit Select
                                        End If
                                    Else
                                        s = "ERROR: Some function param is not valid. Valid example: FunctionName(param1)(param2)(param3)" + vbCrLf + _
                                            "                And remember that immediate strings must be quoted."
                                        Exit Select
                                    End If
                                End If

                                If n < 0 Then
                                    If sa(0).Contains(" ") Or IsSomeVariable(sa(0)) Then
                                        s = "WARNING: Label " + Quoted(saCase(0)) + " not found. The destination could be wrong at run-time."
                                    Else
                                        If line.Length > 0 Then
                                            s = "Suggestions: " + AllKeywords + " " + GetLabelNamesStartingWith(line.Trim)
                                            If s = "Suggestions: " Then
                                                's = "ERROR: Label " + Quoted(saCase(0)) + " not found."

                                                s = "Suggestions: " + AllKeywords
                                            End If
                                        End If
                                    End If
                                Else
                                    s = "OK: The program will call the subprogram starting with the label " + Quoted(saCase(0))
                                End If
                            Else
                                If IsNumericVariable(sa(0)) OrElse IsStringVariable(sa(0)) Then
                                    s = "Keywords: ="
                                Else
                                    s = "Keywords: " + AllKeywords
                                End If
                            End If
                        End If
                End Select
            Case 2
                Select Case sa(0)
                    Case "beep"
                        If (sa(1).Length > 4 And sa(1).Contains("-")) Or IsValidString(sa(1)) Then
                            s = "OK: This will emit a pure sinusoidal sound specified with a sequence of Frequency(Hz) and Duration(milliseconds) pairs separated by spaces or other separators." + vbCrLf + _
                                "In alternative you could do a single beep using two numbers (frequence and duration)"
                        Else
                            s = "ERROR: Please specify a valid string containing a sequence like this: ""440-100,300-100,600-100""" + vbCrLf + _
                                    "In alternative you could do a single beep using two numbers (frequence and duration)"
                            s = GetVariablesStartingWith(sa(1))
                            If s <> "" Then
                                s = "Suggestions: " + s
                            Else
                                s = "ERROR: Please specify the Frequency (in Hz) by a numeric value or variable."
                                Dim s2 As String = GetVariablesStartingWith(sa(1))
                                If s2 <> "" Then
                                    s = s + vbCrLf + "Suggestions: " + s2
                                End If
                            End If
                        End If
                    Case "button"
                        If IsValidButtonIdentifier(sa(1)) Then
                            s = "Keywords: Label / Slot / Text / Color / Enabled / Disabled"
                        Else
                            s = "ERROR: Invalid button number or identifier"
                        End If
                    Case "com"
                        If sa(1) = "close" Then
                            s = "OK: This line will close the COM-Port."
                        ElseIf sa(1) = "discardinbuffer" Then
                            s = "OK: This line will discard the COM-Port input buffer."
                        ElseIf sa(1) = "discardoutbuffer" Then
                            s = "OK: This line will discard the COM-Port output buffer."
                        ElseIf sa(1) = "enabledtr" Then
                            s = "OK: This line will enable the DTR output."
                        ElseIf sa(1) = "enablerts" Then
                            s = "OK: This line will enable the RTS output."
                        ElseIf sa(1) = "disabledtr" Then
                            s = "OK: This line will disable the DTR output."
                        ElseIf sa(1) = "disablerts" Then
                            s = "OK: This line will disable the RTS output."
                        Else
                            s = "Keywords: Open / Close / WriteString /  WriteStringHEX / WriteStringDEC / WriteLine / DiscardOutBuffer / DiscardInBuffer / EnableDTR / EnableRTS / DisableDTR / DisableRTS"
                        End If
                    Case "controls"
                        If sa(1) = "opentextbox" Then
                            s = "OK: This line will open the print text box"
                        ElseIf sa(1) = "closetextbox" Then
                            s = "OK: This line will close the print text box"
                        ElseIf sa(1) = "cleartextbox" Then
                            s = "OK: This line will clear the print text box"
                        ElseIf sa(1) = "setcursorpos" Then
                            s = "ERROR: The ""SetCursorPos"" must be followed by two numeric values."
                        Else
                            s = "Keywords: OpenTextBox / CloseTextBox / ClearTextBox / SetCursorPos"
                        End If
                    Case "exit"
                        Dim n As Int32 = CInt(Val(sa(1)))
                        If n < 1 Then
                            s = "ERROR: After the word ""Exit"" you could specify only an integer that is the number of For-Next structures to skip."
                        Else
                            If TestNext(nline, n) Then
                                s = "OK: The execution of the current FOR continues after the " + n.ToString + GetOrdinalSuffix(n) + " nested NEXT."
                            Else
                                n = CountNestedNext(nline)
                                s = "WARNING: The specified """ + sa(1) + """ is more than the max nesting level """ + n.ToString + """"
                            End If
                        End If
                    Case "for"
                        If IsNumericVariable(sa(1)) Then
                            s = "Keywords: ="
                        Else
                            s = "ERROR: Invalid control variable"
                        End If
                    Case "gosub"
                        If GetLabelLine(GetFunctionNameWithoutParams(line.Substring(6))) >= 0 Then
                            ' ------------------------------------------------------------------------------- GOSUB WITH PARAMETERS
                            ' -------------------------------------------------- Test the label-name without parameters 
                            If FunctionParamsAreValid(sa(1)) Then
                                Dim p() As String = ExtractFunctionParams(sa(1))
                                If p.Length > 0 Then
                                    s = "OK: The program will call the function """ + saCase(1) + """ with params."
                                Else
                                    s = "OK: The program will call the function """ + saCase(1) + """ "
                                End If
                            Else
                                s = "ERROR: Some function param is not valid. Valid example: FunctionName(param1)(param2)(param3)" + vbCrLf + _
                                    "                And remember that immediate strings must be quoted."
                            End If
                        ElseIf line.Contains("""") And Not line.Contains("+") Then
                            s = "ERROR: The character "" is not allowed in label names."
                        ElseIf FindLabel(RTB_Lines, sa(1)) < 0 Then
                            If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                s = "WARNING: Label " + Quoted(saCase(1)) + " not found. The destination could be wrong at run-time."
                            Else
                                If line.Length > 6 Then
                                    s = "Suggestions: " + GetLabelNamesStartingWith(line.Substring(6))
                                    If s = "Suggestions: " Then
                                        s = "ERROR: Label " + Quoted(saCase(1)) + " not found."
                                    End If
                                End If
                            End If
                        Else
                            s = "OK: The program will call the subprogram starting with the label " + Quoted(saCase(1))
                        End If
                    Case "goto"
                        If line.Contains("""") And Not line.Contains("+") Then
                            s = "ERROR: The character "" is not allowed in label names."
                        ElseIf FindLabel(RTB_Lines, sa(1)) < 0 Then
                            If sa(1).Contains(" ") Or IsSomeVariable(sa(1)) Then
                                s = "WARNING: Label " + Quoted(saCase(1)) + " not found. The destination could be wrong at run-time."
                            Else
                                If line.Length > 5 Then
                                    s = "Suggestions: " + GetLabelNamesStartingWith(line.Substring(5))
                                    If s = "Suggestions: " Then
                                        s = "ERROR: Label " + Quoted(saCase(1)) + " not found."
                                    End If
                                End If
                            End If
                        Else
                            s = "OK: The program will jump to the label " + Quoted(saCase(1))
                        End If
                    Case "if"
                        If IsValidBoolean(sa(1)) Then
                            If TestElseOrEndif(nline) Then
                                s = "OK: the expression """ + saCase(1) + """ will be evalued"
                            Else
                                s = "WARNING: This line has no effect without a corresponding Else or EndIf"
                            End If
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            s = "Keywords: Goto / Gosub"
                        Else
                            s = "NoKeywords: Please use a valid Key: 0..9 / a..z / F1..F12 / Num0..Num9 / Left / Right / Up / Down / Pageup / Pagedown / Space / Esc / Home / End"
                        End If
                    Case "label"
                        Dim n As Int32 = FindLabel(RTB_Lines, sa(1), False, nline)
                        If line.Contains("""") And Not line.Contains("+") Then
                            s = "ERROR: The character "" is not allowed in label names."
                        ElseIf n > -1 Then
                            If n >= 0 Then
                                s = "ERROR: The label """ + saCase(1) + """ is already declared."
                            Else
                                s = "ERROR: Expressions are not allowed."
                            End If
                        Else
                            If sa(1) = "eventstop" Then
                                If TestTheFinalEndForTheEventStopLabel(nline) Then
                                    s = "OK: This label will be called when the program stops."
                                Else
                                    s = "ERROR: This structure must be closed with ""End"""
                                End If
                            ElseIf sa(1) = "eventtimer" Then
                                Form1.EventTimer_Init()
                                If Form1.EventTimer_StartLine >= 0 Then
                                    s = "OK: This label will be called ten times per second." + vbCrLf + _
                                        "WARNING: Keep the EventTimer structure very light and fast."
                                Else
                                    s = "ERROR: The EventTimer structure must be closed with Return."
                                End If
                            ElseIf sa(1) = "event_externalcommands" Then
                                If GetNumericUserVariable("Slot_ExternalCommands") > 0 Then
                                    If Form1.GetReturnLineNumber(nline) > 0 Then
                                        s = "OK: This label will be called when there is some text on the Event_ExternalCommands." + vbCrLf + _
                                            "WARNING: This structure must clear the command with: SlotText(Slot_ExternalCommands) = """""
                                    Else
                                        s = "ERROR: The Event_ExternalCommands must terminate with a Return."
                                    End If
                                Else
                                    s = "ERROR: The Event_ExternalCommands requires a line like this: ""Variable Numeric Slot_ExternalCommands = nn""."
                                End If
                            Else
                                s = "OK: This label will be called by ""GoSub"", ""GoTo"" and by ""Buttons"" with text """ + saCase(1) + """"
                            End If
                        End If
                    Case "load"
                        If sa(1) = "slots" Then
                            s = "OK: All the slots (0 to 999) will be loaded from the ""_Slots.txt"" file."
                        ElseIf sa(1) = "vars" Then
                            s = "OK: All the vars will be loaded from the ""_Vars.txt"" file."
                        ElseIf sa(1) = "pause" Then
                            s = "OK: Sounds and video will be paused here"
                        ElseIf sa(1) = "play" Then
                            s = "OK: Sounds and video will play (if paused)"
                        ElseIf sa(1).StartsWith("position") Then
                            Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c)
                            If sa2.Length = 2 Then
                                s = "OK: Sounds and video will be moved to this position in seconds"
                            Else
                                s = "ERROR: Wrong instruction." + vbCrLf + _
                                    "SUGGESTION: Add position in seconds and fractions."
                            End If
                        ElseIf sa(1) = "stop" Then
                            s = "OK: Sounds and video will be stopped here"
                        ElseIf sa(1) = "hide" Then
                            s = "OK: The running program will be visualized until a new image or video will be loaded."
                        ElseIf sa(1) = "openapps" Then
                            s = "OK: This line will opens all the applications starting with ""Theremino_"" and ending with "".exe""."
                        ElseIf sa(1).StartsWith("vars") Then
                            If ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c).Length > 1 Then
                                s = "ERROR: Too many parameters."
                            End If
                        ElseIf sa(1).StartsWith("openapps") Then
                            If ReplaceMultipleSpacesAndTabAndTrim(sa(1)).Split(" "c).Length > 1 Then
                                s = "ERROR: Too many parameters."
                            End If
                        ElseIf sa(1).StartsWith("closeapps") Then
                            If sa(1) = "closeapps" Then
                                s = "OK: This line will close all the applications previously opened."
                            Else
                                Dim fname As String = sa(1).Substring(9).Trim.Replace("""", "")
                                If fname.ToLower = "windows" Then
                                    s = "OK --- WARNING --- THIS LINE WILL SHUT DOWN THE WINDOWS SYSTEM !!!"
                                Else
                                    s = "OK: This line will close the opened application: """ + fname + """"
                                End If
                            End If
                        ElseIf sa(1) = "closeall" Then
                            s = "OK: This line will close all the applications previously opened." + vbCrLf + _
                                "And then closes Automation too (to avoid this keep CTRL presses)"
                        ElseIf sa(1).StartsWith("slot") Then
                            Dim n() As Double = ExtractNumericParams(sa(1), 3)
                            If n.Length > 0 Then
                                If n.Length = 2 Then
                                    s = "OK: Slots in the range will be loaded."
                                Else
                                    s = "ERROR: Please write two numbers from 0 to 999, or two expressions without spaces." + vbCrLf + _
                                        "               The first and second values are the first and last Slots to be loaded."
                                End If
                            Else
                                s = "ERROR: Too many parameters."
                            End If
                        Else
                            If Not DoubleQuotesAreBalanced(sa(1)) Then
                                s = "ERROR: The number of double quotes is not balanced." + vbCrLf + _
                                    "Each string starting with double quotes must terminate with double quotes too."
                            Else
                                sa(1) = sa(1).Replace("""", "")
                                Dim str As String = sa(1)
                                Select Case IO.Path.GetExtension(str).ToLower
                                    Case ".txt"
                                        ' --------------------------------------------- test load file
                                        Dim sa2() As String = sa(1).Split(" "c)
                                        If sa2.Length > 1 Then
                                            If IsStringVariable(sa2(0)) Then
                                                s = "OK: The variable """ + sa2(0) + """ will be loaded from the file """ + sa2(sa2.Length - 1) + """"
                                            Else
                                                Dim s2 As String = EvalString_AutoQuoting(sa(1))
                                                If IsValidString_AutoQuoting(s2) Then
                                                    s = "OK: The program """ + sa(1) + """ will be loaded from ""Programs"" folder"
                                                Else
                                                    s = "ERROR: The variable """ + sa2(0) + """ is not declared"
                                                End If
                                            End If
                                        End If
                                    Case ".jpg", ".jpeg", ".bmp", ".png", ".exif", ".tiff", ".svg"
                                        s = "OK: The image will be loaded from ""Media"" folder and displayed"
                                    Case ".avi", ".wmv", ".mpg", ".mpeg", ".m2ts", ".m3u", ".mp4", ".m4v", ".mp4v", ".3g2", ".3gp2", ".3gp", ".3gpp", ".mov", ".gif"
                                        s = "OK: The video will be loaded from ""Media"" folder and played with the WMP player"
                                    Case ".mp3", ".wav", ".m4a", ".aif", ".aifc", ".aiff", ".au", ".snd", ".aac", ".adt", ".adts", ".flac"
                                        s = "OK: The audio file will be loaded from ""Media"" folder and played (with the WMP player)"
                                    Case ".mid", ".midi", ".rmi"
                                        s = "OK: The MIDI file will be loaded from ""Media"" folder and played (with the WMP player)"
                                    Case ".exe"
                                        s = "OK: The application will be started (if exists near to file 'Automation.exe' or in the 'Full-Path' specified)"
                                    Case Else
                                        If IsStringVariable(sa(1)) Then
                                            s = "OK: The image, video or program, will be loaded from ""Media"" or ""Programs"" folder"
                                        Else
                                            If str.StartsWith("http") Then
                                                s = "OK: The site will be opened (if exists)"
                                            Else
                                                s = "OK: Will try to load the """ + str + """ (if exists)" + vbCrLf
                                                s += "Suggestions: Slots / Vars / Stop / Pause / Play / Position / Hide / OpenApps / CloseApps / CloseAll"
                                            End If
                                        End If
                                End Select
                            End If
                        End If
                    Case "option"
                        Select Case sa(1)
                            Case "globalkeys" : s = "Keywords: Enabled / Disabled"
                            Case "speed" : s = "NoKeywords: Values 1 to 9 - Normal values are: 5 (slow execution), 7 (fast execution), 9 (only for short loops)"
                            Case "transparency" : s = "NoKeywords: Valid values are 0 (full opaque) to 9 (maximum transparency)"
                            Case Else : s = "GlobalKeys / Speed / Transparency"
                        End Select
                    Case "presskeys"
                        s = "Keywords: " + GetPressKeysNames()
                    Case "sendkeys"
                        s = "Keywords: " + GetSendKeysNames()
                    Case "print"
                        If IsValidBooleanStringOrDouble(sa(1)) Then
                            s = "OK: The expression """ + saCase(1) + """ will be evalued and printed"
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "save"
                        If sa(1) = "slots" Then
                            s = "OK: All the slots (0 to 999) will be saved to the ""_Slots.txt"" file."
                        ElseIf sa(1) = "vars" Then
                            s = "OK: All the vars will be saved to the ""_Vars.txt"" file."
                        Else
                            If sa(1).StartsWith("stringtofile") Then
                                s = "NoKeywords: Please write a String and a Filename"
                            Else
                                s = "Keywords: Slots / Vars / StringToFile / DeleteFile"
                            End If
                        End If
                    Case "select"
                        If TestCase(nline) And TestEndselect(nline) Then
                            If IsValidBooleanStringOrDouble(sa(1)) Then
                                s = "OK: The expression """ + saCase(1) + """ will be compared with ""Case"" expressions"
                            Else
                                s = "NoKeywords: Please write a valid expression"
                            End If
                        Else
                            s = "WARNING: This line has no effect without corresponding Case and EndSelect"
                        End If
                    Case "case"
                        If TestEndselect(nline) Then
                            If IsValidBooleanStringOrDouble(sa(1)) Then
                                s = "OK: Following instructions are executed when the ""Case"" value equals the ""Select"" value"
                            Else
                                s = "NoKeywords: Please write a valid expression"
                            End If
                        Else
                            s = "WARNING: This line has no effect without a corresponding EndSelect"
                        End If
                    Case "slot", "slottext"
                        If IsValidDouble(sa(1)) Then
                            s = "Keywords: ="
                        Else
                            s = "Suggestions: " + GetVariablesStartingWith(sa(1).Trim)
                        End If
                    Case "tts"
                        Select Case sa(1)
                            Case "stop", "pause", "resume"
                                s = "OK: This TTS operation is valid."
                            Case "selectvoice"
                                s = "NoKeywords: Please write a three letter string containing an ISO voice identifier, eventually followed by the gender, as in the following examples: " + _
                                     """ENG"" ""ENGMALE"" ""ENGM"" ""ENGF"" ""ENGN"" ""ENG Neutral"" ""FRA"" ""ITA FEM."" ""CHI"" ""CHI F"" ""ChiMale"" ""ZHO M"" ""ZHO"" ""CHI neutral"""
                            Case "setvolume"
                                s = "NoKeywords: Please write a number from 0 to 100"
                            Case "setspeed"
                                s = "NoKeywords: Please write a number from -10 to 10"
                            Case "speak", "speakwait"
                                s = "NoKeywords: Please write a string with the text to be spoken"
                            Case Else
                                s = "Keywords: Stop / Pause / Resume / SelectVoice / SetVolume / SetSpeed / Speak / SpeakWait"
                        End Select
                    Case "variable"
                        Select Case sa(1)
                            Case "numeric", "string"
                                s = "NoKeywords: Please write a name for this variable"
                            Case Else
                                s = "Keywords: Numeric / String"
                        End Select
                    Case "varsfromfile"
                        If IsValidString_AutoQuoting(sa(1)) Then
                            s = "OK: The variables will be initialized from the file: """ + saCase(1) + """"
                        Else
                            s = "ERROR: Please write a valid filename. The filename starts in the folder ""Files"" and you can use ""\"" for subfolders, or "".."" for up folders"
                        End If
                    Case "wait"
                        Select Case sa(1)
                            Case "seconds" : s = "NoKeywords: Please write a time in seconds (also with decimals)"
                            Case "button"
                                s = "NoKeywords: BUTTON-NAME"
                            Case "application"
                                If IsValidString(sa(1)) Then
                                    s = "NoKeywords: APP-NAME"
                                Else
                                    s = "ERROR: Invalid app-name."
                                End If
                            Case Else
                                If IsValidBoolean(sa(1)) Then
                                    s = "OK: This line waits until the expression """ + saCase(1) + """ will be TRUE"
                                Else
                                    s = "Keywords: Seconds / Button / Key / Application / Condition" + LastError
                                End If
                        End Select
                    Case "window"
                        If sa(1) = "setfocus" OrElse _
                           sa(1) = "minsize" OrElse _
                           sa(1) = "sizable" OrElse _
                           sa(1) = "maximized" OrElse _
                           sa(1) = "fullscreen" Then
                            s = "OK: The window will be set to " + saCase(1)
                        Else
                            s = "Keywords: SetFocus / MinSize / Sizable / Maximized / FullScreen"
                        End If
                    Case Else
                        If sa(1) = "=" Then
                            If IsNumericVariable(sa(0)) Then
                                s = "NoKeywords: Please write a valid numerical expression, or the keywords: Input or Message"
                            ElseIf IsStringVariable(sa(0)) Then
                                s = "NoKeywords: Please write a valid string (with "" initial and final), or the keywords: Input or Message"
                            Else
                                s = "ERROR: " + LastError
                            End If
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                End Select
            Case 3
                Select Case sa(0)
                    Case "beep"
                        If IsValidDouble(sa(1)) And IsValidDouble(sa(2)) Then
                            s = "OK: The Beep will emit a pure sinusoidal sound specified by Frequency(Hz) and Duration(mS)." + vbCrLf + _
                                "    You could also specify a sequence of Frequency and Duration pairs separated by spaces or other separators."
                        Else
                            s = "ERROR: Please specify the Frequency(Hz) and Duration(mS) by two numeric values or variables."
                            Dim s2 As String = GetVariablesStartingWith(sa(2))
                            If s2 <> "" Then
                                s = s + vbCrLf + "Suggestions: " + s2
                            End If
                        End If
                    Case "button"
                        If IsValidButtonIdentifier(sa(1)) Then
                            Select Case sa(2)
                                Case "enabled", "disabled"
                                    s = "OK: Button will be " + saCase(2)
                                Case "slot" : s = "NoKeywords: SLOT-NUMBER"
                                Case "text" : s = "NoKeywords: BUTTON-TEXT  (enclosing the text in double quotes is possible, but not necessary)"
                                Case "color"
                                    s = "NoKeywords: Please write ""Flashing"" or a color name like ""Yellow"", ""Blue"", ""Red""..."
                                    SuggestionsButtonType = "Colors"
                                Case Else : s = "Keywords: Label / Slot / Text / Color / Enabled / Disabled"
                            End Select
                        Else
                            s = "ERROR: Invalid button number or identifier."
                        End If
                    Case "com"
                        Select Case sa(1)
                            Case "writestring", "writestringhex", "writestringdec"
                                If IsValidString_AutoQuoting(sa(2)) Then s = "OK: The string " + saCase(2) + " will be written on the serial line"
                            Case "writeline"
                                If IsValidString_AutoQuoting(sa(2)) Then s = "OK: The string " + saCase(2) + " will be written on the serial line, followed by a final CRLF"
                            Case Else : s = "Keywords: Open / WriteString /  WriteStringHEX / WriteStringDEC / WriteLine"
                        End Select
                    Case "controls"
                        If sa(1) = "setcursorpos" Then
                            s = "ERROR: The ""SetCursorPos"" must be followed by two numeric values."
                        Else
                            s = "ERROR:  Incomplete or wrong instruction."
                        End If
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" Then
                            s = "NoKeywords: Please write a numeric expression (without spaces), or a numeric variable."
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            If line.Contains("""") And Not line.Contains("+") Then
                                s = "ERROR: The character "" is not allowed in label names."
                            ElseIf FindLabel(RTB_Lines, sa(2)) < 0 Then
                                If sa(2).Contains(" ") Or IsSomeVariable(sa(2)) Then
                                    s = "WARNING: Label " + Quoted(saCase(2)) + " not found. The destination could be wrong at run-time."
                                Else
                                    s = "Suggestions: " + GetLabelNamesStartingWith(sa(2))
                                    If s = "Suggestions: " Then
                                        s = "ERROR: Label " + Quoted(saCase(2)) + " not found."
                                    End If
                                End If
                            Else
                                s = "OK: Pressing the key """ + saCase(1) + """ the Label """ + saCase(2) + """ will be called by a GoSub"
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If

                    Case "option"
                        Select Case sa(1)
                            Case "globalkeys"
                                Select Case sa(2)
                                    Case "enabled" : s = "OK: The keyboard keys are enabled even when this application is not focused (not selected)"
                                    Case "disabled" : s = "OK: The keyboard keys are disabled when this application is not focused (not selected)"
                                    Case Else
                                        s = "Suggestions: Enabled / Disabled"
                                End Select
                            Case "speed"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The speed will be set to """ + saCase(2) + """"
                                Else
                                    s = "NoKeywords: Please write 1 to 9 or a valid numeric expression"
                                End If
                            Case "transparency"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The transparency will be set to """ + saCase(2) + """"
                                Else
                                    s = "NoKeywords: Please write 0 to 9 or a valid numeric expression"
                                End If
                            Case Else
                                s = "ERROR:  Incomplete or wrong instruction."
                        End Select
                    Case "save"
                        If sa(1) = "deletefile" And IsValidString(sa(2)) Then
                            s = "OK: The file " + Quoted(saCase(2).Trim) + " will be deleted"
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case "slot"
                        If IsValidDouble(sa(1)) Then
                            If sa(2) = "=" Then
                                s = "NoKeywords: Please write RESET, SLEEP, RECOGNIZE, CALIBRATE or a valid numerical expression"
                            Else
                                s = "NoKeywords: Please write a ""="""
                            End If
                        Else
                            s = "ERROR: The slot number is not valid"
                        End If
                    Case "slottext"
                        If IsValidDouble(sa(1)) Then
                            If sa(2) = "=" Then
                                s = "NoKeywords: Please write a valid string expression (max length = 2046 chars)"
                            Else
                                s = "NoKeywords: Please write a ""="""
                            End If
                        Else
                            s = "ERROR: The slot number is not valid"
                        End If
                    Case "tts"
                        Select Case sa(1)
                            Case "selectvoice"
                                If IsValidString(sa(2)) Then
                                    s = "OK: The TTS voice """ + saCase(2).Replace("""", "") + """ will be selected (if installed)."
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "setvolume"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The TTS volume will be set to """ + saCase(2) + """"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "setspeed"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The TTS speed will be set to """ + saCase(2) + """"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "speak"
                                If IsValidString_AutoQuoting(sa(2)) Then
                                    s = "OK: The TTS will speak the text: """ + saCase(2) + """"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "speakwait"
                                If IsValidString_AutoQuoting(sa(2)) Then
                                    s = "OK: The TTS will speak and wait the text: """ + saCase(2) + """"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case Else
                                s = "ERROR: Invalid TTS instruction"
                        End Select
                    Case "wait"
                        Select Case sa(1)
                            Case "seconds"
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: The program will be paused for """ + saCase(2) + """ seconds"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            Case "button"
                                s = "OK: The program will wait the button """ + saCase(2) + """ if exists. (for buttons with long names the first word is used)"
                            Case "application"
                                If IsValidString_AutoQuoting(sa(2)) Then
                                    s = "OK: The program will wait the application """ + EvalString(sa(2)) + """"
                                Else
                                    s = "ERROR: Invalid application name."
                                End If
                            Case Else
                                s = "OK: The expression """ + saCase(2) + """ will be evaluated."
                        End Select
                    Case "variable"
                        Select Case sa(1)
                            Case "numeric"
                                ' ------------------------------------- suggestion: OK or ERROR
                                If IsNumericUserVariable(sa(2)) Then
                                    If IsNumericUserVariable(sa(2), nline) Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: The variable """ + saCase(2) + """ is already declared."
                                    End If
                                Else
                                    s = TestUserVariableName(sa(2))
                                    If s = "OK" Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: " + s
                                    End If
                                End If
                            Case "string"
                                ' ------------------------------------- suggestion: OK or ERROR 
                                If IsStringUserVariable(sa(2)) Then
                                    If IsStringUserVariable(sa(2), nline) Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: The variable """ + saCase(2) + """ is already declared."
                                    End If
                                Else
                                    s = TestUserVariableName(saCase(2))
                                    If s = "OK" Then
                                        If UnusedVariable(sa(2), nline) Then
                                            s = "WARNING: The label " + Quoted(saCase(2)) + " is never used."
                                        Else
                                            s = "OK: The variable """ + saCase(2) + """ is valid."
                                        End If
                                    Else
                                        s = "ERROR: " + s
                                    End If
                                End If
                            Case Else
                                s = "ERROR: Please correct with ""Numeric"" or ""String"""
                        End Select
                    Case Else
                        If sa(1) = "=" Then
                            If IsNumericVariable(sa(0)) Then
                                If IsValidDouble(sa(2)) Then
                                    s = "OK: Numeric variable """ + saCase(0) + """ = " + saCase(2)
                                ElseIf IsValidBoolean(sa(2)) Then
                                    s = "OK: Boolean expression """ + saCase(0) + """ = " + saCase(2)
                                ElseIf sa(2) = "input" Then
                                    s = "OK: Numeric variable """ + saCase(0) + """ will be manually entered."
                                ElseIf sa(2) = "message" Then
                                    s = "OK: The message will be prompted and the variable """ + saCase(0) + """ will contain 0 or 1 for True/False."
                                ElseIf sa(2).StartsWith("i") Then
                                    s = "Keywords: Input / Message"
                                Else
                                    s = "ERROR: " + LastError
                                End If
                            ElseIf IsStringVariable(sa(0)) Then
                                If IsValidString(sa(2)) Then
                                    s = EvalString(saCase(2))
                                    If s = "" Then s = saCase(2)
                                    s = "OK: String variable """ + saCase(0) + """ = " + Quoted(s)
                                ElseIf sa(2) = "input" Then
                                    s = "OK: String variable """ + saCase(0) + """ will be manually entered."
                                ElseIf sa(2) = "message" Then
                                    s = "OK: The message will be prompted and the variable """ + saCase(0) + """ will contain True or False."
                                ElseIf sa(2).StartsWith("i") Then
                                    s = "Keywords: Input / Message"
                                Else
                                    s = "ERROR: Invalid string. " + LastError
                                End If
                            Else
                                s = "ERROR: The variable """ + saCase(0) + """ is not declared."
                            End If
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                End Select
            Case 4
                Select Case sa(0)
                    Case "button"
                        If IsValidButtonIdentifier(sa(1)) Then
                            Select Case sa(2)
                                Case "label", "text"
                                    If FindLabel(RTB_Lines, sa(3), True) < 0 Then
                                        If sa(3).Contains(" ") Or _
                                          (sa(3).StartsWith("""") And sa(3).EndsWith("""")) Or _
                                           sa(3).Contains("+") Or IsSomeVariable(sa(3)) Then
                                            If sa(2) = "text" Then
                                                ' OK if Button text includes spaces or "+" or some variable (button used as label)
                                                s = "OK: The Button text will be " + Quoted(saCase(3)) + vbCrLf + _
                                                    "        In many cases it is better to change ""Text"" with ""Label""."
                                            Else
                                                s = "OK: This Button label is " + Quoted(saCase(3)) + vbCrLf + _
                                                    "        Use this label instead of the button number in the all the Button commands."
                                            End If
                                        Else
                                            If sa(3).Replace("""", "").Length > 0 Then
                                                s = "WARNING: Label " + Quoted(saCase(3)) + " not found." + vbCrLf + _
                                                    "( to suppress this message you could enclose all the name in double quotes )"
                                            Else
                                                s = "OK: This instruction will hide the button """ + saCase(1) + """"
                                            End If
                                        End If
                                    Else
                                        Dim s2 As String = GetLabelNamesStartingWith(sa(3))
                                        If FindLabel(RTB_Lines, sa(3).Replace("""", "")) >= 0 Then
                                            's = "OK: Button """ + saCase(1) + """ text is """ + saCase(3) + _
                                            '    """ --- Clicking this button the program jumps to the label """ + saCase(3) + """" + vbCrLf + _
                                            '    "       (the first part of this text is valid for the corrispondence with ""Labels"" and ""Wait Button"")"

                                            If sa(2) = "text" Then
                                                ' OK if Button text includes spaces or "+" or some variable (button used as label)
                                                s = "OK: The Button text will be " + Quoted(saCase(3)) + vbCrLf + _
                                                    "        In many cases it is better to change ""Text"" with ""Label""."
                                            Else
                                                s = "OK: This Button label is " + Quoted(saCase(3)) + vbCrLf + _
                                                    "        Use this label instead of the button number in the all the Button commands."
                                            End If

                                        Else
                                            s = GetLabelNamesStartingWith(sa(3).Replace("""", ""))
                                            If s <> "" Then
                                                s = "Suggestions: " + s
                                            Else
                                                s = "WARNING: Label " + Quoted(saCase(3)) + " not found." + vbCrLf + _
                                                    "( to suppress this message you could enclose all the name in double quotes )"
                                            End If
                                        End If
                                    End If
                                    If DuplicateButtonIdentifier(sa(3)) Then s = "ERROR: Duplicate identifier"
                                Case "slot"
                                    Dim slot As Int32 = CInt(EvalDouble(sa(3)))
                                    If slot > 0 And slot <= 999 Or IsSomeVariable(sa(3)) Then
                                        s = "OK: Button """ + saCase(1) + """ will be pressed by slot """ + saCase(3) + """ (when value exceeds 500)"
                                    Else
                                        s = "ERROR: Invalid Slot"
                                    End If
                                Case "color"
                                    Dim sa2() As String = ReplaceMultipleSpacesAndTabAndTrim(sa(3)).Split(" "c)
                                    If sa(3).StartsWith("flashing") Then
                                        If sa2.Length = 1 Then
                                            s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                            SuggestionsButtonType = "Colors"
                                        ElseIf sa2.Length = 2 Then
                                            s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                            SuggestionsButtonType = "Colors"
                                        ElseIf sa2.Length = 3 Then
                                            If ColorFromColorName(sa2(1)) = Color.Empty Or _
                                               ColorFromColorName(sa2(2)) = Color.Empty Then
                                                s = "Keywords: " + GetColorNames.Replace(" ", " / ")
                                                SuggestionsButtonType = "Colors"
                                            Else
                                                s = "OK: The button """ + saCase(1) + """ background will be set with the flashing colors: """ + saCase(3) + """" + vbCrLf + _
                                                    "(use ""Flashing1"" to ""Flashing9"" to change the flashing speed)"
                                            End If
                                        Else
                                            s = "ERROR: Incomplete or wrong instruction. "
                                        End If
                                    Else
                                        If sa2.Length < 2 Then
                                            If ColorFromColorName(sa(3)) = Color.Empty Then
                                                s = "Keywords: Flashing / " + GetColorNames.Replace(" ", " / ")
                                                SuggestionsButtonType = "Colors"
                                            Else
                                                s = "OK: The button """ + saCase(1) + """ background will be set with the color: """ + saCase(3) + """"
                                            End If
                                        Else
                                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                                        End If
                                    End If
                                Case Else
                                    s = "ERROR: Incomplete or wrong instruction. " + LastError
                            End Select
                        Else
                            s = "ERROR: Invalid button number or identifier."
                        End If
                    Case "com"
                        If sa(1) = "open" Then
                            If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                s = "OK: The COM port " + sa(2) + " will be opened with Bauds " + sa(3)
                            Else
                                s = "ERROR: " + LastError
                            End If
                        End If


                    Case "controls"
                        If sa(1) = "setcursorpos" Then
                            If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                                s = "OK: This line will set the cursor position"
                            Else
                                s = "ERROR: SetCursorPos must be followed by two numeric values."
                            End If
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If

                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            s = "Keywords: To"
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                    Case "key"
                        If IsValidKey(sa(1)) Then
                            If line.Contains("""") And Not line.Contains("+") Then
                                s = "ERROR: The character "" is not allowed in label names."
                            ElseIf FindLabel(RTB_Lines, sa(3)) < 0 Then
                                If sa(3).Contains(" ") Or IsSomeVariable(sa(3)) Then
                                    s = "WARNING: Label " + Quoted(saCase(3)) + " not found. The destination could be wrong at run-time."
                                Else
                                    s = "Suggestions: " + GetLabelNamesStartingWith(sa(3))
                                    If s = "Suggestions: " Then
                                        s = "ERROR: Label " + Quoted(saCase(3)) + " not found."
                                    End If
                                End If
                            Else
                                If sa(2) = "goto" Then
                                    s = "OK: Pressing the key """ + saCase(1) + """ the Label """ + saCase(3) + """ will be called by a GoTo"
                                ElseIf sa(2) = "gosub" Then
                                    s = "OK: Pressing the key """ + saCase(1) + """ the Label """ + saCase(3) + """ will be called by a GoSub (remember to use a Return)"
                                Else
                                    s = "Keywords: Goto / Gosub"
                                End If
                            End If
                        Else
                            s = "ERROR: The key is not valid"
                        End If
                    Case "load"
                        If IsValidDouble(sa(2)) And IsValidDouble(sa(3)) Then
                            s = "OK: This statement will load Slots from " + saCase(2).Trim + " to " + saCase(3).Trim
                        Else
                            s = "ERROR: Please write two simple numeric values from 0 to 999, or variables V1 to V99 " + vbCrLf + _
                                "                (the first value is first Slot and the second value is the last Slot to load)"
                        End If
                    Case "save"
                        If sa(1) = "stringtofile" And IsValidString(sa(2)) And IsValidString(sa(3)) Then
                            s = "OK: The string " + Quoted(saCase(2).Trim) + " will be appended to the file: " + Quoted(saCase(3).Trim)
                        Else
                            s = "ERROR: Please write a string to write and a string with the filename"
                        End If
                    Case "slot"
                        If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) Then
                            s = "OK: Slot """ + saCase(1).Trim + """ = expression: """ + saCase(3).Trim + """"
                        Else
                            Select Case sa(3).Trim
                                Case "reset", "sleep", "recognize", "calibrate"
                                    s = "OK: Slot """ + saCase(1).Trim + """ will be set with the special value: NAN+" + saCase(3).Trim
                                Case Else
                                    s = "ERROR: Please write RESET, SLEEP, RECOGNIZE, CALIBRATE or a valid numeric expression"
                            End Select
                        End If
                    Case "slottext"
                        If IsValidDouble(sa(1)) And sa(2) = "=" And IsValidString_AutoQuoting(sa(3)) Then
                            s = "OK: SlotText """ + saCase(1).Trim + """ = expression: " + saCase(3).Trim + ""
                        Else
                            s = "ERROR:  Incomplete or wrong instruction."
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
                Select Case sa(2)
                    Case "input"
                        s = "OK: The variable """ + saCase(0) + """ will be manually entered." + vbCrLf + _
                            "         (after the keyword ""Input"" you could also write a prompt for the user)"
                    Case "message"
                        s = "OK: The message will be prompted and the variable """ + saCase(0) + _
                            """ will contain 0 or 1 for True/False."
                End Select
            Case 5
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And IsValidDouble(sa(3)) And sa(4) = "to" Then
                            s = "NoKeywords: Please write a numeric expression (without spaces), or a numeric variable."
                        Else
                            s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End If
                    Case "variable"
                        ' -------------------------------------------- Immediate vars suggestions
                        Select Case sa(1)
                            Case "numeric"
                                If Not IsNumericUserVariable(sa(2), nline) Then s = "ERROR: Duplicate variable name" : Exit Select
                                If sa(3) <> "=" Then s = "ERROR: Invalid assignment" : Exit Select
                                If Not IsValidDouble(sa(4)) Then s = "ERROR: Invalid number" : Exit Select
                                If UnusedVariable(sa(2), nline) Then
                                    s = "WARNING: Unused variable"
                                Else
                                    s = "OK: Numeric variable """ + saCase(2) + """ initialized with " + saCase(4)
                                End If
                            Case "string"
                                If Not IsStringUserVariable(sa(2), nline) Then s = "ERROR: Duplicate variable name" : Exit Select
                                If sa(3) <> "=" Then s = "ERROR: Invalid assignment" : Exit Select
                                If Not IsValidString_AutoQuoting(sa(4)) Then s = "ERROR: Invalid string" : Exit Select
                                If UnusedVariable(sa(2), nline) Then
                                    s = "WARNING: Unused variable"
                                Else
                                    s = "OK: String variable """ + saCase(2) + """ initialized with " + Quoted(sa(4))
                                End If
                            Case Else
                                s = "ERROR: Incomplete or wrong instruction. " + LastError
                        End Select
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 6
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) Then
                            If sa(2) = "=" And IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                                If TestForNext(nline) Then
                                    s = "OK: Initially the variable """ + sa(1) + """ is set with the """ + sa(3) + """ numeric value." + vbCrLf
                                    s += "       Then the lines between For and Next will be executed until """ + sa(1) + """ reaches or exceeds """ + sa(5) + """"
                                Else
                                    s = "ERROR: This FOR has not the closing NEXT"
                                End If
                            Else
                                s = "ERROR: " + LastError
                            End If
                        Else
                            s = "ERROR: The control variable """ + sa(1) + """ is not declared (or not numeric)."
                        End If
                    Case "sewinit"
                        If SewFunctionsEnabled Then
                            s = "OK: The ""SewInit"" instruction is correct."
                            For i As Int32 = 1 To 5
                                If Not IsValidDouble(sa(i)) Then
                                    s = "ERROR: The variable """ + saCase(i) + """ is not numeric or not declared."
                                End If
                            Next
                        Else
                            s = "ERROR: Sew functions are enabled in the industrial version only."
                        End If
                    Case "sewto"
                        If SewFunctionsEnabled Then
                            s = "OK: This instruction will ""Sew To XY"" using the Sew-Inizialized Slots."
                            For i As Int32 = 1 To 4
                                If Not IsValidDouble(sa(i)) Then
                                    s = "ERROR: The variable """ + saCase(i) + """ is not numeric or not declared."
                                End If
                            Next
                            If Not IsValidString_AutoQuoting(sa(5)) Then
                                s = "ERROR: The variable """ + saCase(5) + """ is not a valid string."
                            End If
                        Else
                            s = "ERROR: Sew functions are enabled in the industrial version only."
                        End If
                    Case "sewlimits"
                        If SewFunctionsEnabled Then
                            s = "OK: The ""SewLimits"" instruction is correct."
                            For i As Int32 = 1 To 5
                                If Not IsValidDouble(sa(i)) Then
                                    s = "ERROR: The variable """ + saCase(i) + """ is not numeric or not declared."
                                End If
                            Next
                        Else
                            s = "ERROR: Sew functions are enabled in the industrial version only."
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 7
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                           IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) Then
                            If sa(6) = "step" Then
                                s = "OK: Please write a step value"
                            Else
                                s = "OK: Please write the word ""STEP"", and a step value"
                            End If
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case 8
                Select Case sa(0)
                    Case "for"
                        If IsNumericVariable(sa(1)) And sa(2) = "=" And _
                           IsValidDouble(sa(3)) And sa(4) = "to" And IsValidDouble(sa(5)) And _
                           sa(6) = "step" And IsValidDouble(sa(7)) Then
                            If TestForNext(nline) Then
                                s = "OK: Initially the variable """ + sa(1) + """ is set with the """ + sa(3) + """ numeric value." + vbCrLf
                                s += "       Then the variable """ + sa(1) + """ will be incremented by """ + sa(7) + """ until reaches or exceeds """ + sa(5) + """"
                            Else
                                s = "ERROR: This FOR has not the closing NEXT"
                            End If
                        Else
                            s = "ERROR: " + LastError
                        End If
                    Case Else
                        s = "ERROR: Incomplete or wrong instruction. " + LastError
                End Select
            Case Else
                s = "ERROR: Incomplete or wrong instruction. " + LastError
        End Select
        ' -----------------------------------------------------------
        Form1.SetStatusLabelFontProportional()
        s = Strings.Left(s, 999)
        ' -----------------------------------------------------------
        With Form1.StatusLabel1
            ' ------------------------------------------------------- status label colors
            If s.StartsWith("OK") Then
                If s.ToLower.Contains("suggestions:") Then
                    .BackColor = Color.FromArgb(210, 255, 150)
                Else
                    .BackColor = Color.FromArgb(250, 255, 200)
                End If
                .Text = Strings.Left(s, 999)
            ElseIf s.ToLower.StartsWith("error") Then
                If s.ToLower.Contains("suggestions:") Then
                    .BackColor = Color.FromArgb(210, 255, 150)
                Else
                    .BackColor = Color.FromArgb(255, 160, 150)
                End If
                .Text = s
                SuggestionsButtonType = "Errors"
            ElseIf s.ToLower.StartsWith("warning") Then
                .BackColor = Color_WarningBack
                .Text = s
                SuggestionsButtonType = ""
            ElseIf s = "" Then
                .BackColor = Color.AliceBlue
                .Text = ""
            Else
                If sa.Length > 0 Then
                    ' ----------------------------------------------- Keyword auto selection
                    SelectedWord = SelectedWord.Replace("""", "")
                    SelectedWord = SelectedWord.Replace("(", "")
                    Dim s2 As String = ""
                    If s.ToLower.StartsWith("nokeywords: ") Then
                        s2 = s
                    ElseIf s.ToLower.Contains("suggestions:") Then
                        s2 = "Suggestions: " + ExtractKeywords(s, SelectedWord)
                    Else
                        s2 = ExtractKeywords(s, SelectedWord)
                    End If
                    If s2 <> "" Then
                        s = s2
                        If s2.ToLower.StartsWith("nokeywords: ") Then
                            s = s.Substring(12)
                        Else
                            If Not s.ToLower.Trim.StartsWith("keywords") And Not s.ToLower.Contains("suggestions:") Then
                                s = "Keywords: " + s
                            End If
                        End If
                    Else
                        If sa(0) IsNot Nothing AndAlso _
                                 (sa(0).ToLower = "sendkeys" Or sa(0).ToLower = "presskeys") Then
                            s = "OK: The statement is valid"
                        Else
                            If s.Contains("Suggestions:") Then
                            Else
                                .BackColor = Color.FromArgb(255, 160, 150)
                                .Text = "ERROR:  Incomplete or wrong instruction. " + LastError
                                SuggestionsButtonType = "Errors"
                            End If
                        End If
                    End If
                End If
                ' --------------------------------------------------- corrections
                If s.ToLower.Trim = "suggestions:" Then
                    s = ""
                End If
                If s.ToLower.Trim.StartsWith("suggestions: suggestions:") Then
                    s = s.Substring(13)
                End If
                If s.Trim.StartsWith("Suggestions: /") Then
                    s = s.Replace("Suggestions: /", "Suggestions:")
                End If
                ' --------------------------------------------------- Keyword suggestions
                .BackColor = Color.FromArgb(220, 255, 200)
                .Text = s
            End If
            ' ------------------------------------------------------- No color-button if the cursor is in the first part of the line or after the comment char
            If SuggestionsButtonType = "Colors" Then
                Dim i As Int32 = Me.SelectionStart - GetFirstCharIndexOfCurrentLine()
                Dim commentpos As Int32 = _line.IndexOf("'")
                If commentpos < 0 Then commentpos = 9999
                If i < line.Length Or i > commentpos Then
                    SuggestionsButtonType = ""
                End If
            End If
            ' ------------------------------------------------------- If not "color" or "error" then Hide the Suggestions-Button
            If Not s.ToLower.StartsWith("ok:") AndAlso _
               (s.ToLower.Contains("suggestion") Or s.ToLower.Contains("error") Or _
                (s.ToLower.Contains("keywords") AndAlso IsErrorColor(Me.SelectionBackColor))) Then
                '
                If sa.Length >= 3 AndAlso sa(0) = "button" AndAlso sa(2) = "color" Then
                    SuggestionsButtonType = "Colors"
                End If
            Else
                If Form1.chk_Suggestions.Visible Then
                    If Form1.chk_Suggestions.Text = "Color selector" AndAlso s.ToLower.StartsWith("ok") Then
                        SuggestionsButtonType = ""
                    End If
                End If
            End If
            ' ------------------------------------------------------- open-close the Suggestions-Button 
            OpenCloseSuggestionButton()
        End With
    End Sub

    Private Sub OpenCloseSuggestionButton()
        Select Case SuggestionsButtonType
            Case "Errors"
                Form1.chk_Suggestions.Text = "Select next error"
                Form1.chk_Suggestions.Visible = True
                Form1.chk_Suggestions.BackColor = Color_ErrorBack
            Case "Colors"
                Form1.chk_Suggestions.Text = "Color selector"
                Form1.chk_Suggestions.Visible = True
                Form1.chk_Suggestions.BackColor = Color_WarningBack
            Case Else
                Form1.chk_Suggestions.Visible = False
                Form1.chk_Suggestions.BackColor = Color.FromArgb(210, 240, 190)
        End Select
        Form1.ResizeStatusLabel()
    End Sub

    Private Function Quoted(ByVal s As String) As String
        If Not s.StartsWith("""") And Not s.EndsWith("""") Then
            Return """" + s + """"
        End If
        Return s
    End Function

    Private Function UnusedVariable(ByVal s As String, ByVal nline As Int32) As Boolean
        Dim l As String
        ' ------------------------------------------------ add spaces before and after 
        s = " " + s + " "
        ' ------------------------------------------------ test all the lines
        For i As Int32 = 0 To RTB_Lines.Length - 1
            If i = nline Then Continue For
            l = RTB_Lines(i).Trim.ToLower
            ' -------------------------------------------- string with spaces around operators
            l = StringSpacedAroundOperators(l)
            ' -------------------------------------------- 
            If l.Contains(s) Then
                If l.StartsWith("beep ") Then Continue For
                If l.StartsWith("key ") Then Continue For
                If l.StartsWith("label ") Then Continue For
                If l.StartsWith("variable ") Then Continue For
                If l.IndexOf("'") >= 0 AndAlso l.IndexOf(s) > l.IndexOf("'") Then Continue For
                Return False
            End If
        Next
        Return True
    End Function



    ' ===============================================================================================
    '  AUTOMATIC COMPLETION
    ' ===============================================================================================
    Private Function ExtractKeywords(ByVal s1 As String, ByVal s2 As String) As String
        If s2 = "" Then Return s1
        ExtractKeywords = s1
        Dim MatchesCount As Int32 = 0
        s1 = s1.Replace("Keywords:", "")
        's1 = s1.Replace("Suggestions:", "")
        Dim s() As String = Split(s1.Replace(" / ", " "), " ")
        s1 = ""
        For Each word As String In s
            If word.StartsWith(s2, StringComparison.InvariantCultureIgnoreCase) Then
                If s1 <> "" Then s1 += " / "
                s1 += word
            End If
        Next
        Return s1
    End Function

    Private Sub CompleteKeyword(ByVal keyword As String, ByVal AddSpace As Boolean)
        ' --------------------------------------------------------------- save
        Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
        Dim OldPaintEventEnabled As Boolean = PaintEventEnabled
        EventsAreEnabled = False
        PaintEventEnabled = False
        ' No "Enabled = false" here otherwise the scroll bar flashes
        Dim OldCursorPosition As Integer = SelectionStart
        ' --------------------------------------------------------------- change
        Dim start, length As Int32
        FindWordPointers_ShortWords(Me.Text, Me.SelectionStart, start, length)
        SelectionStart = start
        SelectionLength = length
        If AddSpace Then
            If keyword.ToLower.StartsWith(SelectedText.ToLower) Then
                SelectedText = keyword & " "
            Else
                SelectedText = SelectedText + " " + keyword & " "
            End If
        Else
            SelectedText = keyword
        End If
        ' --------------------------------------------------------------- restore
        SelectionLength = 0
        PaintEventEnabled = OldPaintEventEnabled
        EventsAreEnabled = OldEventsAreEnabled
        ' --------------------------------------------------------------- update suggestions
        SelectionStart = SelectionStart - 1
        SelectionStart = SelectionStart + 1
    End Sub

    Friend Sub CompleteKeyword()
        Dim s As String = Form1.StatusLabel1.Text.Trim
        If s.Contains("/") Then Return
        If s.ToLower.StartsWith("keywords") Then
            Dim sa() As String
            sa = s.Split(" "c)
            If sa.Length > 1 And sa.Length < 8 Then
                CompleteKeyword(sa(sa.Length - 1).Trim, True)
                ColorSelectedLine()
            End If
        Else
            Dim n As Int32 = s.IndexOf("Suggestions")
            If n >= 0 Then
                s = Mid(s, n + 14).Trim
                Dim sa() As String
                sa = s.Split(" "c)
                If sa.Length = 1 Then
                    CompleteKeyword(sa(0).Trim, True)
                    ColorSelectedLine()
                End If
            End If
        End If
    End Sub

    Friend Sub CompleteKeyword_MouseSelected(ByVal str As String)
        CompleteKeyword(str, True)
        ColorSelectedLine()
        ShowSuggestions()
    End Sub


    ' ===============================================================================================
    '  CUT COPY and PASTE
    ' ===============================================================================================
    Friend Sub ExecCut()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
            ' ---------------------------------------------------------- 
            ExecDelete()
        End If
    End Sub
    Friend Sub ExecCopy()
        If Me.SelectionLength > 0 Then
            SendTextToClipboard()
        End If
    End Sub
    Private Sub SendTextToClipboard()
        ' ------------------------------------------------------- clipboard text must have the CRLF
        Dim s As String = Me.SelectedText.Replace(vbLf, vbCrLf)
        ' ------------------------------------------------------- send to clipboard
        Dim data As IDataObject
        data = New DataObject(DataFormats.Text, s)
        Clipboard.SetDataObject(data, True)
    End Sub

    Friend Sub ExecPaste()
        ' ----------------------------------------------------------------
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        ' ----------------------------------------------------------------
        If Clipboard.GetDataObject().GetDataPresent(DataFormats.Text) Then
            ' ------------------------------------------------------------ PUSH UNDO
            PushUndoState()
            ' ------------------------------------------------------------ gray backcolor
            Dim oldcolor As Color = BackColor
            BackColor = Color.FromArgb(255, 230, 190)
            Refresh()
            ' ------------------------------------------------------------ wait cursor
            Cursor.Current = Cursors.WaitCursor
            Refresh()
            Application.DoEvents()
            WaitKeyOff(Keys.V)
            ' ------------------------------------------------------------ 
            TestKeywords_PrepareKeywordsToRecolor()
            ' ------------------------------------------------------------ PaintEvent and Event (BUT NOT Enabled)
            Dim oldEventsAreEnabled As Boolean = EventsAreEnabled
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            PaintEventEnabled = False
            EventsAreEnabled = False
            ' ------------------------------------------------------------ correct CRLF from clipbbard to single LF
            Dim TextToPaste As String = Clipboard.GetText.Replace(vbCrLf, vbLf)
            '
            ' ------------------------------------------------------------ prepare StartLine and EndLine
            Dim StartSelectionPoint As Int32 = SelectionStart
            Dim StartLine As Int32 = GetLineFromCharIndex(SelectionStart)
            Dim EndLine As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            '
            ' ------------------------------------------------------------ paste the text and show immediately
            PaintEventEnabled = True
            EventsAreEnabled = True
            Me.SelectedText = TextToPaste
            RTB_Lines = Me.Lines
            PaintEventEnabled = False
            EventsAreEnabled = False
            '
            ' ------------------------------------------------------------ 
            InitButtonsIdentifiers()
            ReinitUserVariables()
            ReinitLabelNames()


            ' TODO - COLOR ALL LINES - PASTE ORIGINAL - REMOVE ???
            ' ------------------------------------------------------------ color lines starting with variable
            ColorAllLines_StartingWithVariable()
            Cursor.Current = Cursors.WaitCursor
            RTB_Lines = Me.Lines
            ' ------------------------------------------------------------ color specific keywords
            Me.SelectLines(StartLine, EndLine)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
            Cursor.Current = Cursors.WaitCursor
            RTB_Lines = Me.Lines
            '
            ' ------------------------------------------------------------ ensure to remove buttons identifiers errors
            ' ------------------------------------------------------------ required to remove errors when pasting many buttons
            InitButtonsIdentifiers()
            PaintEventEnabled = False
            EventsAreEnabled = False
            Me.Enabled = False
            Dim StartLineToRecolor As Int32 = GetLineFromCharIndex(SelectionStart)
            Dim EndLineToRecolor As Int32 = GetLineFromCharIndex(SelectionStart + TextToPaste.Length)
            ColorLines(StartLineToRecolor, EndLineToRecolor)
            Cursor.Current = Cursors.WaitCursor

            ' TODO - COLOR ALL LINES - PASTE V2 - ADD ???
            'ColorAllLines()

            ' ------------------------------------------------------------
            PaintEventEnabled = True
            RTB_Lines = Me.Lines
            '
            ' ------------------------------------------------------------
            ShowSuggestions()
            '
            ' ------------------------------------------------------------ set the final cursor position
            Me.SelectionLength = 0
            Me.SelectionStart = StartSelectionPoint + TextToPaste.Length
            '
            ' ------------------------------------------------------------ ensure the text indented
            AutoIndent()
            RTB_Lines = Me.Lines
            '
            ' ------------------------------------------------------------ 
            PaintEventEnabled = oldPaintEvent
            EventsAreEnabled = oldEventsAreEnabled
            Me.Enabled = True
            ' ------------------------------------------------------------
            SetTextModifiedFlags_And_ReinitVariables()
            Me.Focus()
            Cursor.Current = Cursors.Default
            ' ------------------------------------------------------------ restore the original backcolor
            BackColor = oldcolor
            Refresh()
        End If
        ' -----------------------------------------------------------------
        'Form1.Text = sw.ElapsedMilliseconds.ToString
    End Sub

    Friend Sub ExecDelete()
        If Me.SelectionLength > 0 Then
            ' ------------------------------------------------------------ PUSH UNDO
            PushUndoState()
            ' ------------------------------------------------------------
            Dim oldSelectionStart As Int32 = SelectionStart
            Dim startline As Int32 = GetLineFromCharIndex(SelectionStart)
            ' ------------------------------------------------------------ 
            TestKeywords_PrepareKeywordsToRecolor()
            ' ------------------------------------------------------------ 
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            Dim oldEnabled As Boolean = Me.Enabled
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------ 
            PaintEventEnabled = True
            Me.Enabled = True
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
            PaintEventEnabled = False
            Me.Enabled = False
            ' ------------------------------------------------------------
            Dim endline As Int32 = GetLineFromCharIndex(SelectionStart + SelectionLength)
            ' ------------------------------------------------------------ color lines
            ReinitUserVariables()
            ReinitLabelNames()
            ColorLines(startline, endline)
            ColorAllLines_StartingWithVariable()
            ' ------------------------------------------------------------
            ShowSuggestions()
            ' ------------------------------------------------------------
            Me.SelectLines(startline, endline)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
            ' ------------------------------------------------------------
            Me.SelectionStart = oldSelectionStart
            Me.SelectionLength = 0
            ' ------------------------------------------------------------
            PaintEventEnabled = oldPaintEvent
            Me.Enabled = oldEnabled
            ' ------------------------------------------------------------
            SetTextModifiedFlags_And_ReinitVariables()
            Me.Focus()
            ' ------------------------------------------------------------
            ScrollToCaret_if_DeletingLastLines()
        End If
    End Sub

    ' This enable to scroll immediately when deleting many lines at the end of the file
    ' (test with the file ERRORS_EDIT-1.txt)
    Private Sub ScrollToCaret_if_DeletingLastLines()
        If Me.GetPositionFromCharIndex(Me.Text.Length).Y < Me.ClientSize.Height - 32 Then
            Me.ScrollToCaret()
        End If
    End Sub

    Friend Sub ExecReplaceTabsWithSpaces()
        If RTB_Lines.Length > 0 Then
            Cursor.Current = Cursors.WaitCursor
            Dim sd As ScrollData
            Form1.RTB.SaveSelectionAndScroll(sd)
            SetTextModifiedFlags_And_ReinitVariables()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            If Me.SelectedText.Length = 0 Then
                Me.Text = Me.Text.Replace(vbTab, "    ")
                SelectionLength = 0
                ColorAllLines()
                Form1.RTB.RestoreScroll(sd)
            Else
                Me.SelectedText = Me.SelectedText.Replace(vbTab, "    ")
                SelectLines(FirstSelectedLine, LastSelectedLine)
                ColorLines(FirstSelectedLine, LastSelectedLine)
                Form1.RTB.RestoreSelectedLinesAndScroll(sd)
            End If
            ' ----------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            SetTextModifiedFlags_And_ReinitVariables()
            Cursor.Current = Cursors.Default
        End If
        Me.Focus()
    End Sub

    Friend Sub ExecReplaceSpacesWithTabs()
        If RTB_Lines.Length > 0 Then
            Cursor.Current = Cursors.WaitCursor
            Dim sd As ScrollData
            Form1.RTB.SaveSelectionAndScroll(sd)
            SetTextModifiedFlags_And_ReinitVariables()
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            If Me.SelectedText.Length = 0 Then
                Me.Text = Me.Text.Replace("    ", vbTab)
                SelectionLength = 0
                ColorAllLines()
                Form1.RTB.RestoreScroll(sd)
            Else
                Me.SelectedText = Me.SelectedText.Replace("    ", vbTab)
                SelectLines(FirstSelectedLine, LastSelectedLine)
                ColorLines(FirstSelectedLine, LastSelectedLine)
                Form1.RTB.RestoreSelectedLinesAndScroll(sd)
            End If
            ' ----------------------------------------------------------------
            Me.Enabled = True
            PaintEventEnabled = True
            SetTextModifiedFlags_And_ReinitVariables()
            Cursor.Current = Cursors.Default
        End If
        Me.Focus()
    End Sub


    ' ===============================================================================================
    '  BREAKPOINTS and WATCHES
    ' ===============================================================================================
    Friend Sub ToggleBreakPoint(ByVal mousepos As Point)
        If RTB_Lines.Length <= 0 Then Return
        '
        Dim charindex As Int32 = Me.GetCharIndexFromPosition(mousepos)
        Dim nline As Int32 = GetNumLineFromPoint(mousepos)
        Dim lineText As String = RTB_Lines(nline).TrimEnd
        lineText = ReplaceMultipleSpacesAndTabAndTrim(lineText)
        '
        If lineText.EndsWith("'Breakpoint") Then
            DeleteBreakpointFromLine(nline)
        Else
            Dim commentpos As Int32 = lineText.IndexOf("'")
            If commentpos <> 0 Then
                AddBreakPoint(nline)
                ' ------------------------------------------------- open the Debug Form
                If Form_Debug.WindowState = FormWindowState.Minimized Then
                    Form_Debug.WindowState = FormWindowState.Normal
                End If
                Form_Debug.Show()
            Else
                'Console.Beep(800, 50)
                PlayBeepAsync(800, 50, "")
                Form1.StatusLabel1.Text = "Breakpoints are not valid on comment lines."
                Form1.StatusLabel1.BackColor = Form1.RTB.Color_WarningBack
                Form1.StatusLabel1.Invalidate()
            End If
        End If
    End Sub

    Friend Sub AddBreakPoint(ByVal nline As Int32)
        If RTB_Lines.Length <= 0 Then Return
        If RTB_Lines(nline).TrimEnd.EndsWith("'Breakpoint") Then Return
        ' ---------------------------------------------------------------------
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length
        Dim len As Int32 = GetLineLengthAsEquivalentSpaces(nline)
        Dim ntabs As Int32 = (48 - len + 3) \ 4
        ' ---------------------------------------------------------------------
        If ntabs < 0 Then
            endpos -= (-ntabs + 1) * 4
            ntabs = 0
        End If
        ' ---------------------------------------------------------------------
        If endpos - Me.GetFirstCharIndexFromLine(nline) < RTB_Lines(nline).Length Then
            endpos = Me.GetFirstCharIndexFromLine(nline) + RTB_Lines(nline).Length
        End If
        ' ---------------------------------------------------------------------
        Me.Select(endpos, 0)
        Me.SelectedText = StrDup(ntabs, vbTab) + " 'Breakpoint"
        ' ---------------------------------------------------------------------
        RTB_Lines = Me.Lines
        ColorSingleLine(nline)
        RTB_Lines = Me.Lines
        ' ---------------------------------------------------- remove red border from the following lines
        ColorRangeOfLines(nline + 1, nline + 3)
        Hilighter_RedBorderToSelectedLine()
    End Sub

    Friend Sub DeleteAllBreakPoints()
        Dim i As Int32 = RTB_Lines.Length - 1
        While i >= 0
            Dim lineText As String = RTB_Lines(i).Trim
            If lineText.EndsWith("'Breakpoint") Then
                DeleteBreakpointFromLine(i)
            Else
                i = i - 1
            End If
        End While
    End Sub

    Private Sub DeleteBreakpointFromLine(ByVal i As Int32)
        If RTB_Lines.Length <= 0 Then Return
        ' --------------------------------------------- select the word "'Breakpoint"
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).IndexOf("'Breakpoint")
        Me.Select(startpos, endpos - startpos)
        ' --------------------------------------------- remove breakpoint and TrimRightSpacesFromLine
        Me.ReadOnly = False
        Me.SelectedText = ""
        RTB_Lines = Me.Lines
        TrimRightSpacesFromLine_TrimAlsoEmptyLines(i)
        If Running Then Me.ReadOnly = True
        RTB_Lines = Me.Lines
        Me.Select(Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length, 0)
    End Sub

    Private Sub TrimRightSpacesFromLine_TrimAlsoEmptyLines(ByVal i As Int32)
        If i < 0 Or i >= RTB_Lines.Length Then Return
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        RTB_Lines(i) = RTB_Lines(i).TrimEnd
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        If endpos > startpos Then
            Me.Select(startpos, endpos - startpos)
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
        End If
    End Sub

    Private Sub TrimRightSpacesFromLine(ByVal i As Int32)
        If i < 0 Or i >= RTB_Lines.Length Then Return
        Dim endpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        If RTB_Lines(i).Trim = "" Then
            While RTB_Lines(i).EndsWith(" ")
                RTB_Lines(i) = RTB_Lines(i).Remove(RTB_Lines(i).Length - 1, 1)
            End While
        Else
            RTB_Lines(i) = RTB_Lines(i).TrimEnd
        End If
        Dim startpos As Int32 = Me.GetFirstCharIndexFromLine(i) + RTB_Lines(i).Length
        If endpos > startpos Then
            Me.Select(startpos, endpos - startpos)
            Me.SelectedText = ""
            RTB_Lines = Me.Lines
        End If
    End Sub

    ' ===============================================================================================
    '  GoToNextOccurrence / GoToDeclaration
    ' ===============================================================================================
    Friend Sub GoToNextOccurrence()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTB_Lines.Length - 1 Then Beep() : testLine = 0
            If testLine = startLine Then Exit For
            l = RTB_Lines(testLine).ToLower.Trim.Replace("""", "")
            'l = l.Replace("'", "")
            l = l.TrimEnd
            If l.Contains(s) Then
                If Form_FindReplace.IsWholeWord(l, l.IndexOf(s), s) Then
                    Me.ScrollToCenterLine(testLine)
                    Dim n As Int32 = RTB_Lines(testLine).ToLower.IndexOf(s)
                    Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                    History_AddEntry()
                    Me.Focus()
                    Exit For
                End If
            End If
        Next
    End Sub

    Friend Sub GoToDeclaration()
        ' ---------------------------------------------------------- get the word to be searched
        Dim s As String = Me.SelectedText
        If s = "" Then
            s = GetCompleteWord_Short(Me.Text, Me.SelectionStart)
        End If
        s = s.ToLower.Replace("""", "").Replace("'", "")
        If s = "" Then Return
        ' ---------------------------------------------------------- search all the lines starting from cursor
        Dim l As String
        Dim startLine As Int32 = Me.GetLineFromCharIndex(Me.SelectionStart)
        For testLine As Int32 = startLine + 1 To 99999999
            If testLine >= RTB_Lines.Length - 1 Then Beep() : testLine = 0
            If testLine = startLine Then Exit For
            l = RTB_Lines(testLine).ToLower.Trim.Replace("""", "")
            'l = l.Replace("'", "")
            l = l.TrimEnd
            RemoveComments(l)
            If l.Contains(s) Then
                If Form_FindReplace.IsWholeWord(l, l.IndexOf(s), s) Then
                    If l.StartsWith("label") Or l.StartsWith("variable") Or l.StartsWith("button") Or l.StartsWith("key") Then
                        Me.ScrollToCenterLine(testLine)
                        Dim n As Int32 = RTB_Lines(testLine).ToLower.IndexOf(s)
                        Me.Select(GetFirstCharIndexFromLine(testLine) + n, s.Length)
                        History_AddEntry()
                        Me.Focus()
                        Exit For
                    End If
                End If
            End If
        Next
    End Sub


    ' -------------------------------------------------------------------------------------------------------
    '  If some unicode character (for example the "Bidirectional text control") character is assigned to RTB
    '  then RTB stops to work correctly and a colored character is visible at the end of each line.
    ' -------------------------------------------------------------------------------------------------------
    '
    ' Hebrew and Arabic chars
    ' 0590 to 097F
    '
    ' Chars for "Bidirectional text control"
    ' U+061C, U+200E, U+200F, U+202A, U+202B, U+202C, U+202D, U+202E, U+2066, U+2067, U+2068, U+2069
    '
    ' Note that in the HEX-Editors the bytes are exchanged so that "20 2C" becomes "2C 20"
    '
    Private Const InvalidUnicodes As String = "[\u0590-\u097F]|" + _
                                              "\u200E|\u200F|\u202A|\u202B|\u202C|\u202D|\u202E"
    ' --------------------------------------------------------------------------------------------------------


    Friend Sub LoadProgramFile(ByVal fname As String)
        '
        ' ----------------------------------------------- load as unicode (or any other format specified in the txt)
        Dim s As String() = IO.File.ReadAllLines(fname)
        '
        ' ----------------------------------------------- trim end chars and remove unwanted unicode chars
        For i As Int32 = 0 To s.Length - 1
            s(i) = s(i).TrimEnd
            ' -------------------------------------------------------------------------------
            '  Remove the unwanted unicode chars 
            ' -------------------------------------------------------------------------------
            s(i) = System.Text.RegularExpressions.Regex.Replace(s(i), InvalidUnicodes, "")
            ' TEST (use the program "Examples\_Software_Tests\FINAL CHAR RED\Test1.txt")
            's(i) = s(i).Replace("12345", "1234" + StrDup(10, ChrW(&H750)) + "5")
            ' -------------------------------------------------------------------------------
        Next
        '
        ' ----------------------------------------------- assign lines to RTB
        Me.Text = ""
        Me.Lines = s
        Me.AppendText(vbCrLf)
        RTB_Lines = Me.Lines
    End Sub

    Friend Sub ExtractWatches()
        RTB_Lines = Me.Lines
        Dim startline As Int32 = 0
        ' ------------------------------------------- a maximum of 100 watches to increase the loading speed
        If RTB_Lines.Length > 101 Then
            startline = RTB_Lines.Length - 100
        End If
        ' -------------------------------------------
        Form_Debug.DeleteAllWatches()
        Form_Debug.txt_Exec.Text = ""
        Dim i As Int32 = startline
        While i < RTB_Lines.Length - 1
            Dim lineText As String = RTB_Lines(i).Trim
            If lineText.StartsWith("'WATCH") Then
                Form_Debug.AddWatchLine(Mid(lineText, 7).Trim)
                DeleteLine(i)
                RTB_Lines = Me.Lines
                i -= 1
            End If
            If lineText.StartsWith("'EXEC") Then
                Form_Debug.txt_Exec.Text = Mid(lineText, 6).Trim
                DeleteLine(i)
                RTB_Lines = Me.Lines
                i -= 1
            End If
            i += 1
        End While
        Form_Debug.ResizeListViewColumns()
    End Sub

    Friend Sub DeleteLine(ByVal line As Int32)
        Dim startpos As Int32 = GetFirstCharIndexFromLine(line)
        Dim count As Int32 = Lines(line).Length + 1
        If startpos + count > Me.Text.Length Then Return
        Me.Text = Me.Text.Remove(startpos, count)
    End Sub

    Friend Sub DeleteFinalEmptyLines()
        RTB_Lines = Me.Lines
        For i As Int32 = RTB_Lines.Length - 1 To 0 Step -1
            If RTB_Lines(i).Trim = "" Then
                ReDim Preserve RTB_Lines(RTB_Lines.Length - 2)
            Else
                Exit For
            End If
        Next
        Me.Lines = RTB_Lines
        Me.AppendText(vbCrLf)
    End Sub

    ' ===============================================================================================
    '  INDENT-UNINDENT
    ' ===============================================================================================
    Friend Sub IndentSelectedLines(ByVal increment As Int32)
        If RTB_Lines.Length > 0 Then
            ' ---------------------------------------------------------- 
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            Dim oldEnabled As Boolean = Me.Enabled
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim ntabs As Int32 = CountInitialTabs_NotAllowingSpaces(lineText) + increment
                If ntabs >= 0 Then
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    Me.Select(start, CountInitialChars(lineText))
                    Me.SelectedText = StrDup(ntabs, vbTab)
                Else
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    Me.Select(start, lineText.Length)
                    Me.SelectedText = lineText.TrimStart
                End If
            Next
            ' ----------------------------------------------------------------
            SelectLines(FirstSelectedLine, LastSelectedLine)
            ' ----------------------------------------------------------------
            PaintEventEnabled = oldPaintEvent
            Me.Enabled = oldEnabled
            SetTextModifiedFlags_And_ReinitVariables()
        End If
        Me.Focus()
    End Sub

    ' ===============================================================================================
    '  COMMENT UN-COMMENT
    ' ===============================================================================================
    Friend Sub CommentSelectedLines(ByVal comment As Boolean)
        ' ------------------------------------------------------------ 
        TestKeywords_PrepareKeywordsToRecolor()
        ' ------------------------------------------------------------ 
        RTB_Lines = Me.Lines
        If RTB_Lines.Length > 0 Then
            ' ----------------------------------------------------------------
            Dim oldPaintEvent As Boolean = PaintEventEnabled
            Dim oldEnabled As Boolean = Me.Enabled
            PaintEventEnabled = False
            Me.Enabled = False
            ' ----------------------------------------------------------------
            Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
            Dim LastSelectedLine As Int32 = GetLastSelectedLine()
            ' ----------------------------------------------------------------
            Dim CommentColumn As Int32 = 999999
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch < CommentColumn Then CommentColumn = firstch
            Next
            ' ----------------------------------------------------------------
            For i As Int32 = FirstSelectedLine To LastSelectedLine
                Dim lineText As String = RTB_Lines(i)
                Dim firstch As Int32 = CountInitialChars(lineText)
                If firstch >= 0 Then
                    Dim start As Int32 = Me.GetFirstCharIndexFromLine(i)
                    If comment Then
                        Me.Select(start + CommentColumn, 0)
                        Me.SelectedText = "'"
                    Else
                        Me.Select(start + firstch, 1)
                        If Me.SelectedText = "'" Then
                            Me.SelectedText = ""
                        End If
                    End If
                End If
            Next
            ' ----------------------------------------------------------------
            ReinitUserVariables()
            ReinitLabelNames()
            ColorLines(FirstSelectedLine, LastSelectedLine)
            ColorAllLines_StartingWithVariable()
            ' ----------------------------------------------------------------
            PaintEventEnabled = oldPaintEvent
            Me.Enabled = oldEnabled
            ' ----------------------------------------------------------------
            SetTextModifiedFlags_And_ReinitVariables()
            ' ----------------------------------------------------------------
            Me.SelectLines(FirstSelectedLine, LastSelectedLine)
            TestKeywords_PrepareKeywordsToRecolor()
            TestKeywords_RecolorLines()
        End If
        Me.Focus()
    End Sub

    Private Function CountInitialChars(ByVal s As String) As Int32
        Dim CharCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : CharCount += 1
                Case " " : CharCount += 1
                Case Else : Exit For
            End Select
        Next
        Return CharCount
    End Function

    Private Function CountInitialTabs(ByVal s As String) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    Private Function CountInitialTabs_NotAllowingSpaces(ByVal s As String) As Int32
        Dim TabCount As Int32 = 0
        For i As Int32 = 0 To s.Length - 1
            Select Case s.Substring(i, 1)
                Case vbTab : TabCount += 1
                Case " " : Return -9999           ' Spaces are not allowed
                Case Else : Exit For
            End Select
        Next
        Return TabCount
    End Function

    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '  AUTO COMPLETION - AUTOCOMPLETION
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Private AutoCompletion_FinalCursorPosition As Int32

    Private Function TestAutoCompletion() As Boolean
        ' --------------------------------------------------------
        AutoCompletion_FinalCursorPosition = 0
        Dim FirstSelectedLine As Int32 = GetFirstSelectedLine()
        If FirstSelectedLine < 0 Then Return False
        If FirstSelectedLine >= RTB_Lines.Length Then Return False
        ' ------------------------------------------- do not complete if cursor is before the end of line
        If CursorIsBeforeTheEndOfLine() Then
            Return False
        End If
        ' ------------------------------------------- do not complete if the following line is used
        If FirstSelectedLine + 1 < RTB_Lines.Length AndAlso _
           RTB_Lines(FirstSelectedLine + 1).Trim <> "" Then
            Return False
        End If
        ' --------------------------------------------------------
        Dim s As String = RTB_Lines(FirstSelectedLine).Trim.ToLower
        Dim status As String = Form1.StatusLabel1.Text.ToLower
        If s.StartsWith("if") Then
            If status.StartsWith("warning") And status.EndsWith("endif") Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + vbLf + "EndIf"
                RTB_Lines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
        If s.StartsWith("for") Then
            If status.StartsWith("please") Or (status.StartsWith("error") And status.EndsWith("next")) Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + vbLf + "Next"
                RTB_Lines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
        If s.StartsWith("select") Then
            If status.StartsWith("warning") And status.EndsWith("endselect") Then
                Me.SelectionLength = 0
                Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
                Me.SelectedText = vbLf + "Case " + vbLf + "EndSelect"
                RTB_Lines = Me.Lines()
                AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
                TestAutoCompletion = True
            End If
        End If
        If s.StartsWith("label") Then
            Me.SelectionLength = 0
            Me.SelectionStart = Me.GetLastCharIndexFromLine(FirstSelectedLine)
            If s.EndsWith("eventstop") Then
                Me.SelectedText = vbLf + vbLf + "End"
            Else
                Me.SelectedText = vbLf + vbLf + "Return"
            End If
            RTB_Lines = Me.Lines()
            AutoCompletion_FinalCursorPosition = Me.GetLastCharIndexFromLine(FirstSelectedLine + 1)
            TestAutoCompletion = True
        End If
    End Function

    Private Function CursorIsBeforeTheEndOfLine() As Boolean
        Return Me.SelectionStart < Me.GetLastCharIndexFromLine(GetFirstSelectedLine)
    End Function

    Private Sub SetAutoCompletionCursorPosition()
        If AutoCompletion_FinalCursorPosition > 0 Then
            Me.SelectionStart = AutoCompletion_FinalCursorPosition
            AutoCompletion_FinalCursorPosition = 0
        End If
    End Sub

    Friend Sub SetTextModifiedFlags_And_ReinitVariables()
        ' ----------------------------------------------------------------
        If Not PaintEventEnabled Then Exit Sub
        ' ----------------------------------------------------------------
        TextModified = True
        TextMustBeSaved = True
        'TextMustBeIndented = True  ' --- REMOVED
        ' ---------------------------------------------------------------- reinit user variables
        ReinitUserVariables()
        ReinitLabelNames()
    End Sub


    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '   UNDO and HISTORY
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    ' ===============================================================================================
    '  PRIVATE UNDO FUNCTIONS
    ' ===============================================================================================
    Private Structure TextState
        Dim RTF_Text As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private UndoStack As New Stack(Of TextState)
    Private RedoStack As New Stack(Of TextState)
    Friend Sub Undo_Clear()
        UndoStack.Clear()
        RedoStack.Clear()
        TextModified = False
    End Sub
    Private Function GetTextState() As TextState
        GetTextState.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        GetTextState.RTF_Text = Me.Rtf
        GetTextState.SelectionStart = Me.SelectionStart
        GetTextState.SelectionLength = Me.SelectionLength
    End Function

    Private Sub SetTextState(ByVal ts As TextState)
        Me.Enabled = False
        PaintEventEnabled = False
        '
        Me.Rtf = ts.RTF_Text
        Me.Select(ts.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.SelectionStart = ts.SelectionStart
        Me.SelectionLength = ts.SelectionLength
        '
        Form1.ApplyEditorProps(True)
        '
        PaintEventEnabled = True
        Me.Enabled = True

        Me.Focus()
    End Sub
    Friend Sub PushUndoState()
        UndoStack.Push(GetTextState)
        RedoStack.Clear()
    End Sub

    Friend Function Editor_CanUndo() As Boolean
        Return UndoStack.Count > 0
    End Function
    Friend Function Editor_CanRedo() As Boolean
        Return RedoStack.Count > 0
    End Function

    ' ===============================================================================================
    '  UNDO and REDO commands
    ' ===============================================================================================
    Friend Sub ExecUndo()
        If Editor_CanUndo() Then
            RedoStack.Push(GetTextState)
            SetTextState(UndoStack.Pop())
        End If
    End Sub
    Friend Sub ExecRedo()
        If Editor_CanRedo() Then
            UndoStack.Push(GetTextState)
            SetTextState(RedoStack.Pop())
        End If
    End Sub


    ' ===============================================================================================
    '   PRIVATE HISTORY FUNCTIONS
    ' ===============================================================================================
    Private Structure HistoryEntry
        Dim ActionName As String
        Dim FirstVisibleChar As Int32
        Dim SelectionStart As Int32
        Dim SelectionLength As Int32
    End Structure
    Private HistoryBuffer As List(Of HistoryEntry) = New List(Of HistoryEntry)
    Private History_Index As Int32
    Friend Sub History_Clear()
        HistoryBuffer.Clear()
        History_Index = 0
        History_AddEntry()
    End Sub
    Friend Sub History_AddEntry()
        ' ----------------------------------------------------------------
        Dim newentry As HistoryEntry
        newentry.FirstVisibleChar = Me.GetCharIndexFromPosition(New Point(1, 1))
        newentry.SelectionStart = Me.SelectionStart
        newentry.SelectionLength = Me.SelectionLength
        newentry.ActionName = ""
        ' ----------------------------------------------------------------
        HistoryBuffer.Add(newentry)
        History_Index += 1
    End Sub

    ' ===============================================================================================
    '  HISTORY commands
    ' ===============================================================================================
    Friend Sub History_Back()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- back
        History_Index -= 1
        If History_Index < 1 Then History_Index = 1
        Dim hist As HistoryEntry
        hist = HistoryBuffer(History_Index - 1)
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ---------------------------------------------------------------- 
        ShowLinesCount()
        ' ---------------------------------------------------------------- 
        Refresh()
    End Sub
    Friend Sub History_Forward()
        ' ---------------------------------------------------------------- 
        PaintEventEnabled = False
        Me.Enabled = False
        ' ---------------------------------------------------------------- forward
        If History_Index > HistoryBuffer.Count - 1 Then History_Index = HistoryBuffer.Count - 1
        Dim hist As HistoryEntry = HistoryBuffer(History_Index)
        History_Index += 1
        ' ---------------------------------------------------------------- apply props must be before the scroll
        Form1.ApplyEditorProps(True)
        ' ---------------------------------------------------------------- update scroll and selection
        Me.Select(hist.FirstVisibleChar, 0)
        Me.ScrollToCaret()
        Me.Select(hist.SelectionStart, hist.SelectionLength)
        ' ----------------------------------------------------------------
        Me.ShowSuggestions()
        ' ----------------------------------------------------------------
        Me.Enabled = True
        PaintEventEnabled = True
        ' ----------------------------------------------------------------
        ShowLinesCount()
        ' ---------------------------------------------------------------- 
        Refresh()
    End Sub
    Friend Function Editor_CanHistoryBack() As Boolean
        Return History_Index >= 2
    End Function
    Friend Function Editor_CanHistoryForward() As Boolean
        Return History_Index <= HistoryBuffer.Count - 1
    End Function

    ' ===============================================================================================================
    '  FIND - REPLACE and Options
    ' ===============================================================================================================
    Friend Sub Editor_Find()
        If SelectedText = "" Then SelectCompleteWord_Short()
        With Form_FindReplace
            .OwnerCodeBox = Me
            .InitFindText()
            .TopMost = True
            .DialogType = Form_FindReplace.DialogTypeValue.Find
            .Show()
            .txtFind.Focus()
        End With
    End Sub
    Friend Sub Editor_FindReplace()
        If SelectedText = "" Then SelectCompleteWord_Short()
        With Form_FindReplace
            .OwnerCodeBox = Me
            .InitFindText()
            .TopMost = True
            .DialogType = Form_FindReplace.DialogTypeValue.Replace
            .Show()
            .txtFind.Focus()
        End With
    End Sub



    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '   EVENTS
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    ' ===============================================================================================
    '  DISABLE MOUSE SELECTION WHEN NOT EDITING
    ' ===============================================================================================
    'Private Sub RTB_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.MouseLeave
    '    If Not Running Then Enabled = True
    'End Sub
    'Private Sub RTB_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
    '    If Not Running Then Enabled = False
    'End Sub

    ' ===============================================================================================
    '  ENABLE DISABLE PAINT and DISABLE ZOOM with CTRL WHEEL
    ' ===============================================================================================
    Private Const MK_CONTROL As Int32 = &H8
    Private Const WM_MOUSEWHEEL As Int32 = &H20A
    Private Const WM_PAINT As Int32 = &HF
    Friend PaintEventEnabled As Boolean = True
    <DebuggerStepThrough()> _
    Protected Overrides Sub WndProc(ByRef m As Message)
        If m.Msg = WM_PAINT Then
            If PaintEventEnabled Then
                MyBase.WndProc(m)
                Hilighter_RedBorderToSelectedLine()
            Else
                m.Result = IntPtr.Zero
            End If
        ElseIf m.Msg = WM_MOUSEWHEEL Then
            If (m.WParam.ToInt32 And MK_CONTROL) = MK_CONTROL Then
                SetZoomByMouseWheel(m.WParam.ToInt32)
                m.WParam = IntPtr.Zero
                m.Result = IntPtr.Zero
            Else
                MyBase.WndProc(m)
            End If
        Else
            MyBase.WndProc(m)
        End If
    End Sub

    Private Sub SetZoomByMouseWheel(ByVal Delta As Int32)
        Form1.tkbar_Zoom.Value += If(Delta < 0, -1, 1)
    End Sub

    ' ===============================================================================================
    '   KEY DOWN / KEY UP
    ' ===============================================================================================
    Private LastKeyPressed As Keys
    Private KeyPressedFlag As Boolean = False

    Private Sub SyntaxRichTextBox_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown

        'Return
        ''
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        ''
        If Not Enabled Then Return
        If Not PaintEventEnabled Then Return
        '
        ' ---------------------------------------------------------------------- NO KEY DOWN WHILE RUNNING
        If Running Then Return
        '
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Cancel, Keys.Delete, Keys.Back
                ' -------------------------------------------------------------- autoindent flag
                TextMustBeIndented = True
            Case Keys.Tab
                If SelectedText.Length > 0 Then
                    IndentSelectedLines(If(e.Shift, -1, 1))
                    e.Handled = True
                    e.SuppressKeyPress = True
                End If
            Case Keys.Insert
                If e.Modifiers <> Keys.None Then
                    e.Handled = True
                Else
                    OverwriteMode_Toggle(Me)
                End If
        End Select
        ' -------------------------------------------------- 
        Select Case e.KeyCode
            Case Keys.Enter, Keys.Up, Keys.Down
                TrimRightSpacesFromLine(Me.GetLineFromCharIndex(Me.SelectionStart))
        End Select
        ' -------------------------------------------------- 
        LastKeyPressed = e.KeyCode
        RTB_Lines = Me.Lines()
        ' -------------------------------------------------- auto completion
        If e.KeyCode = Keys.Enter Then
            If TestAutoCompletion() Then
                e.Handled = True
                ColorVisibleLines()
            End If
        End If
        ' -------------------------------------------------- test keywords and eventually recolor
        TestKeywords_PrepareKeywordsToRecolor()
        ' -------------------------------------------------- suppress RTB sounds
        RichTextBox_SuppressSounds(sender, e)
        ' -------------------------------------------------- remove the red lines indicating the cursor
        If Not KeyPressedFlag Then
            KeyPressedFlag = True
            If e.KeyCode <> Keys.ControlKey Then
                Refresh()
            End If
        End If
        '
        'Form1.Text = sw.Elapsed.TotalMilliseconds.ToString

        If Not Form1.PrintTextBox.Focused Then
            If My.Computer.Keyboard.CtrlKeyDown Then
                ' ------------------------------------------------------- CUT / COPY / PASTE / UNDO / REDO
                RTB_Lines = Me.Lines
                e.SuppressKeyPress = True
                e.Handled = True
                Select Case e.KeyCode
                    Case Keys.F : PushUndoState() : Editor_Find()
                    Case Keys.R : PushUndoState() : Editor_FindReplace()
                    Case Keys.X : PushUndoState() : Me.ExecCut()
                    Case Keys.C : Me.ExecCopy()
                    Case Keys.V : PushUndoState() : Me.ExecPaste()
                    Case Keys.Z : Me.ExecUndo()
                    Case Keys.Y : Me.ExecRedo()
                    Case Keys.Space : CompleteKeyword()
                    Case Else
                        e.Handled = False
                        e.SuppressKeyPress = False
                End Select
                'Me.Focus()
            Else
                ' ------------------------------------------------------- UNDO PUSH
                Select Case e.KeyCode
                    Case Keys.Up, Keys.Down, Keys.Left, Keys.Right
                    Case Else
                        PushUndoState()
                End Select
            End If
        End If

    End Sub

    Private Sub SyntaxRichTextBox_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp

        'Return
        '
        'Dim sw As Stopwatch = New Stopwatch
        'sw.Start()
        '
        ' -------------------------------------------------- NO KEYUP WHILE RUNNING
        If Running Then Return
        If Not Enabled Then Return
        If Not PaintEventEnabled Then Return
        '
        ' --------------------------------------------------
        KeyPressedFlag = False
        '
        ' -------------------------------------------------- THIS ENABLES MULTIPAGE SELECTION WITH SHIFT
        If e.KeyCode = Keys.ShiftKey Then Return
        '
        ' -------------------------------------------------- THIS ENABLES MULTIPAGE CTRL-C and CTRL-A
        '                                                                     and also SHIFT + ARROWS
        Static LastKeyWasControlOrShift_With_C_A_Arrows As Boolean = False
        If LastKeyWasControlOrShift_With_C_A_Arrows Then
            LastKeyWasControlOrShift_With_C_A_Arrows = False
            If e.KeyCode = Keys.ControlKey Or e.KeyCode = Keys.ShiftKey Then
                Return
            End If
        End If
        If (e.KeyCode = Keys.C Or e.KeyCode = Keys.A Or _
            e.KeyCode = Keys.Up Or e.KeyCode = Keys.Down Or _
            e.KeyCode = Keys.PageUp Or e.KeyCode = Keys.PageDown Or _
            e.KeyCode = Keys.Left Or e.KeyCode = Keys.Right) AndAlso (e.Control Or e.Shift) Then
            LastKeyWasControlOrShift_With_C_A_Arrows = True
            Return
        End If
        '
        ' -------------------------------------------------------------- init RTB_Lines
        RTB_Lines = Me.Lines()
        '
        ' -------------------------------------------------------------- finish the autocompletion
        SetAutoCompletionCursorPosition()
        ' -------------------------------------------------------------- TEST KEYWORDS AND EVENTUALLY RECOLOR LINES
        If TextChangedFlag Then
            TextChangedFlag = False
            ' ---------------------------------------------------------- init SelectedLine
            Dim SelectedLine As Int32 = GetFirstSelectedLine()
            ' ---------------------------------------------------------- prepare the LineTrimmedLowcase
            Dim LineTrimmedLowcase As String
            If RTB_Lines.Length = 0 Then
                ' ------------------------------------------------------ avoid errors with lines array empty
                LineTrimmedLowcase = ""
            Else
                LineTrimmedLowcase = RTB_Lines(SelectedLine).TrimStart.ToLower
            End If
            ' ---------------------------------------------------------- REINIT USER VARIABLES
            If RTB_Lines.Length > 0 Then
                ReinitUserVariables()
            End If
            ' ---------------------------------------------------------- REINIT LABEL NAMES
            If RTB_Lines.Length > 0 Then
                ReinitLabelNames()
            End If
            ' ---------------------------------------------------------- Reinit Buttons Identifiers (SLOW)
            If LineTrimmedLowcase.Contains("button") Then
                InitButtonsIdentifiers()
            End If
            ' ---------------------------------------------------------- recolor lines after one second
            TestKeywords_PrepareKeywordsToRecolor()
            SetUpdateTimer()
            ' ---------------------------------------------------------- recolor this line and the precedent line (if splitted with ENTER)
            If e.KeyCode <> Keys.Home And e.KeyCode <> Keys.End Then
                ColorRangeOfLines(SelectedLine - 1, SelectedLine)
            End If
            ' ---------------------------------------------------------- recolor following (text and color) button lines  
            If LineTrimmedLowcase.Contains("button") Then
                ColorRangeOfLines(SelectedLine + 1, SelectedLine + 3)
            End If
        End If
        ' -------------------------------------------------------------- update the status bar SUGGESTIONS
        ShowSuggestions()
        ' -------------------------------------------------------------- show the Overwrite Caret
        OverwriteMode_ShowCaret(Me)
        ' -------------------------------------------------------------- 
        'If e.KeyCode <> Keys.ControlKey Then
        ScrollToCaret_if_DeletingLastLines()
        'End If
        '
        'Form1.Text = sw.ElapsedMilliseconds.ToString
    End Sub


    ' ===============================================================================================
    '   ON TEXT CHANGED
    ' ===============================================================================================
    Private TextChangedFlag As Boolean = False
    Private LastEditedLine As Int32 = -1
    Protected Overrides Sub OnTextChanged(ByVal e As EventArgs)
        If Not PaintEventEnabled Then Return
        ' ------------------------------------------------------- Set FLAGS
        TextChangedFlag = True
        TextModified = True
        TextMustBeSaved = True
        TextMustBeIndented = True
    End Sub


    ' ===============================================================================================
    '   UPDATE TIMER
    ' ===============================================================================================
    Private WithEvents UpdateTimer As Windows.Forms.Timer = New Windows.Forms.Timer
    Private Sub SetUpdateTimer()
        UpdateTimer.Interval = 1500
        UpdateTimer.Stop()
        UpdateTimer.Start()
    End Sub
    Private Sub UpdateTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles UpdateTimer.Tick
        UpdateTimer.Stop()
        TestKeywords_RecolorLines()
    End Sub

    ' ===============================================================================================
    '  DOUBLE CLICK (complete words)
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_MouseDoubleClick(ByVal sender As Object, _
                                                   ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDoubleClick
        '
        ' --------------------------------------------------- complete word - QuotedString or single word
        Dim start, length As Int32
        FindQuotedStringPointers(Me.Text, Me.SelectionStart, start, length)
        If length > 0 Then
            SelectionStart = start
            SelectionLength = length
        Else
            FindWordPointers_ShortWords(Me.Text, Me.GetCharIndexFromPosition(e.Location), start, length)
            SelectionStart = start
            SelectionLength = length
        End If
    End Sub

    ' ===============================================================================================
    '   MOUSE DOWN
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        ' 
        Dim MouseDownSelectionIndex As Int32
        '
        If Not Running Then
            '
            ' --------------------------------------------------------- update the status bar SUGGESTIONS
            ShowSuggestions()
            '
            MouseDownSelectionIndex = Me.GetCharIndexFromPosition(e.Location)
            '
            ' This corrects the GetCharIndexFromPosition when clicking on the last char of the file
            '  (test with ERRORS_EDIT-2.txt)
            If MouseDownSelectionIndex = Me.Text.Length - 1 AndAlso Me.Text(Me.Text.Length - 1) <> vbLf Then
                MouseDownSelectionIndex += 1
            End If
            '
            If e.Button = MouseButtons.Right And Me.SelectionLength = 0 Then
                Me.Select(MouseDownSelectionIndex, 0)
            End If
            '
            OverwriteMode_ShowCaret(Me)
            '
            LastKeyPressed = Keys.None
            '
            If e.Button = Windows.Forms.MouseButtons.Left And Not My.Computer.Keyboard.ShiftKeyDown Then
                Me.Select(MouseDownSelectionIndex, 0)
            End If
        End If
        ' -------------------------------------------------------------
        If e.Button = Windows.Forms.MouseButtons.Left Then
            If Control.ModifierKeys = Keys.Control Then
                ' ----------------------------------------------------- toggle Breakpoint
                Dim p As Point = e.Location
                Me.ToggleBreakPoint(p)
                ' ----------------------------------------------------- toggle Bookmark
                'ElseIf Control.ModifierKeys = Keys.Alt Then
                '    Dim p As Point = e.Location
                '    Me.ToggleBreakPoint(p)
                '    '    'Me.ToggleBookMark(p)
                '    Console.Beep(880, 100)
            Else
                If Running Then
                    ' ------------------------------------------------- stop running
                    Me.PaintEventEnabled = False
                    SetFirstVisibleLine_BeforeMouseStop()
                    Form1.StopRunning()
                    RestoreFirstVisibleLine_BeforeMouseStop()
                    ' ------------------------------------------------- enable and focus RTB
                    Application.DoEvents()
                    Me.Enabled = True
                    Me.Focus()
                End If
            End If
        End If
        If Debugging Then
            Form1.JumpingFromNEXT = False
        End If
    End Sub

    Private FirstVisibleLine_BeforeMouseStop As Int32 = -1
    Private SelectionStart_BeforeMouseStop As Int32

    Friend Sub SetFirstVisibleLine_BeforeMouseStop()
        SelectionStart_BeforeMouseStop = Me.SelectionStart
        FirstVisibleLine_BeforeMouseStop = Me.GetFirstVisibleLine()
    End Sub

    Friend Sub RestoreFirstVisibleLine_BeforeMouseStop()
        If FirstVisibleLine_BeforeMouseStop < 0 Then Return
        Me.ScrollToLine(FirstVisibleLine_BeforeMouseStop)
        Me.SelectionStart = SelectionStart_BeforeMouseStop
        Me.SelectionLength = 0
        Me.PaintEventEnabled = True
    End Sub

    ' ===============================================================================================
    '   MOUSE UP
    ' ===============================================================================================
    Private Sub SyntaxRichTextBox_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        '
    End Sub

    ' ===============================================================================================
    '   SELECTION CHANGED
    ' ===============================================================================================
    Private OldCursorFirstLine As Int32 = -1
    Private OldCursorLastLine As Int32 = -1

    Private Sub RTB_SelectionChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.SelectionChanged
        '
        ' --------------------------------------------------------------------- AUTO WORD SELECTION BUG CORRECTION 
        Me.AutoWordSelection = True
        Me.AutoWordSelection = False
        ' --------------------------------------------------------------------- REMOVE LINE FEED FROM SELECTION
        RemoveLineFeedFromSelection()
        ' --------------------------------------------------------------------- TEST BEFORE AUTO INDENT
        If Not EventsAreEnabled Then Return
        If Running Then Return
        If Not Enabled Then Return
        ' --------------------------------------------------------------------- AUTO INDENT
        Dim Firstline As Int32 = GetLineFromCharIndex(Me.SelectionStart)
        Dim LastLine As Int32 = GetLineFromCharIndex(Me.SelectionStart + Me.SelectionLength)
        If Firstline <> OldCursorFirstLine Or LastLine <> OldCursorLastLine Then
            OldCursorFirstLine = Firstline
            OldCursorLastLine = LastLine
            ' ----------------------------------------------------------------- HISTORY
            History_AddEntry()
            ' ----------------------------------------------------------------- INDENT
            If LastKeyPressed <> Keys.Back And LastKeyPressed <> Keys.Cancel Then
                If TextMustBeIndented Then
                    Dim OldEventsAreEnabled As Boolean = EventsAreEnabled
                    EventsAreEnabled = False
                    AutoIndent() ' <------------------------------------------- INDENT ALL THE TEXT
                    EventsAreEnabled = OldEventsAreEnabled
                End If
            End If
            ' ----------------------------------------------------------------- REFRESH AND RED BORDER
            Refresh()
            Hilighter_RedBorderToSelectedLine()
        End If
        ' --------------------------------------------------------------------- SHOW LINES COUNT
        ShowLinesCount()
    End Sub

    Private Sub RemoveLineFeedFromSelection()
        If My.Computer.Keyboard.ShiftKeyDown Then Return
        If Me.SelectedText.EndsWith(vbLf) Then
            If Not Me.SelectedText.EndsWith(vbLf + vbLf) Then
                Me.SelectionLength -= 1
                Refresh()
                Hilighter_RedBorderToSelectedLine()
            End If
        End If
    End Sub


    ' ===============================================================================================
    '   LOST FOCUS / Mouse Leave / VScroll
    ' ===============================================================================================
    'Private Sub SyntaxRichTextBox_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
    '    '
    '    ' TODO -1- UNDO HYSTORY REMOVED
    '    'If UndoAndHistoryTextReady Then
    '    '    UndoAndHistory_AddEntry()
    '    'End If
    '    '
    '    'Dim line As Int32 = GetLineFromCharIndex(Me.SelectionStart)
    '    'If line <> OldCursorLine Then
    '    '    OldCursorLine = line
    '    '    Static OldFirstLine As Int32
    '    '    Dim firstline As Int32 = Me.GetLineFromCharIndex(Me.GetCharIndexFromPosition(New Point(1, 1)))
    '    '    If firstline <> OldFirstLine Then
    '    '        OldFirstLine = firstline
    '    '        History_AddEntry()
    '    '    End If
    '    'End If
    'End Sub




    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################
    ' 
    '   GRAPHICS - RED CURSOR LINE and HILIGHT RUNNING LINES
    ' 
    ' ###############################################################################################
    ' ###############################################################################################
    ' ###############################################################################################

    Private GFX As Graphics
    Private GFX_Font As Font
    Private GFX_Size As Size
    Private GFX_Zoom As Single
    Private GFX_PenRed As Pen = New Pen(Color.FromArgb(255, 100, 0), 2)

    Private Sub GFX_Init()
        If GFX Is Nothing Or Me.ClientSize <> GFX_Size Or ZoomFactor <> GFX_Zoom Then
            GFX_Size = Me.ClientSize
            GFX_Zoom = ZoomFactor
            GFX = Me.CreateGraphics()
        End If
    End Sub

    Friend Sub Hilighter_RedBorderToSelectedLine()
        If Not PaintEventEnabled Then Return
        If Running Then Return
        '
        Dim h As Single
        Dim DeltaUP As Int32
        Dim DeltaDW As Int32
        '
        h = Me.Font.GetHeight() * Me.ZoomFactor
        DeltaUP = -1
        DeltaDW = 1
        '
        Dim y1 As Int32 = Me.GetPositionFromCharIndex(Me.GetFirstCharIndexFromLine(GetFirstSelectedLine)).Y
        Dim y2 As Int32 = Me.GetPositionFromCharIndex(Me.GetFirstCharIndexFromLine(GetLastSelectedLine)).Y
        '
        If My.Computer.Keyboard.ShiftKeyDown Then
            If GetFirstSelectedLine() <> GetLastSelectedLine() Then
                y2 = Me.GetPositionFromCharIndex(Me.GetFirstCharIndexFromLine(GetLastSelectedLine_WithoutLineFeed())).Y
            End If
        End If
        GFX_Init()
        GFX.DrawLine(GFX_PenRed, 2, y1 + DeltaUP, Me.ClientSize.Width - 4, y1 + DeltaUP)
        GFX.DrawLine(GFX_PenRed, 2, y2 + h + DeltaDW, Me.ClientSize.Width - 4, y2 + h + DeltaDW)
    End Sub

    ' ==========================================================================================
    '  HILIGHTER FOR EXECUTED LINES 
    ' ==========================================================================================
    Private ExecutedLines(-1) As Single

    Friend Sub Hilighter_ClearExecutedLinesBuffer()
        ReDim ExecutedLines(RTB_Lines.Length - 1)
    End Sub

    Private Sub Hilighter_InitExecutedLinesBuffer()
        If ExecutedLines.Length <> RTB_Lines.Length Then
            Hilighter_ClearExecutedLinesBuffer()
        End If
    End Sub

    Friend Sub Hilighter_AddExecutedLine(ByVal n As Int32)
        Hilighter_InitExecutedLinesBuffer()
        ExecutedLines(n) = 1
    End Sub

    Private Sub HilightLine(ByVal line As Int32, ByVal b As Brush)
        Dim h As Single = 20 * Me.ZoomFactor
        Dim p As PointF = Me.GetPositionFromCharIndex(Me.GetLastCharIndexFromLine(line))
        If p.X < Me.ClientSize.Width - 20 Then
            GFX.FillRectangle(b, p.X + 2, p.Y, Me.ClientSize.Width - p.X, h + 2)
        End If
    End Sub

    Friend Sub Hilighter_ShowAllBufferedLines()
        If ExecutedLines.Length = 0 Then Return
        GFX_Init()
        Dim br1 As Brush
        '
        For i As Int32 = GetFirstVisibleLine() To GetLastVisibleLine()
            Dim n As Single = ExecutedLines(i)
            n = n * n
            Dim r As Int32 = CInt(220 + n * 35)
            Dim g As Int32 = CInt(225 - n * 80)  ' was 120
            Dim b As Int32 = CInt(225 - n * 225) ' was 200
            br1 = New SolidBrush(Color.FromArgb(r, g, b))
            HilightLine(i, br1)
        Next
        For i As Int32 = 0 To ExecutedLines.Length - 1
            ExecutedLines(i) -= 0.1F
            If ExecutedLines(i) < 0 Then
                ExecutedLines(i) = 0
            End If
        Next
    End Sub

End Class
