Imports System.Text

Public Class EvalFunctions

    ' ==================================================================
    '  SLOTS
    ' ==================================================================
    Function Slot(ByVal value As Double) As Double
        Return Slots.ReadSlot_NoNan(CInt(value))
    End Function

    Function DecodeCommandSlot(ByVal value As Double) As String
        Return NAN_GetName(Slots.ReadSlot(CInt(value)))
    End Function

    Function SlotText(ByVal value As Double) As String
        Return StringSlots.ReadString(CInt(value))
    End Function

    ' ==================================================================
    '  DecodeXML
    ' ==================================================================
    Function DecodeXML(ByVal xml As String, _
                       ByVal key As String, _
                       Optional ByVal n As Double = 1) As String
        If n <= 0 Then n = 1
        Dim i1 As Int32 = 0
        Dim i2 As Int32
        For j As Int32 = 1 To CInt(n)
            i2 = xml.IndexOf("<" + key + ">", i1, StringComparison.InvariantCultureIgnoreCase)
            If i2 < 0 Then
                i2 = xml.IndexOf("<" + key + " ", i1, StringComparison.InvariantCultureIgnoreCase)
            End If
            i1 = i2
            If i1 < 0 Then Return ""
            i1 = xml.IndexOf(">", i1 + key.Length + 1)
            If i1 < 0 Then Return ""
            i2 = xml.IndexOf("</" + key, i1 + 1, StringComparison.InvariantCultureIgnoreCase)
            If i2 < 0 Then Return ""
        Next
        Return xml.Substring(i1 + 1, i2 - i1 - 1)
    End Function

    ' ==================================================================
    '  FormatXML
    ' ==================================================================
    Function FormatXML(ByVal xml As String, Optional ByVal IndentSize As Double = 2) As String
        xml = xml.Replace("><", ">" + vbCrLf + "<")
        Dim str() As String = Strings.Split(xml, vbCrLf)
        Dim sb As StringBuilder = New StringBuilder
        Dim IndentSizeInt As Int32 = CInt(IndentSize)
        Dim Indent As Int32 = 0
        For n As Int32 = 0 To str.Length - 1
            '
            If str(n).StartsWith("</") Then
                Indent -= IndentSizeInt
                If Indent < 0 Then Indent = 0
            End If
            '
            sb.Append(StrDup(Indent, " ") + str(n) + vbCrLf)
            '
            If str(n).StartsWith("<?") Then
            ElseIf str(n).StartsWith("<!") Then
            ElseIf str(n).StartsWith("</") Then
            ElseIf str(n).StartsWith("<") Then
                Indent += IndentSizeInt
            End If
            '
            If Not str(n).StartsWith("</") And str(n).Contains("</") Then
                Indent -= IndentSizeInt
                If Indent < 0 Then Indent = 0
            End If
            '
            If str(n).Contains("/>") Then
                Indent -= IndentSizeInt
                If Indent < 0 Then Indent = 0
            End If
        Next
        Return sb.ToString
    End Function

    ' ==================================================================
    '  SELECT FILE and FOLDER
    ' ==================================================================
    Function SelectFile(ByVal Folder As String) As String
        '
        If Not Running Then Return ""
        '
        Folder = NormalizeFilename(Folder)
        '
        Return BrowseForFile(Folder, "Select file name")
    End Function


    ' ==================================================================
    '  ReadFileNames
    ' ==================================================================
    Function GetFileNames(ByVal TopFolder As String, Optional ByVal Sorted As Boolean = False, Optional ByVal Extension As String = "") As String
        '
        If Not Running Then Return ""
        '
        TopFolder = NormalizeFilename(TopFolder)
        '
        Dim s As String = ""
        Try
            ' -------------------------------------------------- get all files   
            Dim files() As String = IO.Directory.GetFiles(TopFolder, _
                                                          "*.*", _
                                                          IO.SearchOption.AllDirectories)
            ' -------------------------------------------------- sort the string array 
            If Sorted Then Array.Sort(files)
            ' -------------------------------------------------- copy to the output string
            For i As Int32 = 0 To files.Length - 1
                If Extension = "" OrElse files(i).ToLower.EndsWith(Extension) Then
                    s += IO.Path.GetFullPath(files(i)) + vbCrLf
                End If
            Next
            ' -------------------------------------------------- remove the final crlf
            If s.EndsWith(vbCrLf) Then
                s = s.Remove(s.Length - 2)
            End If
        Catch
        End Try
        Return s
    End Function


    ' ==================================================================
    '  GetFolderNames
    ' ==================================================================
    Function GetFolderNames(ByVal TopFolder As String, Optional ByVal Sorted As Boolean = False) As String
        '
        If Not Running Then Return ""
        '
        TopFolder = NormalizeFilename(TopFolder)
        '
        Dim s As String = ""
        Try
            ' -------------------------------------------------- get all files   
            Dim folders() As String = IO.Directory.GetDirectories(TopFolder, _
                                                                   "*.*", _
                                                                  IO.SearchOption.AllDirectories)
            ' -------------------------------------------------- sort the string array 
            If Sorted Then Array.Sort(folders)
            ' -------------------------------------------------- copy to the output string
            For i As Int32 = 0 To folders.Length - 1
                s += IO.Path.GetFullPath(folders(i)) + vbCrLf
            Next
            ' -------------------------------------------------- remove the final crlf
            If s.EndsWith(vbCrLf) Then
                s = s.Remove(s.Length - 2)
            End If
        Catch
        End Try
        Return s
    End Function

    ' ==================================================================
    '  FileExists
    ' ==================================================================
    Function FileExists(ByVal fname As String) As Boolean
        Return IO.File.Exists(fname)
    End Function

    ' ==================================================================
    '  AppRunning
    ' ==================================================================
    Function AppRunning(ByVal fname As String) As Boolean
        Return ProcessRunning(fname)
    End Function


    ' ==================================================================
    '  GetTextFile
    ' ==================================================================
    Function GetTextFile(ByVal fname As String) As String
        '
        If Not Running Then Return ""
        '
        '
        fname = NormalizeFilename(fname)
        '
        If Not FileIO.FileSystem.FileExists(fname) Then
            Return ""
        End If
        Dim reader As System.IO.TextReader
        reader = New System.IO.StreamReader(fname)
        Dim text As String = reader.ReadToEnd
        reader.Close()
        'text = Replace(text, vbCrLf, vbLf)
        'text = Replace(text, vbCr, vbLf)
        Return text
    End Function

    ' ==================================================================
    '  GetApplicationPath
    ' ==================================================================
    Function GetApplicationPath() As String
        Return Application.StartupPath
    End Function

    ' ==================================================================
    '  GetFilePath
    ' ==================================================================
    Function GetFilePath(ByVal FilePathAndName As String) As String
        If Not Running Then Return ""
        Return IO.Path.GetDirectoryName(IO.Path.GetFullPath(FilePathAndName))
    End Function

    ' ==================================================================
    '  GetFileNameWithoutPath
    ' ==================================================================
    Function GetFileNameWithoutPath(ByVal FilePathAndName As String) As String
        If Not Running Then Return ""
        Return IO.Path.GetFileName(FilePathAndName)
    End Function


    ' ==================================================================
    '  MATH
    ' ==================================================================
    Function sin(ByVal v As Double) As Double
        Return Math.Sin(v)
    End Function
    Function cos(ByVal v As Double) As Double
        Return Math.Cos(v)
    End Function
    Function asin(ByVal v As Double) As Double
        Return Math.Asin(v)
    End Function
    Function acos(ByVal v As Double) As Double
        Return Math.Acos(v)
    End Function
    Function tan(ByVal v As Double) As Double
        Return Math.Tan(v)
    End Function
    Function atan(ByVal v As Double) As Double
        Return Math.Atan(v)
    End Function
    Function atan2(ByVal v1 As Double, ByVal v2 As Double) As Double
        Return Math.Atan2(v1, v2)
    End Function
    Function sqrt(ByVal v1 As Double) As Double
        Return Math.Sqrt(v1)
    End Function
    Function abs(ByVal v1 As Double) As Double
        Return Math.Abs(v1)
    End Function
    Function sign(ByVal v1 As Double) As Double
        Return Math.Sign(v1)
    End Function
    Function min(ByVal v1 As Double, ByVal v2 As Double) As Double
        Return Math.Min(v1, v2)
    End Function
    Function max(ByVal v1 As Double, ByVal v2 As Double) As Double
        Return Math.Max(v1, v2)
    End Function
    Function [mod](ByVal n As Double, ByVal md As Double) As Double
        Return n Mod md
    End Function

    Function pow(ByVal v1 As Double, ByVal v2 As Double) As Double
        Return Math.Pow(v1, v2)
    End Function
    Function Exp(ByVal v As Double) As Double
        Return Math.Pow(Math.E, v)
    End Function
    Function Exp2(ByVal v As Double) As Double
        Return Math.Pow(2, v)
    End Function
    Function Exp10(ByVal v As Double) As Double
        Return Math.Pow(10, v)
    End Function
    Function Log(ByVal v As Double) As Double
        Return Math.Log(v, Math.E)
    End Function
    Function Log2(ByVal v As Double) As Double
        Return Math.Log(v, 2)
    End Function
    Function Log10(ByVal v As Double) As Double
        Return Math.Log(v, 10)
    End Function
    Function int(ByVal value As Double) As Double
        Return Microsoft.VisualBasic.Int(value)
    End Function
    Function round(ByVal value As Double) As Double
        Return Math.Round(value)
    End Function

    ' ==================================================================
    '  STRINGS
    ' ==================================================================
    Function mid(ByVal str As String, ByVal index As Double, Optional ByVal len As Double = 999999) As String
        If str = "" Or str Is Nothing Or index <= 0 Then Return ""
        Return Microsoft.VisualBasic.Mid(str, CInt(index), CInt(len))
    End Function
    Function len(ByVal str As String) As Double
        Return str.Length
    End Function
    Function trim(ByVal str As String) As String
        Return str.Trim
    End Function
    Function left(ByVal str As String, ByVal len As Double) As String
        If len < 0 Then len = 0
        If len > str.Length Then len = str.Length
        Return Microsoft.VisualBasic.Left(str, CInt(len))
    End Function
    Function right(ByVal str As String, ByVal len As Double) As String
        If len < 0 Then len = 0
        If len > str.Length Then len = str.Length
        Return Microsoft.VisualBasic.Right(str, CInt(len))
    End Function

    Function padleft(ByVal str As String, ByVal places As Double) As String
        Return str.PadLeft(CInt(places))
    End Function
    Function padright(ByVal str As String, ByVal places As Double) As String
        Return str.PadRight(CInt(places))
    End Function
    Function str(ByVal value As Object, Optional ByVal style As String = "0") As String
        Return Microsoft.VisualBasic.Format(value, style).Replace(",", ".")
    End Function
    Function ucase(ByVal value As String) As String
        Return Microsoft.VisualBasic.UCase(value)
    End Function
    Function lcase(ByVal value As String) As String
        Return Microsoft.VisualBasic.LCase(value)
    End Function
    Function wcase(ByVal value As String) As String
        If len(value) = 0 Then Return ""
        Return Microsoft.VisualBasic.UCase(value.Substring(0, 1)) & _
              Microsoft.VisualBasic.LCase(value.Substring(1))
    End Function
    Function chars(ByVal str As String, ByVal repetitions As Double) As String
        If str = "" Or str Is Nothing Or repetitions <= 0 Then Return ""
        Return Microsoft.VisualBasic.StrDup(CInt(repetitions), str)
    End Function
    Function Replace(ByVal str As String, ByVal oldStr As String, ByVal newStr As String) As String
        Return str.Replace(oldStr.ToString, newStr.ToString)
    End Function
    Function RemoveComments(ByVal str As String) As String
        Dim i As Int32 = str.IndexOf("'")
        If i = 0 Then str = ""
        If i > 0 Then
            i = str.Length
            Dim InsideString As Boolean = False
            For j As Int32 = 0 To str.Length - 1
                Select Case str(j)
                    Case """"c : InsideString = Not InsideString
                    Case "'"c
                        If Not InsideString Then
                            i = j
                            str = str.Remove(i).TrimEnd
                            Exit For
                        End If
                End Select
            Next
        End If
        Return str
    End Function

    Function StringContains(ByVal str As String, ByVal SubstringToSearch As String) As Boolean
        Return str.ToLower.Contains(SubstringToSearch.ToLower)
    End Function


    ' ==================================================================
    '  CONVERSIONS
    ' ==================================================================
    Function chr(ByVal n As Double) As String
        Return ChrW(CInt(n))
    End Function
    Function asc(ByVal s As String) As Double
        s = s.Replace("""", "")
        If s = "" Then Return 0
        Return AscW(s)
    End Function
    Function bin(ByVal n As Double) As String
        Return Convert.ToString(CInt(n), 2)
    End Function
    Function hex(ByVal n As Double) As String
        Return Convert.ToString(CInt(n), 16).ToUpper
    End Function
    Function frombin(ByVal s As String) As Double
        s = s.Replace("""", "")
        If s = "" Then Return 0
        Return Convert.ToInt32(s, 2)
    End Function
    Function fromhex(ByVal s As String) As Double
        s = s.Replace("""", "")
        If s = "" Then Return 0
        Return Convert.ToInt32(s, 16)
    End Function
    Function val(ByVal s As String) As Double
        If s = "" Then Return 0
        Return Val_Double(s)
    End Function
    Function format(ByVal value As Object, Optional ByVal style As String = "0") As String
        Return Microsoft.VisualBasic.Format(value, style).Replace(",", ".")
        'Return CDbl(value).ToString(style).Replace(",", ".")
    End Function

    ' ==================================================================
    '  CONSTANTS
    ' ==================================================================
    Function crlf() As String
        Return vbCrLf
    End Function
    Function cr() As String
        Return vbCr
    End Function
    Function lf() As String
        Return vbLf
    End Function
    Function nbsp() As String
        Return chr(160)
    End Function
    Function cls() As String
        Return chr(12)
    End Function
    Function tab() As String
        Return vbTab
    End Function
    Function math_pi() As Double
        Return Math.PI
    End Function
    Function math_e() As Double
        Return Math.E
    End Function
    Function infinity() As Double
        Return 9.0E+99
    End Function

    ' ==================================================================
    '  DATE AND TIME
    ' ================================================================== 
    Function now() As DateTime
        Return Microsoft.VisualBasic.Now
    End Function
    Function today() As DateTime
        Return Microsoft.VisualBasic.Today
    End Function
    Function [date](ByVal year As Double, ByVal month As Double, ByVal day As Double) As DateTime
        Return New Date(CInt(year), CInt(month), CInt(day))
    End Function

    ' ==================================================================
    '  STOPWATCH
    ' ==================================================================
    Function elapsedtime() As Double
        Return StartTime.Elapsed.TotalSeconds
    End Function

    ' ==================================================================
    '  RND
    ' ==================================================================
    Function rnd() As Double
        Microsoft.VisualBasic.Randomize()
        Return Microsoft.VisualBasic.Rnd()
    End Function

    ' ==================================================================
    '  COM
    ' ==================================================================
    Function com_status() As String
        Return COM_GetStatus()
    End Function
    Function com_received() As String
        Return COM_GetReceivedBuffer()
    End Function
    Function com_receiveuntil(ByVal s As String) As String
        Return COM_GetReceivedUntil(s)
    End Function
    Function com_receivenchars(ByVal n As Double) As String
        Return COM_GetReceivedNBytes(CInt(n))
    End Function
    Function com_portnames() As String
        Dim sa() As String = COM_GetPortNames()
        Dim names As String = ""
        For Each s As String In sa
            names += s + " "
        Next
        Return names.Trim
    End Function
    Function com_dsr() As Boolean
        Return COM_GetDSR()
    End Function
    Function com_cts() As Boolean
        Return COM_GetCTS()
    End Function

    ' ==================================================================
    '  GET SEPARATED STRINGS 
    '  - The default separator default is a single space
    '  - Valid separators are single chars or strings
    ' ==================================================================
    Function GetSeparatedStringCount(ByVal s As String, _
                                     Optional ByVal separator As String = " ") As Double
        '
        s = ReplaceMultipleSeparators(s, separator)
        Return Split(s, separator).Length
    End Function
    Function GetSeparatedString(ByVal s As String, ByVal n As Double, _
                                Optional ByVal separator As String = " ") As String
        '
        s = ReplaceMultipleSeparators(s, separator)
        Dim sa() As String = Split(s, separator)

        Dim i As Int32 = CInt(n)
        If i < 0 Or i >= sa.Length Then Return ""
        Return sa(i)
    End Function

    ' ==================================================================
    '  MOUSE
    ' ==================================================================
    Function MouseX() As Double
        Return GetNormalizedCursorX()
    End Function

    Function MouseY() As Double
        Return GetNormalizedCursorY()
    End Function

    Function MouseXP() As Double
        Return GetCursorX()
    End Function

    Function MouseYP() As Double
        Return GetCursorY()
    End Function

    Function MouseButtons() As Double
        Return Control.MouseButtons \ 1048576
    End Function

    ' ==================================================================
    '  KEY
    ' ==================================================================
    Function Key(ByVal k As String) As Boolean
        If Form.ActiveForm IsNot Nothing Or GlobalKeysAreEnabled Then
            Dim kk As Keys = GetKeyCode(k)
            If kk <> Nothing Then
                If KeyPressedNow(kk) Then Return True
            Else
                If KeyPressedNow(KeyCodeToVirtualKey(AscW(k))) Then Return True
            End If
        End If
        Return False
    End Function

    ' ==================================================================
    '  ExternalCommands - Function CommandText
    ' ==================================================================
    Function CommandText() As String
        Return String_ExternalCommand
    End Function


    ' ==================================================================
    '  GET CURSOR PIXEL COLOR
    ' ==================================================================
    Function GetCursorPixelColor() As Double
        Dim bmp As Bitmap = New Bitmap(1, 1)
        Dim g As Graphics = Graphics.FromImage(bmp)
        g.CopyFromScreen(Windows.Forms.Cursor.Position, New Point(0, 0), New Size(1, 1))
        Return bmp.GetPixel(0, 0).ToArgb And &HFFFFFF
    End Function

    ' --------------------------------------------------------------- GET XY PIXEL COLOR (0,0 is the lower left corner)
    Function GetPixelColor(ByVal x As Double, ByVal y As Double) As Double
        'Dim s As Screen = Screen.FromPoint(New Point(CInt(x), CInt(y)))
        Dim bmp As Bitmap = New Bitmap(1, 1)
        Dim g As Graphics = Graphics.FromImage(bmp)
        g.CopyFromScreen(New Point(CInt(x), CInt(1079 - y)), New Point(0, 0), New Size(1, 1))
        GetPixelColor = bmp.GetPixel(0, 0).ToArgb And &HFFFFFF
    End Function

    ' ==================================================================
    '  GET Red/Green/Blue FROM COLOR 
    ' ==================================================================
    Function GetColorRed(ByVal color As Double) As Double
        If color < 0 Or color > &HFFFFFF Then Return 0
        Return (CInt(color) And &HFF0000) \ &H10000
    End Function
    Function GetColorGreen(ByVal color As Double) As Double
        If color < 0 Or color > &HFFFFFF Then Return 0
        Return (CInt(color) And &HFF00) \ &H100
    End Function
    Function GetColorBlue(ByVal color As Double) As Double
        If color < 0 Or color > &HFFFFFF Then Return 0
        Return CInt(color) And &HFF
    End Function

    ' ==================================================================
    '  GET Hue/Saturation/Lightness FROM COLOR 
    ' ==================================================================
    Function GetColorHue(ByVal color As Double) As Double
        Return RGBToHSL(color).H
    End Function
    Function GetColorSaturation(ByVal color As Double) As Double
        Return RGBToHSL(color).S
    End Function
    Function GetColorLightness(ByVal color As Double) As Double
        Return RGBToHSL(color).L
    End Function

    ' ==================================================================
    '  GET ColorIndex(0 to 7) FROM COLOR 
    ' ==================================================================
    Function GetColorIndex(ByVal c As Double) As Double
        Return HslToColorIndex(RGBToHSL(c))
    End Function

    ' ==================================================================
    '  GET COLOR NAMES
    ' ==================================================================
    Function GetSimpleColorName(ByVal c As Double) As String
        Return HslToColorName(RGBToHSL(c))
    End Function
    Function GetNearestColorName(ByVal c As Double) As String
        If c < 0 Or c > &HFFFFFF Then Return ""
        Return Utils_GetNearestColorName(Color.FromArgb(CInt(c)))
    End Function

    ' ==================================================================
    '  LIMIT
    ' ==================================================================
    Private LimitFlag As Boolean = False

    Function Limit(ByVal n As Double, ByVal Min As Double, ByVal Max As Double) As Double
        If CSng(n) > CSng(Max) Then
            n = Max
            LimitFlag = True
        End If
        If CSng(n) < CSng(Min) Then
            n = Min
            LimitFlag = True
        End If
        Return n
    End Function

    Function LimitReached() As Boolean
        LimitReached = LimitFlag
        LimitFlag = False
    End Function

    ' ==================================================================
    '  MEDIA PLAYER
    ' ==================================================================
    Friend Shared WmpIsPlaying As Boolean = False
    Friend Shared WmpMediaLength As Double
    Friend Shared WmpMediaPosition As Double
    Function MediaPlaying() As Boolean
        Return WmpIsPlaying
    End Function

    Function MediaLength() As Double
        Return WmpMediaLength
    End Function

    Function MediaPosition() As Double
        Return WmpMediaPosition
    End Function

    ' ==================================================================
    '  FILTERS
    ' ==================================================================
    Function FilterLP(ByVal Value As Double, _
                      ByVal Freq As Double, _
                      ByVal Poles As Double) As Double
        Return ExecFilter(Value, Freq, Poles, False, LineInExecution)
    End Function
    Function FilterHP(ByVal Value As Double, _
                      ByVal Freq As Double, _
                      ByVal Poles As Double) As Double
        Return ExecFilter(Value, Freq, Poles, True, LineInExecution)
    End Function

    ' ==================================================================
    '  SEW FUNCTIONS
    ' ==================================================================
    Function SewProgress() As Double
        Return SewGetProgress()
    End Function

    ' ==================================================================
    '  TTS - TEXT TO SPEECH - FUNCTIONS
    ' ==================================================================
    Function TTSvoices() As String
        Return TTS_GetInstalledVoices()
    End Function

    Function TTSlanguage() As String
        Return TTS_GetLanguage()
    End Function

    Function TTSready() As Boolean
        Return TTS_SynthesizerReady()
    End Function

    Function TTSpaused() As Boolean
        Return TTS_SynthesizerPaused()
    End Function

    Function TTSspeaking() As Boolean
        Return TTS_SynthesizerSpeaking()
    End Function

End Class

