﻿
Module Module_BeepPlayer

    ' =============================================================================================
    '  PLAY BEEP ASYNC
    ' =============================================================================================
    Friend Sub PlayBeepAsync(ByVal frequency As Double, _
                             ByVal millisec As Double, _
                             ByVal sequence As String)
        If sequence = "" Then
            If frequency <= 0 And millisec <= 0 Then
                StopAllPlayThreads()
                Return
            End If
            BeepAsyncSequence = frequency.ToString("0") + " " + millisec.ToString("0")
        Else
            If IsStringVariable(sequence) Then
                sequence = EvalString(sequence)
            End If
            BeepAsyncSequence = sequence
        End If
        ' --------------------------------------------------------- asynchronously start the Thread 
        Dim objThread As Threading.Thread
        objThread = New Threading.Thread(AddressOf StartBeepAsyncThread)
        ' --------------------------------------------------------- make the thread as background thread.
        objThread.IsBackground = False
        ' --------------------------------------------------------- set the Priority of the thread.
        objThread.Priority = Threading.ThreadPriority.AboveNormal
        ' --------------------------------------------------------- start the thread
        StopAsyncThread = False
        objThread.Start()
    End Sub

    Private ThreadCounter As Int32 = 0
    Private StopAsyncThread As Boolean
    Private BeepAsyncSequence As String

    Private Sub StartBeepAsyncThread()
        ThreadCounter += 1
        BeepAsyncSequence = BeepAsyncSequence.Replace(vbLf, " ")
        BeepAsyncSequence = BeepAsyncSequence.Replace(vbCr, " ")
        BeepAsyncSequence = BeepAsyncSequence.Replace(",", " ")
        BeepAsyncSequence = BeepAsyncSequence.Replace(";", " ")
        BeepAsyncSequence = BeepAsyncSequence.Replace("-", " ")
        Dim sa() As String = ReplaceMultipleSpacesAndTabAndTrim(BeepAsyncSequence).Split(" "c)
        For i As Int32 = 0 To sa.Length - 2 Step 2
            If StopAsyncThread Then Exit For
            PlayBeepOrPause(Val(sa(i)), Val(sa(i + 1)))
            Threading.Thread.Sleep(1)
        Next
        ThreadCounter -= 1
    End Sub

    Friend Sub StopAllPlayThreads()
        MCI.Volume = 0
        MCI.StopWaiting()
        StopAsyncThread = True
        WaitAllThreadClosed()
        MCI.Volume = 100
    End Sub

    Private Sub WaitAllThreadClosed()
        For i As Int32 = 1 To 500
            If ThreadCounter <= 0 Then Exit For
            Threading.Thread.Sleep(1)
        Next
        ThreadCounter = 0
    End Sub

    Private Sub PlayBeepOrPause(ByVal frequency As Double, ByVal millisec As Double)
        If frequency = 0 And millisec > 0 Then
            Threading.Thread.Sleep(CInt(millisec))
        Else
            PlayBeep(frequency, millisec / 1000)
        End If
    End Sub

    ' =============================================================================================
    '  PLAY BEEP 
    ' =============================================================================================
    Private MCI As SoundPlayer = New SoundPlayer

    ' Frequency(Hz) 
    ' Duration(seconds)
    Private Sub PlayBeep(ByVal Frequency As Double, ByVal Duration As Double)
        ' ---------------------------------------------------------- Limit Frequency and Duration
        If Frequency < 1 Then Frequency = 1
        If Frequency > 20000 Then Frequency = 20000
        If Duration < 0.01 Then Duration = 0.01
        If Duration > 60 Then Duration = 60
        ' ---------------------------------------------------------- Recalculate duration to get a whole number of cycles
        Duration = CInt(Duration * Frequency) / Frequency
        ' ---------------------------------------------------------- Generate the sound
        Dim SampleRate As Int32 = 44100
        Dim FadeDuration As Double = 0.01
        Dim FadeSamples As Integer = CInt(FadeDuration * SampleRate)
        Dim Samples As Integer = CInt(Duration * SampleRate) + 1
        Dim Bytes(Samples * 2 - 1) As Byte
        Dim krad As Double = 2 * Math.PI * Frequency / SampleRate
        Dim s As Short
        For i As Integer = 0 To Samples - 1
            ' ------------------------------------------------------ Sinusoidal sample
            s = CShort(Short.MaxValue * Math.Sin(krad * i))
            ' ------------------------------------------------------ Add initial and final fading
            If i <= FadeSamples Then
                s = CShort(s * i / FadeSamples)
            End If
            If i >= Samples - FadeSamples Then
                s = CShort(s * (Samples - i) / FadeSamples)
            End If
            ' ------------------------------------------------------ Lower and Upper Bytes
            Bytes(i * 2) = CByte(s And &HFF)
            Bytes(i * 2 + 1) = CByte((s >> 8) And &HFF)
        Next
        ' ---------------------------------------------------------- New file name
        Dim fname As String = GetNewFname()
        ' ---------------------------------------------------------- Create the WAV file
        On Error Resume Next
        WriteWaveFile(fname, Bytes, SampleRate, 16, 1)
        ' ---------------------------------------------------------- Play the file 
        MCI.Play(fname, True)
        ' ---------------------------------------------------------- Delete the file
        IO.File.Delete(fname)
    End Sub

    Private FnameIndex As Int32 = 0
    Private Function GetNewFname() As String
        FnameIndex = (FnameIndex + 1) Mod 99999
        Return Application.StartupPath + "\Files\BeepTemp" + FnameIndex.ToString + ".wav"
    End Function

End Module
