﻿
Imports System.IO.Ports

Module Module_COM

    Private WithEvents ComPort As New SerialPort()

    Friend Sub COM_Open(ByVal PortName As String, ByVal BaudRate As Int32)
        If PortName = "" Then PortName = "COM1"
        If BaudRate < 1 Then BaudRate = 1
        ' -------------------------------------------- If the port is Open then close it
        COM_Close()
        ComPort.PortName = PortName
        ComPort.BaudRate = BaudRate
        ComPort.Parity = Parity.None
        ComPort.DataBits = 8
        ComPort.StopBits = StopBits.One
        ' -------------------------------------------- Set params
        ComPort.ReceivedBytesThreshold = 1
        ComPort.ReadBufferSize = 8192   ' minimum = 4096
        ComPort.WriteBufferSize = 8192  ' minimum = 2048
        ' -------------------------------------------- Start with DTR disabled
        ComPort.DtrEnable = False
        ' -------------------------------------------- little timout to avoid open close errors 
        ComPort.ReadTimeout = 20
        ComPort.WriteTimeout = 20
        Try
            ComPort.Open()
        Catch
        End Try
        ' -------------------------------------------- start with empty buffers
        COM_DiscardOutBuffer()
        COM_DiscardInBuffer()
    End Sub

    Friend Sub COM_Close()
        If ComPort.IsOpen Then
            Try
                ' ----------------------------- little timout to avoid open close errors 
                'ComPort.ReadTimeout = 1
                'ComPort.WriteTimeout = 1
                'Thread.Sleep(30)
                ComPort.Close()
                'Thread.Sleep(30)
            Catch
            End Try
        End If
    End Sub


    '' ==============================================================================
    ''   DATA RECEIVED EVENT
    '' ==============================================================================
    'Private Sub COM_DataReceived(ByVal sender As Object, _
    '                                 ByVal e As System.IO.Ports.SerialDataReceivedEventArgs) _
    '                                 Handles ComPort.DataReceived
    '    Try
    '        SyncLock LockObj
    '            While ComPort.BytesToRead > 0
    '                Dim c As Char = Chr(ComPort.ReadByte)
    '                If (c >= Chr(32) And c <= Chr(127)) Or c = vbCr Or c = vbLf Then
    '                    ReceivedBuffer += c
    '                End If
    '            End While
    '        End SyncLock
    '    Catch
    '    End Try
    'End Sub

    '' ==========================================================================================
    ''   DECODE A CHAR LINE
    '' ==========================================================================================
    'Private LockObj As Object = New Object
    'Private ReceivedBuffer As String = ""

    'Friend Function COM_GetReceivedBuffer() As String
    '    Dim s As String
    '    SyncLock LockObj
    '        s = ReceivedBuffer
    '        If s.EndsWith(vbCrLf) Or s.EndsWith(vbLf) Or s.EndsWith(vbCr) Then
    '            s = s.Replace(vbLf, vbCr).Replace(vbCr + vbCr, vbCr)
    '            s = s.Replace(vbCr, "")
    '            ReceivedBuffer = ""
    '            Return s
    '        End If
    '    End SyncLock
    '    Return ""
    'End Function



    ' ==============================================================================
    '   DATA RECEIVED EVENT
    ' ==============================================================================
    Private Sub COM_DataReceived(ByVal sender As Object, _
                                 ByVal e As System.IO.Ports.SerialDataReceivedEventArgs) _
                                 Handles ComPort.DataReceived
        Try
            SyncLock LockObj
                While ComPort.BytesToRead > 0
                    ReceivedBuffer += Convert.ToChar(ComPort.ReadByte)
                End While
            End SyncLock
        Catch
        End Try
    End Sub

    ' ==========================================================================================
    '   DECODE A CHAR LINE
    ' ==========================================================================================
    Private LockObj As Object = New Object
    Private ReceivedBuffer As String = ""

    Friend Function COM_GetReceivedBuffer() As String
        Dim s As String
        SyncLock LockObj
            s = ReceivedBuffer
            If s.EndsWith(vbCrLf) Or s.EndsWith(vbLf) Or s.EndsWith(vbCr) Then
                s = s.Replace(vbLf, vbCr).Replace(vbCr + vbCr, vbCr)
                s = s.Replace(vbCr, "")
                ReceivedBuffer = ""
                Return s
            End If
        End SyncLock
        Return ""
    End Function

    Friend Function COM_GetReceivedUntil(ByVal str As String) As String
        Dim s As String = ""
        SyncLock LockObj
            Dim n As Int32 = ReceivedBuffer.IndexOf(str)     ' -1 if not found
            If n >= 0 Then
                n += str.Length
                s = ReceivedBuffer.Substring(0, n)           ' zero based
                ReceivedBuffer = ReceivedBuffer.Substring(n) ' zero based
                Return s
            End If
        End SyncLock
        Return ""
    End Function

    Friend Function COM_GetReceivedNBytes(ByVal n As Int32) As String
        Dim s As String = ""
        SyncLock LockObj
            If ReceivedBuffer.Length >= n Then
                s = ReceivedBuffer.Substring(0, n)           ' zero based
                ReceivedBuffer = ReceivedBuffer.Substring(n) ' zero based
                Return s
            End If
        End SyncLock
        Return ""
    End Function



    ' ==============================================================================
    '   SEND
    ' ==============================================================================
    Friend Sub COM_WriteLine(ByVal text As String)
        Try
            If Not ComPort.IsOpen Then Return
            ComPort.Write(text + vbCrLf)
        Catch
        End Try
    End Sub

    Friend Sub COM_WriteString(ByVal text As String)
        Try
            If Not ComPort.IsOpen Then Return
            ComPort.Write(text)
        Catch
        End Try
    End Sub

    Friend Sub COM_WriteStringHEX(ByVal text As String)
        Try
            If Not ComPort.IsOpen Then Return
            text = ReplaceMultipleSpacesAndTabAndTrim(text)
            Dim sa() As String = text.Split(" "c)
            Dim bytes(sa.Length - 1) As Byte
            For i As Int32 = 0 To sa.Length - 1
                bytes(i) = CByte(Convert.ToInt32(sa(i), 16))
            Next
            ComPort.Write(bytes, 0, bytes.length)
        Catch
        End Try
    End Sub

    Friend Sub COM_WriteStringDEC(ByVal text As String)
        Try
            If Not ComPort.IsOpen Then Return
            text = ReplaceMultipleSpacesAndTabAndTrim(text)
            Dim sa() As String = text.Split(" "c)
            Dim bytes(sa.Length - 1) As Byte
            For i As Int32 = 0 To sa.Length - 1
                bytes(i) = CByte(Val(sa(i)))
            Next
            ComPort.Write(bytes, 0, bytes.Length)
        Catch
        End Try
    End Sub

    ' ==============================================================================
    '   FUNCTIONS
    ' ==============================================================================
    Friend Function COM_GetPortNames() As String()
        If SerialPort.GetPortNames().Length = 0 Then Return New String() {"No ports"}
        Return SerialPort.GetPortNames()
    End Function

    Friend Function COM_GetStatus() As String
        Return If(ComPort.IsOpen, "OPENED", "CLOSED")
    End Function

    Friend Function COM_GetDSR() As Boolean
        If Not ComPort.IsOpen Then Return False
        Return ComPort.DsrHolding
    End Function

    Friend Function COM_GetCTS() As Boolean
        If Not ComPort.IsOpen Then Return False
        Return ComPort.CtsHolding
    End Function

    ' ==============================================================================
    '   PROCEDURES
    ' ==============================================================================
    Friend Sub COM_DiscardInBuffer()
        If ComPort.IsOpen Then ComPort.DiscardInBuffer()
        ReceivedBuffer = ""
    End Sub

    Friend Sub COM_DiscardOutBuffer()
        If ComPort.IsOpen Then ComPort.DiscardOutBuffer()
    End Sub

    Friend Sub COM_EnableDTR()
        If ComPort.IsOpen Then ComPort.DtrEnable = True
    End Sub

    Friend Sub COM_EnableRTS()
        If ComPort.IsOpen Then ComPort.RtsEnable = True
    End Sub

    Friend Sub COM_DisableDTR()
        If ComPort.IsOpen Then ComPort.DtrEnable = False
    End Sub

    Friend Sub COM_DisableRTS()
        If ComPort.IsOpen Then ComPort.RtsEnable = False
    End Sub

End Module
