﻿Module Module_CompareStrings

    ' =========================================================================================
    '  FIND FOLDERS AND FILES
    ' =========================================================================================
    Friend Function FindFoldersAndFiles(ByVal StartingFolder As String, _
                                        ByVal words As String, _
                                        Optional ByVal ExcludeUrlLinks As Boolean = True) As String
        Dim i, j As Int32
        ' ----------------------------------------------------------------- 
        If Not StartingFolder.EndsWith("\") Then
            StartingFolder += "\"
        End If
        ' ----------------------------------------------------------------- 
        words = ReplaceMultipleSpacesAndTabAndTrim(words).ToLower
        ' ----------------------------------------------------------------- 
        Dim wds() As String = words.Split(" "c)
        For i = 0 To wds.Length - 1
            wds(i) = ReplaceAccentedChars(wds(i))
            wds(i) = wds(i).Replace("_", " ")
        Next
        ' ----------------------------------------------------------------- get all folders and files
        Dim folders() As String
        folders = IO.Directory.GetDirectories(StartingFolder, _
                                              "*.*", _
                                              IO.SearchOption.AllDirectories)
        Dim files() As String
        files = IO.Directory.GetFiles(StartingFolder, _
                                      "*.*", _
                                      IO.SearchOption.AllDirectories)
        ' -----------------------------------------------------------------
        Finder_MaxPoints = 0
        Dim points As Int32
        Dim retval As String = ""
        ' ------------------------------------------------- folders
        For i = 0 To folders.Length - 1
            Dim s As String = folders(i)
            s = s.Replace(StartingFolder, "").ToLower
            s = ReplaceAccentedChars(s)
            ' ---------------------------------------------
            points = 0
            For j = 0 To wds.Length - 1
                points += StringSimilarity(s, wds(j)) * 100
            Next
            ' ---------------------------------------------
            If points > Finder_MaxPoints Then
                Finder_MaxPoints = points
                retval = folders(i)
            End If
        Next
        ' ------------------------------------------------- files
        For i = 0 To files.Length - 1
            Dim s As String = files(i)
            s = s.Replace(StartingFolder, "").ToLower
            s = ReplaceAccentedChars(s)
            Dim s2 As String = GetFilenameRemovingInvalidChars(s)
            ' ---------------------------------------------
            points = 0
            For j = 0 To wds.Length - 1
                points += StringSimilarity(s2, wds(j)) * 100
            Next
            ' ---------------------------------------------
            If ExcludeUrlLinks Then
                If files(i).EndsWith(".url") Then Continue For
            End If
            ' ---------------------------------------------
            If points > Finder_MaxPoints Then
                Finder_MaxPoints = points
                retval = files(i)
            End If
        Next
        '
        Dim len As Int32
        'len = My.Computer.FileSystem.GetName(retval).Replace(" ", "").Length
        len = words.Replace(" ", "").Length
        If words.Length > 0 Then
            Finder_MaxPoints \= len
        Else
            Finder_MaxPoints = 0
        End If
        ' ----------------------------------------------------------------- 
        Return retval
    End Function

    Private Function StringSimilarity(ByVal s1 As String, ByVal s2 As String) As Integer
        Return LongestCommonSubstring(s1, s2)
    End Function

    ' ===============================================================================================
    '  GET FILENAME REMOVING INVALID CHARS
    ' ===============================================================================================
    Friend Function GetFilenameRemovingInvalidChars(ByVal filepath As String) As String
        Try
            Return IO.Path.GetFileName(filepath)
        Catch ex As Exception
            Return RemoveInvalidChars(filepath)
        End Try
    End Function

    Friend Function RemoveInvalidChars(ByVal filename As String) As String
        Return String.Concat(filename.Split(IO.Path.GetInvalidFileNameChars()))
    End Function

    Friend Function ReplaceInvalidChars(ByVal filename As String) As String
        Return String.Join("_", filename.Split(IO.Path.GetInvalidFileNameChars()))
    End Function


    ' =========================================================================================
    '  FIND MOST SIMILAR ACTION
    ' =========================================================================================
    Friend Finder_MaxPoints As Int32
    Friend Finder_BestActionText As String = ""
    Friend Finder_BestActionId As String = ""

    Friend Function FindMostSimilarAction(ByVal Text As String, _
                                          ByVal ActionId() As String, _
                                          ByVal ActionText() As String) As String
        ' -----------------------------------------------------------------
        Dim points, i, j As Int32
        Finder_MaxPoints = 0
        ' -----------------------------------------------------------------
        Text = Text.Trim.ToLower.Replace("_", " ")
        If Text = "" Then Return ""
        ' ----------------------------------------------------------------- 
        Text = ReplaceMultipleSpacesAndTabAndTrim(Text).ToLower
        ' ----------------------------------------------------------------- 
        Dim wds() As String = Text.Split(" "c)
        For i = 0 To wds.Length - 1
            wds(i) = ReplaceAccentedChars(wds(i))
            wds(i) = wds(i).Replace("_", " ")
        Next
        ' -----------------------------------------------------------------
        For i = 0 To ActionId.Length - 1
            points = 0
            For j = 0 To wds.Length - 1
                points += 100 * LongestCommonSubstring(wds(j), ActionId(i).Replace("_", " ").ToLower)
            Next
            If points > Finder_MaxPoints Then
                Finder_MaxPoints = points
                Finder_BestActionText = ActionText(i)
                Finder_BestActionId = ActionId(i)
            End If
        Next
        ' -----------------------------------------------------------------
        Dim len As Int32
        len = Finder_BestActionId.Replace("_", "").Length
        len = Math.Max(len, Text.Replace(" ", "").Length)
        '
        If Text.Length > 0 Then
            Finder_MaxPoints \= len
        Else
            Finder_MaxPoints = 0
        End If
        ' ----------------------------------------------------------------- 
        Return Finder_BestActionText
    End Function

    Private Function ReplaceAccentedChars(ByVal s As String) As String
        Dim tempBytes As Byte()
        tempBytes = System.Text.Encoding.GetEncoding("ISO-8859-8").GetBytes(s)
        Return System.Text.Encoding.UTF8.GetString(tempBytes)
    End Function


    ' =========================================================================================
    '  Longest Common Substring - Returns the length of longest common substring
    ' =========================================================================================
    Private Function LongestCommonSubstring(ByVal s1 As String, ByVal s2 As String) As Integer
        Dim l1 As Integer = s1.Length
        Dim l2 As Integer = s2.Length
        Dim Stuff(l1 + 1 - 1, l2 + 1 - 1) As Integer
        Dim result As Integer = 0
        '
        For i As Integer = 0 To l1
            For j As Integer = 0 To l2
                If i = 0 OrElse j = 0 Then
                    Stuff(i, j) = 0
                ElseIf s1(i - 1) = s2(j - 1) Then
                    Stuff(i, j) = Stuff(i - 1, j - 1) + 1
                    result = Math.Max(result, Stuff(i, j))
                Else
                    Stuff(i, j) = 0
                End If
            Next
        Next
        '
        'Debug.Print(s1 + "   " + s2 + "   " + result.ToString)
        Return result
    End Function


End Module
