﻿
Module OpenCloseApps

    ' ==============================================================================================================
    '   OPEN ALL THEREMINC APPS (in the APP Folder and Subfolders)
    ' ==============================================================================================================
    Friend Sub OpenAllThereminicApps()
        ' ------------------------------------------------------------------
        FindThereminicApps()
        ' ------------------------------------------------------------------
        For Each s As String In OpenedProcessesList.Keys
            If ProcessRunning(s) Then Continue For
            Dim p As Process = New Process()
            p.StartInfo.WorkingDirectory = Application.StartupPath
            p.StartInfo.FileName = s
            p.StartInfo.UseShellExecute = False
            p.Start()
        Next
    End Sub

    ' ===========================================================
    '   Find Applications in the APP Folder and Subfolders
    ' ===========================================================
    Private OpenedProcessesList As Generic.Dictionary(Of String, Int32) = New Generic.Dictionary(Of String, Int32)

    Friend Sub ClearTheOpenedProcessesList()
        OpenedProcessesList.Clear()
    End Sub

    Friend Sub AddProcessToOpenedProcessesList(ByVal filename As String)
        If Not OpenedProcessesList.ContainsKey(filename) Then
            OpenedProcessesList.Add(filename, 0)
        End If
    End Sub

    Private Sub FindThereminicApps()
        '
        Dim s As String = Application.StartupPath & "\Apps\"
        If Not My.Computer.FileSystem.DirectoryExists(s) Then Return
        ' ----------------------------------------------------------------------- Get files recursive to allow very long paths
        Dim fc() As String = GetFilesRecursive(s, "Theremino_*.exe")
        ' -----------------------------------------------------------------------
        For i As Int32 = 0 To fc.Count - 1
            s = fc(i).ToLower
            If Not s.Contains("vshost") And Not s.Contains("\sources\") And Not s.Contains("\obj\") Then
                If Not s.ToLower = System.Windows.Forms.Application.ExecutablePath.ToLower Then
                    ' ------------------------------------------------------------------ Add to OpenedProcesses dictionary (if the key is not present)
                    If Not OpenedProcessesList.ContainsKey(s) Then
                        OpenedProcessesList.Add(s, 0)
                    End If
                End If
            End If
        Next
    End Sub

    ' ======================================================================
    '   Find File Path by Name in the specified Folder and Subfolders
    ' ======================================================================
    Friend Function FindFilePathByName(ByVal appname As String, _
                                       ByVal StartingFolder As String) As String
        '
        appname = IO.Path.GetFileName(appname)
        If Not StartingFolder.EndsWith("\") Then StartingFolder += "\"
        ' ----------------------------------------------------------------------- Get files recursive to allow very long paths
        Dim fc() As String = GetFilesRecursive(StartingFolder, appname)
        ' ----------------------------------------------------------------------- Remove VSHOST and OBJ files
        For i As Int32 = 0 To fc.Count - 1
            Dim s As String = fc(i).ToLower
            If IO.Path.GetFileName(s).StartsWith("theremino") Then
                If Not s.Contains("vshost") And Not s.Contains("\sources\") And Not s.Contains("\obj\") Then
                    Return fc(i)
                End If
            Else
                Return fc(i)
            End If
        Next
        '
        Return appname
    End Function





    ' =============================================================================
    '   NORMALIZE FILENAME
    ' =============================================================================
    Friend Function NormalizeFilename(ByVal filename As String) As String
        ' ------------------------------------------------------------------ Prepare path
        If Not IO.Path.IsPathRooted(filename) Then
            filename = Application.StartupPath + "\" + filename
        End If
        ' ------------------------------------------------------------------ Correct path for /../..
        filename = IO.Path.GetFullPath(filename)
        ' ------------------------------------------------------------------ Normalize filename
        filename = filename.Replace("/", "\")
        Return (filename)
    End Function


    ' =============================================================================
    '   Process Running  (testing the complete path of the running app)
    ' =============================================================================
    Friend Function ProcessRunning(ByVal FileName As String) As Boolean
        If FileName = "" Then Return False
        FileName = NormalizeFilename(FileName)
        ' -------------------------------------------------------------- FIND FILE IF NOT EXISTS
        If Not IO.File.Exists(FileName) Then
            FileName = FindFilePathByName(FileName, Application.StartupPath + "\Apps\")
        End If
        ' ------------------------------------------------------------------ Prepare ProcessName to compare
        Dim ProcessName As String = IO.Path.GetFileNameWithoutExtension(FileName)
        ' ------------------------------------------------------------------ Compare with all processes
        Dim p() As System.Diagnostics.Process
        p = System.Diagnostics.Process.GetProcesses()
        For i As Int32 = 0 To p.Length - 1
            Try
                ' ---------------------------------------------------------- test ProcessName for fast execution
                If Not p(i).ProcessName.ToLower = ProcessName.ToLower Then
                    Continue For
                End If
                ' ---------------------------------------------------------- test FileName
                If p(i).MainModule.FileName.ToLower = FileName.ToLower Then
                    Return True
                End If
            Catch
            End Try
        Next
        Return False
    End Function

    ' ==============================================================================================================
    '   CLOSE ALL OPENED PROCESSES
    ' ==============================================================================================================
    Friend Sub CloseAllOpenedProcesses()
        Dim ProcessName As String
        Dim p() As System.Diagnostics.Process
        p = System.Diagnostics.Process.GetProcesses()
        For i As Int32 = 0 To p.Length - 1
            If p(i).Id = System.Diagnostics.Process.GetCurrentProcess.Id Then Continue For
            '
            For Each OpenedProc As System.Collections.Generic.KeyValuePair(Of String, Int32) In OpenedProcessesList
                ' ---------------------------------------------------------- test ProcessName for fast execution
                ProcessName = IO.Path.GetFileNameWithoutExtension(OpenedProc.Key)
                If Not p(i).ProcessName.ToLower = ProcessName.ToLower Then
                    Continue For
                End If
                Try
                    ' ---------------------------------------------------------- test FileName
                    If p(i).MainModule.FileName.ToLower = OpenedProc.Key.ToLower Then
                        '
                        'Debug.Print(p(i).ProcessName)
                        '
                        p(i).CloseMainWindow()
                        p(i).Close()
                        Exit For
                    End If
                Catch
                    ' --------------------- some process as Notepad generates errors when reading the MainModule
                    p(i).CloseMainWindow()
                    p(i).Close()
                    Exit For
                End Try
            Next
        Next
        '
        p = System.Diagnostics.Process.GetProcesses()
        For i As Int32 = 0 To p.Length - 1
            '
            If p(i).Id = System.Diagnostics.Process.GetCurrentProcess.Id Then Continue For
            '
            For Each OpenedProc As System.Collections.Generic.KeyValuePair(Of String, Int32) In OpenedProcessesList
                ' ---------------------------------------------------------- test ProcessName for fast execution
                ProcessName = IO.Path.GetFileNameWithoutExtension(OpenedProc.Key)
                If Not p(i).ProcessName.ToLower = ProcessName.ToLower Then
                    Continue For
                End If
                ' ---------------------------------------------------------- test FileName
                Try
                    If Not p(i).HasExited Then
                        If p(i).MainModule.FileName.ToLower = OpenedProc.Key.ToLower Then
                            '
                            'Debug.Print(p(i).ProcessName)
                            '
                            If Not p(i).WaitForExit(1000) Then
                                Try
                                    p(i).Kill()
                                Catch ex As Exception
                                End Try
                            End If
                            Exit For
                        End If
                    End If
                Catch
                End Try
            Next
        Next
    End Sub




    ' =============================================================================
    '   START PROCESS (IF NOT RUNNING) AND WAIT
    ' =============================================================================
    Friend Sub StartProcess_IfNotRunning_AndWait(ByVal filename As String, ByVal Wait As Boolean)
        ' ------------------------------------------------------------------ Prepare path
        If filename = "" Then Return
        filename = NormalizeFilename(filename)
        If Not IO.File.Exists(filename) Then Return
        ' ------------------------------------------------------------------ Add to OpenedProcesses dictionary
        AddProcessToOpenedProcessesList(filename)
        ' ------------------------------------------------------------------ Test if already running
        If ProcessRunning(filename) Then Return
        ' ------------------------------------------------------------------ Init process
        Dim proc As Process
        Dim psi As New ProcessStartInfo
        psi.WorkingDirectory = IO.Path.GetDirectoryName(filename)
        psi.FileName = filename
        ' ------------------------------------------------------------------ Protect for process errors
        Try
            ' -------------------------------------------------------------- Start process
            proc = Process.Start(psi)
            ' -------------------------------------------------------------- Wait the process really started 
            If Wait Then
                For i As Int32 = 1 To 200  ' 200x50mS = 10 to 15 seconds
                    Threading.Thread.Sleep(50)
                    proc.Refresh()
                    Try
                        If proc.MainWindowTitle.Length > 1 Then
                            Return
                        End If
                    Catch
                    End Try
                Next
            End If
        Catch
        End Try
    End Sub


    ' =============================================================================
    '   START PROCESS (ALSO IF RUNNING) AND WAIT
    ' =============================================================================
    Friend Sub StartProcess_AlsoIfRunning_AndWait(ByVal filename As String, ByVal Wait As Boolean)
        ' ------------------------------------------------------------------ Prepare path
        If filename = "" Then Return
        filename = NormalizeFilename(filename)
        If Not IO.File.Exists(filename) Then Return
        ' ------------------------------------------------------------------ Init process
        Dim proc As Process
        Dim psi As New ProcessStartInfo
        psi.WorkingDirectory = IO.Path.GetDirectoryName(filename)
        psi.FileName = filename
        ' ------------------------------------------------------------------ Protect for process errors
        Try
            ' -------------------------------------------------------------- Start process
            proc = Process.Start(psi)
            ' -------------------------------------------------------------- Wait the process really started 
            If Wait Then
                For i As Int32 = 1 To 200  ' 200x50mS = 10 to 15 seconds
                    Threading.Thread.Sleep(50)
                    proc.Refresh()
                    Try
                        If proc.MainWindowTitle.Length > 1 Then
                            Return
                        End If
                    Catch
                    End Try
                Next
            End If
        Catch
        End Try
    End Sub


    ' ========================================================
    '   PROCESS START AND WAIT PROCESS CLOSED
    ' ========================================================
    Friend Sub ProcessStartAndWait(ByVal fname As String)
        If Not My.Computer.FileSystem.FileExists(fname) Then Return
        Dim p As Process = Process.Start(fname)
        Do
            System.Threading.Thread.Sleep(100)
        Loop Until p.HasExited
    End Sub

    ' ========================================================
    '   PROCESS START  AND WAIT PROCESS CLOSED (NOTEPAD)
    ' ========================================================
    Friend Sub NotepadProcess_StartAndWait(ByVal fname As String)
        If Not My.Computer.FileSystem.FileExists(fname) Then Return
        Dim p As Process = Process.Start("notepad.exe", fname)
        Do
            System.Threading.Thread.Sleep(100)
        Loop Until p.HasExited
    End Sub


    ' =============================================================================
    '   Close App By Filename  (using ProcessName for fast execution)
    ' -----------------------------------------------------------------------------
    '  Valid paths are:
    '   - A complete path, for example C:\xxx\yyyy.exe
    '   - A partial path,  for example Files\yyyy.exe
    '   - A process name,  for example yyyy.exe
    '   - A process name,  for example yyyy
    ' =============================================================================
    Friend Sub CloseAppByFilename(ByVal FileName As String)
        If FileName = "" Then Return
        ' ------------------------------------------------------------------ transform statements and formula to a single string
        FileName = EvalString_AutoQuoting(FileName)
        ' ------------------------------------------------------------------ if the path is rooted the file must exist
        If IO.Path.IsPathRooted(FileName) Then
            If Not IO.File.Exists(FileName) Then Return
        Else
            ' -------------------------------------------------------------- else try to normalize it to automation folders
            Dim f As String = NormalizeFilename(FileName)
            If IO.File.Exists(f) Then
                FileName = f
            End If
        End If
        ' ------------------------------------------------------------------ if the file exists then use the process-name
        Dim p() As System.Diagnostics.Process
        p = System.Diagnostics.Process.GetProcesses()
        Dim ProcessName As String = IO.Path.GetFileNameWithoutExtension(FileName)
        For i As Int32 = 0 To p.Length - 1
            Try
                If p(i).Id = System.Diagnostics.Process.GetCurrentProcess.Id Then Continue For
                '
                If p(i).ProcessName.ToLower = ProcessName.ToLower Then
                    'Debug.Print(p(i).ProcessName)
                    p(i).CloseMainWindow()
                    p(i).Close()
                    Exit For
                End If
                '
                ' ----------------------------------------------------------------------------------
                ' Sometimes not working because the ProcessName is different from filename  
                ' For example when opening File.txt with Notepad
                '  - ProcessName     = "Notepad" 
                '  - MainWindowTitle = "File.txt - Notepad"
                ' ----------------------------------------------------------------------------------
                'If p(i).ProcessName.ToLower.StartsWith("notepad") Then
                '    Beep()
                'End If
                '
                If p(i).MainWindowTitle.ToLower.StartsWith(ProcessName.ToLower) Then
                    'Debug.Print(p(i).ProcessName)
                    p(i).CloseMainWindow()
                    p(i).Close()
                    Exit For
                End If
                '
            Catch
            End Try
        Next
        '
        p = System.Diagnostics.Process.GetProcesses()
        For i As Int32 = 0 To p.Length - 1
            Try
                If p(i).Id = System.Diagnostics.Process.GetCurrentProcess.Id Then Continue For
                If Not p(i).ProcessName.ToLower = ProcessName.ToLower Then Continue For
                '
                'Debug.Print(p(i).ProcessName)
                '
                If Not p(i).WaitForExit(1000) Then
                    Try
                        p(i).Kill()
                    Catch ex As Exception
                    End Try
                End If
            Catch
            End Try
        Next
    End Sub

End Module
