﻿
' =========================================================
'  THIS MODULE WORKS OK
'  BUT OPTIONS BEGIN-END ARE NOT IMPLEMENTED
' =========================================================

Module SewMachine

    ' ==================================================================================
    '   THREAD START / STOP
    ' ==================================================================================
    Private SewThread As Threading.Thread
    Private AbortThreadFlag As Boolean
    Private RunningProgress As Double = -1  ' -1 if not running / 0 to 1 if running

    Private Sub ThreadStart()
        ThreadAbort()
        ' --------------------------------------------------------- 
        SchedulerMaxPrecision()
        ' --------------------------------------------------------- start the Thread 
        SewThread = New Threading.Thread(AddressOf Thread_Execute)
        ' --------------------------------------------------------- make the thread as background thread.
        SewThread.IsBackground = True
        ' --------------------------------------------------------- set the Priority of the thread.
        SewThread.Priority = Threading.ThreadPriority.AboveNormal
        ' --------------------------------------------------------- start the thread
        RunningProgress = 0
        AbortThreadFlag = False
        SewThread.Start()
    End Sub

    Private Sub ThreadAbort()
        If SewThread Is Nothing Then Return
        If Not SewThread.IsAlive Then Return
        ' --------------------------------------------- try to join the thread
        AbortThreadFlag = True
        SewThread.Join(10)
        ' --------------------------------------------- catch because sometimes the thread is already closed
        Try
            For i As Int32 = 1 To 20
                If Not SewThread.IsAlive Then Exit For
                Threading.Thread.Sleep(10)
                Application.DoEvents()
            Next
            ' ----------------------------------------- eventually abort
            If SewThread.IsAlive Then
                MessageBox.Show("SewThread not closed correctly.", _
                                "Message from Automation", _
                                MessageBoxButtons.OK, MessageBoxIcon.Warning)
                SewThread.Abort()
            End If
        Catch
            ' ----------------------------------------- try catch needed to be sure to close the app
            Try
                SewThread.Abort()
            Catch
            End Try
        End Try
        ' ---------------------------------------------
        SewThread = Nothing
    End Sub


    ' ==================================================================================
    '   SEWING FUNCTIONS
    ' ==================================================================================
    Private SlotX As Int32 = 1
    Private SlotY As Int32 = 3
    Private SlotZ As Int32 = 5
    Private StopAngle As Int32 = 90
    Private StartAngle As Int32 = 270
    Private WaitTheStartAngle As Boolean

    Private Xmax As Double = 999
    Private Xmin As Double = 0
    Private Ymax As Double = 999
    Private Ymin As Double = 0
    Private MaxPerSec As Double = 20

    Private DestinationX As Double
    Private DestinationY As Double
    Private DistanceX As Double
    Private DistanceY As Double
    Private TotalMovementMM As Double
    Private IncrementX As Double
    Private IncrementY As Double
    Private IncrementAngle As Double
    Private IncrementZ As Double
    Private Zangle As Double
    Private OldZangle As Double
    Private StitchCount As Int32
    Private LoopsCount As Int32
    Private LoopTimeMS As Int32 = 2

    Private X, Y, Z As Double
    Private PX, PY As Double

    Private Pitch As Double
    Private Radius As Double = 0
    Private ZigZag As Double = 0
    Private StartPoints As Int32 = 0
    Private EndPoints As Int32 = 0

    Friend SewThreadPaused As Boolean = False
    Friend SewFunctionsEnabled As Boolean = False

    ' ------------------------------------------------------------------------------------------------
    '  The Stop_Angle  is the angle of the Z axis when the needle enters the fabric.
    '  The Start_Angle is the angle of the Z axis when the needle comes out of the fabric.
    ' ------------------------------------------------------------------------------------------------
    '  When the Z axis rotates to Start_Angle, the X and Y axes are moved to the next destination 
    '   with the maximum possible acceleration and speed.
    '  To compensate for the acceleration limits of the Z axis and minimize inaccuracies at high speed
    '   the Start_Angle setting should be increased from a standard value of 270° to approximately 350°
    ' ------------------------------------------------------------------------------------------------
    '  The StopAngle is not used in the "Sew" functions.
    '  The StopAngle is the angle at which the needle enters the fabric. 
    '  If the Z axis rotates past StopAngle, and X and Y have not yet reached their destination, 
    '   then the needle moves sideways and the seam becomes inaccurate. 
    '  This limits the maximum linear speed (mm/s) and the maximum sewing speed (stitches/sec).
    ' ------------------------------------------------------------------------------------------------

    Friend Sub SewInit(ByVal Slot_X As Int32, _
                       ByVal Slot_Y As Int32, _
                       ByVal Slot_Z As Int32, _
                       ByVal Stop_Angle As Int32, _
                       ByVal Start_Angle As Int32)
        SlotX = Slot_X
        SlotY = Slot_Y
        SlotZ = Slot_Z
        StopAngle = Stop_Angle
        StartAngle = Start_Angle
    End Sub

    Friend Sub SewLimits(ByVal X_min As Double, ByVal X_max As Double, _
                         ByVal Y_min As Double, ByVal Y_max As Double, _
                         ByVal Max_PerSec As Double)
        Xmax = X_max
        Xmin = X_min
        Ymax = Y_max
        Ymin = Y_min
        MaxPerSec = Max_PerSec
    End Sub

    Friend Sub SewAbort()
        ThreadAbort()
    End Sub

    Friend Function SewGetProgress() As Double
        Return RunningProgress
    End Function

    ' ---------------------------------------------------------------------------
    ' SewTo Options examples (all the options can be combined)
    ' ---------------------------------------------------------------------------
    '        DestX DestY Speed Pitch Options
    ' SewTo  70    35    20    4     "Radius=5"   ' Circle with radius = 5 mm
    ' SewTo  70    35    20    4     "ZigZag=2"   ' ZigZag displaced by +/-2 mm
    ' SewTo  10    30    20    5     "Begin=2"    ' 2 StartPoints with Pitch 5
    ' SewTo  10    30    20    5     "End=3"      ' 3 EndPoints with Pitch 5
    ' ---------------------------------------------------------------------------

    Friend Sub SewTo(ByVal DestX As Double, _
                     ByVal DestY As Double, _
                     ByVal Speed As Double, _
                     ByVal _Pitch As Double, _
                     ByVal Options As String)
        ' -------------------------------------------------
        If Not SewFunctionsEnabled Then Return
        ' ------------------------------------------------- init vars
        DestinationX = DestX
        DestinationY = DestY
        ' ------------------------------------------------- test options
        Dim i, i2 As Int32
        Dim sa() As String
        Options = Options.ToLower + " "
        ' ------------------------------------------------- test options - begin
        StartPoints = 0
        i = Options.IndexOf("begin=")
        If i >= 0 Then
            i += 6
            i2 = Options.IndexOf(" ", i)
            If i2 > i Then
                sa = Options.Substring(i, i2 - i).Split("_"c)
                If sa.Length >= 1 Then StartPoints = CInt(Val(sa(0)))
            End If
        End If
        ' ------------------------------------------------- test options - end
        EndPoints = 0
        i = Options.IndexOf("end=")
        If i >= 0 Then
            i += 4
            i2 = Options.IndexOf(" ", i)
            If i2 > i Then
                sa = Options.Substring(i, i2 - i).Split("_"c)
                If sa.Length >= 1 Then EndPoints = CInt(Val(sa(0)))
            End If
        End If
        ' ------------------------------------------------- test options - radius
        i = Options.IndexOf("radius=")
        Radius = 0
        If i >= 0 Then
            Radius = Val(Options.Substring(i + 6).Replace(",", ".").Replace("=", ""))
        End If
        ' ------------------------------------------------- test options - zigzag
        ZigZag = 0
        i = Options.IndexOf("zigzag=")
        If i >= 0 Then
            ZigZag = Val(Options.Substring(i + 6).Replace(",", ".").Replace("=", ""))
        End If
        ' ------------------------------------------------- test for max speed
        If _Pitch > 0 Then
            If Speed / _Pitch > MaxPerSec Then
                Speed = _Pitch * MaxPerSec
                MessageBox.Show("SEW-MACHINE ERROR - Too much speed" + vbCrLf + _
                                "SewTo_Speed reduced to " + Speed.ToString("0.0", GCI), _
                                "Message from Automation")
            End If
        End If
        ' ------------------------------------------------- read X / Y / Z
        X = Slots.ReadSlot(SlotX)
        Y = Slots.ReadSlot(SlotY)
        Z = Slots.ReadSlot(SlotZ)
        ' ------------------------------------------------- test for very low speed
        If Speed < 0.1 Then Speed = 1
        ' ------------------------------------------------- calc the movement in mm
        If Radius > 0 Then
            ' --------------------------------------------- for circle
            X = DestinationX
            Y = DestinationY
            TotalMovementMM = Radius * Math.PI * 2
        Else
            ' --------------------------------------------- for line
            DistanceX = DestinationX - X
            DistanceY = DestinationY - Y
            TotalMovementMM = CSng(Math.Sqrt(DistanceX * DistanceX + DistanceY * DistanceY))
        End If
        ' ------------------------------------------------- calc the loops count
        LoopsCount = CInt(Int((TotalMovementMM / _
                          (LoopTimeMS / 1000.0F)) / Speed))
        ' ------------------------------------------------- too much speed ?
        If LoopsCount = 0 Then
            Slots.WriteSlot(SlotX, Limit(DestinationX, Xmin, Xmax))
            Slots.WriteSlot(SlotY, Limit(DestinationY, Ymin, Ymax))
            Return
        End If
        ' ------------------------------------------------- calc the stitch number
        If _Pitch > 0 Then
            StitchCount = CInt(Int(TotalMovementMM / _Pitch))
            StitchCount = CInt(Int(TotalMovementMM / _Pitch + 0.8F)) ' accept 20% more pitch before increase
        Else
            StitchCount = 0
        End If
        ' ------------------------------------------------- prepare the increments
        If Radius > 0 Then
            ' --------------------------------------------- for circle
            IncrementAngle = 2 * Math.PI / LoopsCount
        Else
            ' --------------------------------------------- for line
            IncrementX = DistanceX / LoopsCount
            IncrementY = DistanceY / LoopsCount
        End If
        IncrementZ = CSng(StitchCount / LoopsCount)
        ' ------------------------------------------------- goto the circle start point
        If Radius > 0 Then
            ' --------------------------------------------- XY for circle
            Dim r As Double = Radius '+ ZigZag / 2
            PX = DestinationX + r * 0 ' SIN = 0
            PY = DestinationY + r * 1 ' COS = 1
            ' --------------------------------------------- write XY to slots
            Slots.WriteSlot(SlotX, Limit(PX, Xmin, Xmax))
            Slots.WriteSlot(SlotY, Limit(PY, Ymin, Ymax))
            WaitAllMotors()
        End If
        ' ------------------------------------------------- EXECUTE SEGMENT OR CIRCLE
        ThreadStart()
    End Sub

    Private Sub Thread_Execute()
        ' ------------------------------------------------------ EXECUTE START POINTS
        If Not AbortThreadFlag Then
            If StartPoints > 0 Then

            End If
        End If
        ' ------------------------------------------------------ EXECUTE SEGMENT OR CIRCLE
        WaitTheStartAngle = True
        ' ------------------------------------------------------ calculate two times to set the OldZangle too
        ComputeZangle()
        ComputeZangle()
        ' ------------------------------------------------------ loop for LoopsCount times
        Dim LoopIndex As Int32 = 1
        Do
            ' -------------------------------------------------- increment, move and wait
            If Not SewThreadPaused Or Not Z_IsUp() Then
                ' ----------------------------------------------
                IncrementPositionsAndWriteSlots(LoopIndex)
                ' ----------------------------------------------
                LoopIndex += 1
                RunningProgress = LoopIndex / LoopsCount
            End If
            Thread.Sleep(LoopTimeMS - 1)
        Loop Until LoopIndex > LoopsCount Or AbortThreadFlag
        ' ------------------------------------------------------ complete the movement
        If Not AbortThreadFlag Then
            If Radius > 0 Then
                ' ---------------------------------------------- XY for circle
                Dim r As Double = Radius + ZigZag / 2
                PX = DestinationX + r * 0 ' SIN = 0
                PY = DestinationY + r * 1 ' COS = 1
            Else
                ' ---------------------------------------------- XY for segment
                PX = DestinationX
                PY = DestinationY
            End If
            ' -------------------------------------------------- write XY to slots
            Slots.WriteSlot(SlotX, Limit(PX, Xmin, Xmax))
            Slots.WriteSlot(SlotY, Limit(PY, Ymin, Ymax))
        End If
        ' ------------------------------------------------------ round the needle position to UP
        Slots.WriteSlot(SlotZ, Math.Round(Slots.ReadSlot(SlotZ)))
        ' ------------------------------------------------------ wait all motors before to exit
        WaitAllMotors()
        ' ------------------------------------------------------ DO A FINAL POINT
        If Pitch > 0 And Not AbortThreadFlag Then
            Z = Z + 1
            Slots.WriteSlot(SlotZ, Z)
            WaitAllMotors()
        End If
        ' ------------------------------------------------------ EXECUTE END POINTS
        If Not AbortThreadFlag Then
            If EndPoints > 0 Then

            End If
        End If
        ' ------------------------------------------------------
        RunningProgress = -1
    End Sub

    Private Sub IncrementPositionsAndWriteSlots(ByVal LoopIndex As Int32)
        '
        If Zangle < StartAngle Then
            WaitTheStartAngle = True
        End If
        '
        If WaitTheStartAngle And Zangle > StartAngle Then
            WaitTheStartAngle = False

            If Radius > 0 Then
                ' ---------------------------------------------- zigzag for circle
                Dim r As Double = Radius
                If ZigZag > 0 Then
                    If CInt(Z) Mod 2 = 0 Then
                        r += ZigZag / 2
                    Else
                        r -= ZigZag / 2
                    End If
                End If
                ' ---------------------------------------------- XY for circle
                PX = X + r * Math.Sin(IncrementAngle * LoopIndex)
                PY = Y + r * Math.Cos(IncrementAngle * LoopIndex)
            Else
                ' ---------------------------------------------- XY for line
                PX = X + IncrementX * LoopIndex
                PY = Y + IncrementY * LoopIndex
                ' ---------------------------------------------- add zigzag
                If ZigZag > 0 Then
                    If CInt(Z) Mod 2 = 0 Then
                        PX += StitchCount * IncrementY * ZigZag / 2
                        PY += StitchCount * IncrementX * ZigZag / 2
                    Else
                        PX -= StitchCount * IncrementY * ZigZag / 2
                        PY -= StitchCount * IncrementX * ZigZag / 2
                    End If
                End If
            End If
            ' -------------------------------------------------- write XY to slots
            Slots.WriteSlot(SlotX, Limit(PX, Xmin, Xmax))
            Slots.WriteSlot(SlotY, Limit(PY, Ymin, Ymax))
        End If
        '
        ' ------------------------------------------------------ Z
        Z = Z + IncrementZ
        Slots.WriteSlot(SlotZ, Z)
        ComputeZangle()
    End Sub

    Private Sub ComputeZangle()
        OldZangle = Zangle

        ' TEST ZANGLE FROM MASTER ???
        Zangle = 360 * (Z Mod 1)
        'Zangle = 360 * ((Z - Slots.ReadSlot(SlotZ + 1)) Mod 1)

    End Sub

    Private Function Z_IsUp() As Boolean
        If Zangle < OldZangle Then
            Return True
        Else
            Return False
        End If
    End Function

    Private Function Limit(ByVal n As Double, ByVal Min As Double, ByVal Max As Double) As Double
        If n > Max Then n = Max
        If n < Min Then n = Min
        Return n
    End Function

    Private Sub WaitAllMotors()
        Thread.Sleep(10)
        Do
            If Math.Abs(Slots.ReadSlot(SlotX + 1)) < 0.01 Then
                If Math.Abs(Slots.ReadSlot(SlotY + 1)) < 0.01 Then
                    If Math.Abs(Slots.ReadSlot(SlotZ + 1)) < 0.01 Then
                        Exit Do
                    End If
                End If
            End If
            Thread.Sleep(1)
        Loop
    End Sub

End Module


