﻿Imports System.Collections.Generic
Imports System.Runtime.InteropServices

Module Module_Utils

    Friend GCI As Globalization.CultureInfo = Globalization.CultureInfo.InvariantCulture

    Friend Running As Boolean
    Friend Debugging As Boolean

    Friend PauseButtonsString As String = ""
    Friend PauseButtonSlot As Int32 = -1
    Friend PauseButtonSlotTripValue As Int32 = -1
    Friend PauseButtonSlotDirectionPositive As Boolean = True
    Friend PauseButtonOutputSlots() As Int32
    Friend PauseButtonOutputValues() As Int32
    Friend PauseButtonOutputSlotsFrequency As Single

    Friend RTB_Lines() As String
    Friend RTB_Text As String = ""

    Friend FirstErrorLine As Int32
    Friend ErrorsCounter As Int32
    Friend WarningsCounter As Int32

    Friend GlobalKeysAreEnabled As Boolean = False

    Friend SingleStep As Boolean
    Friend GotoLine As Int32 = -1
    Friend LineInExecution As Int32 = -1
    Friend LineBeforeRun As Int32 = -1

    Friend String_ExternalCommand As String = ""

    ' =============================================================================================
    '  WINDOWS SCHEDULER PRECISION
    ' ============================================================================================= 
    Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Friend Sub SchedulerMaxPrecision()
        If OperatingSystemIsWindows Then
            timeBeginPeriod(1)
        End If
    End Sub
    Friend Sub ShedulerDefaultPrecision()
        If OperatingSystemIsWindows Then
            timeEndPeriod(1)
        End If
    End Sub

    ' ----------------------------------------------------------------------------------
    '  StartTime STOP WATCH
    ' ----------------------------------------------------------------------------------
    Friend StartTime As Stopwatch = New Stopwatch
    Friend Sub SetStartTime()
        StartTime.Reset()
        StartTime.Start()
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Swap
    ' ----------------------------------------------------------------------------------
    Friend Sub Swap(Of T)(ByRef d1 As T, ByRef d2 As T)
        Dim d As T = d2
        d2 = d1
        d1 = d
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Concatenate Arrays
    ' ----------------------------------------------------------------------------------
    Friend Function ConcatArrays(Of T)(ByVal A() As T, ByVal B() As T) As T()
        Dim newArray(A.Length + B.Length - 1) As T
        Array.Copy(A, newArray, A.Length)
        Array.Copy(B, 0, newArray, A.Length, B.Length)
        Return newArray
    End Function

    ' ----------------------------------------------------------------------------------
    '  GetOrdinalSuffix
    ' ----------------------------------------------------------------------------------
    Friend Function GetOrdinalSuffix(ByVal n As Int32) As String
        Select Case n
            Case 1 : Return "st"
            Case 2 : Return "nd"
            Case 3 : Return "rd"
            Case Else : Return "th"
        End Select
    End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" the text up the end of line
    ' ----------------------------------------------------------------------------------
    Friend Function GetTextToEndOfFormula(ByVal text As String, _
                                          ByVal selectionIndex As Int32) As String
        GetTextToEndOfFormula = ""
        Dim c As Char
        Dim Counter As Int32 = -1
        For i As Int32 = selectionIndex To text.Length - 1
            c = text(i)
            Select Case c
                Case "("c
                    If Counter < 0 Then
                        Counter = 1
                    Else
                        Counter += 1
                    End If
                Case ")"c
                    Counter -= 1
                Case CChar(vbCr), CChar(vbLf), "'"c
                    Exit For
                Case ","c
                    If Counter <= 0 Then Exit For
            End Select
            If Counter >= -1 Then GetTextToEndOfFormula += text(i)
            If Counter = 0 Then Exit For
        Next
    End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" the text up the end of line
    ' ----------------------------------------------------------------------------------
    'Friend Function GetTextToEndOfLine(ByVal text As String, _
    '                                   ByVal selectionIndex As Int32) As String
    '    GetTextToEndOfLine = ""
    '    For i As Int32 = selectionIndex To text.Length - 1
    '        If text(i) = vbLf Then
    '            Exit For
    '        Else
    '            GetTextToEndOfLine += text(i)
    '        End If
    '    Next
    'End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" the complete word pointed by the "selectionIndex"
    ' ----------------------------------------------------------------------------------
    Friend Function GetCompleteWord(ByVal text As String, _
                                    ByVal selectionIndex As Int32) As String
        Dim start, length As Int32
        FindWordPointers(text, selectionIndex, start, length)
        Return text.Substring(start, length)
    End Function

    ' ----------------------------------------------------------------------------------------------
    '  Extract from "text" the complete word pointed by the "selectionIndex" (short words version)
    ' ----------------------------------------------------------------------------------------------
    Friend Function GetCompleteWord_Short(ByVal text As String, _
                                          ByVal selectionIndex As Int32) As String
        Dim start, length As Int32
        FindWordPointers_ShortWords(text, selectionIndex, start, length)
        Return text.Substring(start, length)
    End Function

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers(ByVal text As String, _
                                ByVal selectionIndex As Int32, _
                                ByRef start As Int32, _
                                ByRef lenght As Int32)
        Const separators As String = " " + vbCr + vbLf + vbTab
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If i1 - 1 < text.Length AndAlso InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    '  This version find shortest words - not including characters like  " , / () ...
    ' ----------------------------------------------------------------------------------
    Friend Sub FindWordPointers_ShortWords(ByVal text As String, _
                                           ByVal selectionIndex As Int32, _
                                           ByRef start As Int32, _
                                           ByRef lenght As Int32)
        '
        Const separators As String = " "";:,/\|()'" + vbCr + vbLf + vbTab
        '
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If i1 - 1 < text.Length AndAlso InStr(separators, text(i1 - 1)) > 0 Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                Exit For
            End If
        Next
        start = i1
        lenght = i2 - i1
    End Sub

    ' ----------------------------------------------------------------------------------
    '  Extract from "text" start and length of the word pointed by the "selectionIndex"
    '  This version find a complete string starting and finishing with double quote
    ' ----------------------------------------------------------------------------------
    Friend Sub FindQuotedStringPointers(ByVal text As String, _
                                        ByVal selectionIndex As Int32, _
                                        ByRef start As Int32, _
                                        ByRef lenght As Int32)
        '
        Const separators As String = ";:,/\|()'+" + vbCr + vbLf
        '
        Dim i1, i2 As Int32
        For i1 = selectionIndex To 1 Step -1
            If InStr(separators, text(i1 - 1)) > 0 Then
                lenght = 0
                Return
            End If
            If text(i1 - 1) = """" Then
                Exit For
            End If
        Next
        For i2 = selectionIndex To text.Length - 1
            If InStr(separators, text(i2)) > 0 Then
                lenght = 0
                Return
            End If
            If text(i2) = """" Then
                Exit For
            End If
        Next
        ' ---------------------------------- return a valid range
        start = i1
        lenght = i2 - i1
    End Sub


    ' ----------------------------------------------------------------------------------
    '  Extract numeric params from a string
    '  Valid only if fields are not more than MaxFieldNumber 
    ' ----------------------------------------------------------------------------------
    Friend Function ExtractNumericParams(ByVal s As String, ByVal MaxFieldsNumber As Int32) As Double()
        Dim n(-1) As Double
        Dim sa As String() = s.Split(" "c)
        Dim paramsCount As Int32 = 0
        For i As Int32 = 0 To sa.Length - 1
            If sa(i) <> "" Then
                If IsValidDouble(sa(i)) Then
                    ReDim Preserve n(n.Length)
                    n(n.Length - 1) = EvalDouble(sa(i))
                End If
                paramsCount += 1
            End If
        Next
        If paramsCount > MaxFieldsNumber Then ReDim n(-1)
        Return n
    End Function

    ' =======================================================================================
    '  TrayWnd
    ' =======================================================================================
    'Private Declare Function FindWindow Lib "user32" Alias "FindWindowA" (ByVal lpClassName As String, ByVal lpWindowName As String) As Int32
    'Private Declare Function ShowWindow Lib "user32" (ByVal hwnd As Int32, ByVal nCmdShow As Int32) As Int32
    'Private Const SW_HIDE As Int32 = 0
    'Private Const SW_SHOW As Int32 = 5

    'Friend Sub HideTrayWnd()
    '    Dim lngHandle As Int32
    '    lngHandle = FindWindow("Shell_TrayWnd", "")
    '    ShowWindow(lngHandle, SW_HIDE)
    'End Sub

    'Friend Sub ShowTrayWnd()
    '    Dim lngHandle As Int32
    '    lngHandle = FindWindow("Shell_TrayWnd", "")
    '    ShowWindow(lngHandle, SW_SHOW)
    'End Sub

   
    ' =============================================================================
    '   MyExecFileIsAlreadyRunning (testing the complete path of the running app)
    ' =============================================================================
    Friend Function MyExecFileIsAlreadyRunning(ByVal FileName As String) As Boolean
        If FileName = "" Then Return False
        FileName = NormalizeFilename(FileName)
        If Not IO.File.Exists(FileName) Then Return False
        ' ------------------------------------------------------------------ Prepare ProcessName to compare
        Dim ProcessName As String = IO.Path.GetFileNameWithoutExtension(FileName)
        ' ------------------------------------------------------------------ Compare with all processes
        Dim p() As System.Diagnostics.Process
        p = System.Diagnostics.Process.GetProcesses()
        For i As Int32 = 0 To p.Length - 1
            Try
                ' ---------------------------------------------------------- test ProcessName for fast execution
                If Not p(i).ProcessName.ToLower = ProcessName.ToLower Then
                    Continue For
                End If
                ' ---------------------------------------------------------- test FileName
                If p(i).MainModule.FileName.ToLower = FileName.ToLower Then
                    ' ------------------------------------------------------ exclude myself
                    If p(i).Id <> System.Diagnostics.Process.GetCurrentProcess().Id Then
                        Return True
                    End If
                End If
            Catch
            End Try
        Next
        Return False
    End Function

    ' ==============================================================
    '  REPLACE MULTIPLE SPACES and TABS and TRIM
    '  (3..5 uS for Strings with 32 chars)
    ' ============================================================== 
    Friend Function ReplaceMultipleSpacesAndTabAndTrim(ByVal s As String) As String
        s = s.Replace(vbTab, " ")
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s.Trim
    End Function

    'Friend Function ReplaceMultipleSpacesAndTrim_PreserveQuotedSpaces(ByVal s As String) As String
    '    ' ------------------------------------------------- replace inside quotes spaces with chr(255)
    '    Dim InsideQuotes As Boolean = False
    '    Dim s2 As String = ""
    '    For i As Int32 = 0 To s.Length - 1
    '        If s(i) = """" Then InsideQuotes = Not InsideQuotes
    '        If InsideQuotes Then
    '            If s(i) = " " Then
    '                s2 += Chr(160)   'NBSP
    '            Else
    '                s2 += s(i)
    '            End If
    '        Else
    '            s2 += s(i)
    '        End If
    '    Next
    '    ' ------------------------------------------------- remove all double spaces
    '    While s2.Contains("  ")
    '        s2 = s2.Replace("  ", " ")
    '    End While
    '    ' ------------------------------------------------- replace inside quotes chr(255) with spaces
    '    's2 = s2.Replace(Chr(160), " ")
    '    ' ------------------------------------------------- trim
    '    Return s2.Trim
    'End Function

    Friend Function ReplaceMultipleSeparators(ByVal s As String, ByVal separator As String) As String
        If separator = " " Then
            Return ReplaceMultipleSpacesAndTabAndTrim(s)
        Else
            While s.Contains(separator + separator)
                s = s.Replace(separator + separator, separator)
            End While
            Return s.Trim
        End If
    End Function

    ' =============================================================================
    '   GetRightPartAfterThisWord
    ' =============================================================================
    Friend Function GetRightPartAfterThisLowerCaseWord(ByVal str As String, ByVal word As String) As String
        Dim i As Int32 = str.ToLower.IndexOf(" " + word + " ")
        If i < 0 Then i = str.ToLower.IndexOf(vbTab + word + " ")
        If i < 0 Then i = str.ToLower.IndexOf(" " + word + vbTab)
        If i < 0 Then i = str.ToLower.IndexOf(vbTab + word + vbTab)
        i += word.Length + 1
        If i > 1 And i < str.Length Then
            Return str.Remove(0, i)
        Else
            Return ""
        End If
    End Function

    ' ======================================================================================================
    '  REMOVE COMMENTS
    ' ======================================================================================================
    Friend Function RemoveComments(ByRef s As String) As Int32
        Dim i As Int32 = s.IndexOf("'")
        If i = 0 Then s = ""
        If i > 0 Then
            i = s.Length
            Dim InsideString As Boolean = False
            For j As Int32 = 0 To s.Length - 1
                Select Case s(j)
                    Case """"c : InsideString = Not InsideString
                    Case "'"c
                        If Not InsideString Then
                            i = j
                            s = s.Remove(i).TrimEnd
                            Exit For
                        End If
                End Select
            Next
        End If
        Return i
    End Function

    ' ======================================================================================================
    '  ASYNC KEY STATE
    ' ======================================================================================================
    Private Declare Function GetAsyncKeyState Lib "user32" (ByVal vKey As Int32) As Int16
    Private Const PRESSED_NOW As Int32 = &H8000
    Private Const PRESSED_AFTER_PREVIOUS_CALL As Int32 = &H1

    Friend Function KeyPressedNow(ByVal k As Int32) As Boolean
        KeyPressedNow = (GetAsyncKeyState(k) And PRESSED_NOW) <> 0
    End Function

    Friend Function KeyPrevious(ByVal k As Int32) As Boolean
        KeyPrevious = (GetAsyncKeyState(k) And PRESSED_AFTER_PREVIOUS_CALL) <> 0
    End Function

    Private Declare Function VkKeyScan Lib "user32" Alias "VkKeyScanA" (ByVal chr As Int16) As Int16
    Friend Function KeyCodeToVirtualKey(ByVal k As Int32) As Int16
        If k > 65535 Then k = 0
        Return VkKeyScan(CShort(k))
    End Function

    Friend Sub WaitMouseOff()
        While KeyPressedNow(Keys.LButton) Or KeyPressedNow(Keys.MButton) Or KeyPressedNow(Keys.RButton)
            Thread.Sleep(1)
        End While
    End Sub

    Friend Sub WaitKeyOff(ByVal WaitKey As Int32)
        Do
            Thread.Sleep(1)
        Loop Until Not KeyPressedNow(WaitKey)
    End Sub

    ' https://learn.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes
    Friend Function GetKeyCode(ByVal str As String) As Keys
        Select Case str.ToLower
            Case "1" : Return Keys.D1
            Case "2" : Return Keys.D2
            Case "3" : Return Keys.D3
            Case "4" : Return Keys.D4
            Case "5" : Return Keys.D5
            Case "6" : Return Keys.D6
            Case "7" : Return Keys.D7
            Case "8" : Return Keys.D8
            Case "9" : Return Keys.D9
            Case "0" : Return Keys.D0
            Case "a" : Return Keys.A
            Case "b" : Return Keys.B
            Case "c" : Return Keys.C
            Case "d" : Return Keys.D
            Case "e" : Return Keys.E
            Case "f" : Return Keys.F
            Case "g" : Return Keys.G
            Case "h" : Return Keys.H
            Case "i" : Return Keys.I
            Case "j" : Return Keys.J
            Case "k" : Return Keys.K
            Case "l" : Return Keys.L
            Case "m" : Return Keys.M
            Case "n" : Return Keys.N
            Case "o" : Return Keys.O
            Case "p" : Return Keys.P
            Case "q" : Return Keys.Q
            Case "r" : Return Keys.R
            Case "s" : Return Keys.S
            Case "t" : Return Keys.T
            Case "u" : Return Keys.U
            Case "v" : Return Keys.V
            Case "w" : Return Keys.W
            Case "x" : Return Keys.X
            Case "y" : Return Keys.Y
            Case "z" : Return Keys.Z
            Case "f1" : Return Keys.F1
            Case "f2" : Return Keys.F2
            Case "f3" : Return Keys.F3
            Case "f4" : Return Keys.F4
            Case "f5" : Return Keys.F5
            Case "f6" : Return Keys.F6
            Case "f7" : Return Keys.F7
            Case "f8" : Return Keys.F8
            Case "f9" : Return Keys.F9
            Case "f10" : Return Keys.F10
            Case "f11" : Return Keys.F11
            Case "f12" : Return Keys.F12
            Case "num0" : Return Keys.NumPad0
            Case "num1" : Return Keys.NumPad1
            Case "num2" : Return Keys.NumPad2
            Case "num3" : Return Keys.NumPad3
            Case "num4" : Return Keys.NumPad4
            Case "num5" : Return Keys.NumPad5
            Case "num6" : Return Keys.NumPad6
            Case "num7" : Return Keys.NumPad7
            Case "num8" : Return Keys.NumPad8
            Case "num9" : Return Keys.NumPad9
            Case "left" : Return Keys.Left
            Case "right" : Return Keys.Right
            Case "up" : Return Keys.Up
            Case "down" : Return Keys.Down
            Case "space" : Return Keys.Space
            Case "esc" : Return Keys.Escape
            Case "pageup" : Return Keys.PageUp
            Case "pagedown" : Return Keys.PageDown
            Case "home" : Return Keys.Home
            Case "end" : Return Keys.End
            Case "shift" : Return Keys.ShiftKey
            Case "ctrl" : Return Keys.ControlKey
            Case "alt" : Return Keys.Menu
            Case "canc" : Return Keys.Delete
            Case "back" : Return Keys.Back
            Case "enter" : Return Keys.Enter
            Case "ins" : Return Keys.Insert
            Case "print" : Return Keys.PrintScreen ' Stamp
            Case "scroll" : Return Keys.Scroll
            Case "pause" : Return Keys.Pause
        End Select
        Return Keys.None
    End Function


    ' ==============================================================
    '  RICH TEXT BOX - SUPPRESS SOUNDS (call from KeyDown)
    ' ==============================================================
    Friend Sub RichTextBox_SuppressSounds(ByVal sender As Object, ByRef e As KeyEventArgs)
        Dim rtb As RichTextBox = CType(sender, RichTextBox)
        If rtb.GetLineFromCharIndex(rtb.SelectionStart) = 0 AndAlso _
                                            (e.KeyData = Keys.Up Or e.KeyData = Keys.PageUp) _
        OrElse rtb.GetLineFromCharIndex(rtb.SelectionStart) = rtb.GetLineFromCharIndex(rtb.TextLength) AndAlso _
                                            (e.KeyData = Keys.Down Or e.KeyData = Keys.PageDown) _
        OrElse rtb.SelectionStart = rtb.TextLength AndAlso e.KeyData = Keys.Right _
        OrElse rtb.SelectionStart = 0 AndAlso _
                (e.KeyData = Keys.Left Or (e.KeyData = Keys.Back And rtb.SelectionLength = 0)) Then
            e.Handled = True
        End If
    End Sub


    ' ======================================================================================================
    '  DEBUG FUNCTIONS (Debug window)
    ' ======================================================================================================
    Private DebugStopWatch As Stopwatch = New Stopwatch
    Friend Sub Debug_InitTime()
        DebugStopWatch.Reset()
        DebugStopWatch.Start()
    End Sub
    Friend Sub Debug_Print(ByVal s As String)
        Debug.Print(DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000", GCI).PadLeft(8) + " " + s)
    End Sub
    Friend Sub Debug_PrintMicro(ByVal s As String)
        Debug.Print(DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000000", GCI).PadLeft(8) + " " + s)
    End Sub
    Friend Sub Debug_PrintDiffTime(ByVal s As String)
        Debug.Print(DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000", GCI).PadLeft(8) + " " + s)
        Debug_InitTime()
    End Sub

    ' ======================================================================================================
    '  DEBUG FUNCTIONS (Form1 text)
    ' ======================================================================================================
    'Private DebugStopWatch As Stopwatch = New Stopwatch
    'Friend Sub Debug_InitTime()
    '    DebugStopWatch.Reset()
    '    DebugStopWatch.Start()
    'End Sub
    'Friend Sub Debug_Print(ByVal s As String)
    '    Form1.Text = DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000", GCI).PadLeft(8) + " " + s
    'End Sub
    'Friend Sub Debug_PrintMicro(ByVal s As String)
    '    Form1.Text = DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000000", GCI).PadLeft(8) + " " + s
    'End Sub
    'Friend Sub Debug_PrintDiffTime(ByVal s As String)
    '    Form1.Text = DebugStopWatch.Elapsed.TotalSeconds.ToString("0.000", GCI).PadLeft(8) + " " + s
    '    Debug_InitTime()
    'End Sub


    ' ======================================================================================================
    '  OVERWRITE MODE CARET
    ' ======================================================================================================
    Private Declare Function CreateCaret Lib "user32" (ByVal hWnd As IntPtr, _
                                                      ByVal hBitmap As IntPtr, _
                                                      ByVal nWidth As Int32, _
                                                      ByVal nHeight As Int32) As Int32

    Private Declare Function ShowCaret Lib "user32" (ByVal hWnd As IntPtr) As Int32

    Private OverwriteMode As Boolean

    Friend Sub OverwriteMode_Toggle(ByRef rtb As Theremino_Automation.SyntaxRichTextBox)
        OverwriteMode = Not OverwriteMode
        If Not OverwriteMode Then
            CreateCaret(rtb.Handle, IntPtr.Zero, 1, CInt((rtb.Font.Height - 1) * rtb.ZoomFactor))
            ShowCaret(rtb.Handle)
        End If
    End Sub

    Friend Sub OverwriteMode_ShowCaret(ByRef rtb As SyntaxRichTextBox)
        If OverwriteMode Then
            CreateCaret(rtb.Handle, IntPtr.Zero, CInt(12 * rtb.ZoomFactor), CInt((rtb.Font.Height - 1) * rtb.ZoomFactor))
            ShowCaret(rtb.Handle)
        End If
    End Sub


    ' ======================================================================================================
    '  STRING SPACED AROUND OPERATORS 
    ' ======================================================================================================
    Friend Function StringSpacedAroundOperators(ByVal s As String) As String
        s = SpacesAroundOperators(s)
        s = ReplaceMultipleSpacesAndTabAndTrim(s)
        Return " " + s + " "
    End Function

    Private SubtractionChars As String = ChrW(&H2D) + " " + _
                                         ChrW(&H2010) + " " + _
                                         ChrW(&H2212) + " " + _
                                         ChrW(&H2013) + " " + _
                                         ChrW(&H2014)

    Dim Operators As String = SubtractionChars + " ( ) * / + = ^ ,"

    Dim OperatorsArray() As String = Operators.Replace("  ", " ").Split(" "c)

    Private Function SpacesAroundOperators(ByVal s As String) As String
        If s = "" Then Return s
        For Each Str As String In OperatorsArray
            If Str <> "" Then
                s = s.Replace(Str, " " + Str + " ")
            End If
        Next
        Return s
    End Function

    ' ======================================================================================================
    '  MOUSE POSITION NORMALIZED AND LIMITED TO THE SCREEN CONTAINING THE FORM
    ' ------------------------------------------------------------------------------------------------------
    '  To call from another Thread, the Form1_Location is updated in the Form_LocationChanged event
    ' ------------------------------------------------------------------------------------------------------
    '  The 0.0 position is the lower left corner
    ' ======================================================================================================
    Friend Form1_Location As Point

    Friend Function GetNormalizedCursorX() As Double
        ' ---------------------------------------------------- get the cursor position in screen coordimates
        Dim mousePoint As Point = Control.MousePosition
        ' ---------------------------------------------------- get the screen containing the form
        Dim s As Screen = Screen.FromPoint(Form1_Location)
        ' ---------------------------------------------------- normalize X position to the screen
        Dim w As Int32 = s.Bounds.Width - 1
        Dim x As Double = (mousePoint.X - s.Bounds.X) / w
        ' ---------------------------------------------------- limit X in the range 0 to 1
        If x < 0 Then x = 0
        If x > 1 Then x = 1
        ' ----------------------------------------------------
        Return x
    End Function

    Friend Function GetNormalizedCursorY() As Double
        ' ---------------------------------------------------- get the cursor position in screen coordimates
        Dim mousePoint As Point = Control.MousePosition
        ' ---------------------------------------------------- get the screen containing the form
        Dim s As Screen = Screen.FromPoint(Form1_Location)
        ' ---------------------------------------------------- normalize Y position to the screen
        Dim h As Int32 = s.Bounds.Height - 1
        Dim y As Double = (mousePoint.Y - s.Bounds.Y) / h
        ' ---------------------------------------------------- limit Y in the range 0 to 1
        If y < 0 Then y = 0
        If y > 1 Then y = 1
        ' ----------------------------------------------------
        Return 1 - y
    End Function


    ' ======================================================================================================
    '  MOUSE POSITION IN PIXEL COORDINATES VALID FOR MULTISCREENS TOO
    ' ------------------------------------------------------------------------------------------------------
    '  The 0.0 position is the lower left corner
    ' ======================================================================================================
    Friend Function GetCursorX() As Double
        ' ---------------------------------------------------- get the cursor position in screen coordimates
        Dim mousePoint As Point = Control.MousePosition
        ' ----------------------------------------------------
        Return mousePoint.X
    End Function

    Friend Function GetCursorY() As Double
        ' ---------------------------------------------------- get the cursor position in screen coordimates
        Dim mousePoint As Point = Control.MousePosition
        ' ---------------------------------------------------- get the screen containing the cursor
        Dim s As Screen = Screen.FromPoint(mousePoint)
        ' ---------------------------------------------------- reverse values to have the zero down
        Return s.Bounds.Height - mousePoint.Y - 1
    End Function

    ' ======================================================================================================
    '  Color(double containing an integer color) to HSL(single containing integers)
    ' ======================================================================================================
    Friend Structure HSL
        Dim H As Single
        Dim S As Single
        Dim L As Single
    End Structure

    Friend Function RGBToHSL(ByVal c As Double) As HSL
        '
        If c < 0 Or c > &HFFFFFF Then Return New HSL
        '
        Dim r As Single = (CInt(c) And &HFF0000) \ &H10000
        Dim g As Single = (CInt(c) And &HFF00) \ &H100
        Dim b As Single = CInt(c) And &HFF
        '
        r /= 256
        g /= 256
        b /= 256
        '
        Dim min As Single = Math.Min(Math.Min(r, g), b)
        Dim max As Single = Math.Max(Math.Max(r, g), b)
        Dim delta As Single = max - min
        '
        With RGBToHSL
            .L = (max + min) / 2
            '
            If delta < 0.00001 Then
                .H = 0
                .S = 0
            Else
                .S = If(.L <= 0.5, delta / (max + min), delta / (2 - max - min))
                '
                If r = max Then
                    .H = ((g - b) / 6) / delta
                ElseIf g = max Then
                    .H = (1.0F / 3) + ((b - r) / 6) / delta
                Else
                    .H = (2.0F / 3) + ((r - g) / 6) / delta
                End If
                '
                If .H < 0 Then .H += 1
                If .H > 1 Then .H -= 1
                '
                .H = .H * 360.0F
            End If
            .H = Int(.H)
            .S = Int(.S * 255)
            .L = Int(.L * 255)
        End With
    End Function

    ' ================================================================================
    '  HUE to ColorIndex (0 to 8) and ColorName
    ' ================================================================================
    Friend Function HslToColorIndex(ByVal hsl1 As HSL) As Int32
        If hsl1.L < 50 Then
            Return 0
        ElseIf hsl1.L > 200 Then
            Return 8
        ElseIf hsl1.S < 50 Then
            If hsl1.L < 128 Then
                Return 0
            Else
                Return 8
            End If
        Else
            Select Case hsl1.H
                Case Is < 25 : Return 1
                Case Is < 50 : Return 2
                Case Is < 70 : Return 3
                Case Is < 160 : Return 4
                Case Is < 210 : Return 5
                Case Is < 270 : Return 6
                Case Is < 320 : Return 7
                Case Else : Return 1
            End Select
        End If
    End Function

    Friend Function HslToColorName(ByVal hsl1 As HSL) As String
        If hsl1.L < 50 Then
            Return "Black"
        ElseIf hsl1.L > 200 Then
            Return "White"
        ElseIf hsl1.S < 50 Then
            If hsl1.L < 128 Then
                Return "Black"
            Else
                Return "White"
            End If
        Else
            Select Case hsl1.H
                Case Is < 25 : Return "Red"
                Case Is < 50 : Return "Orange"
                Case Is < 70 : Return "Yellow"
                Case Is < 160 : Return "Green"
                Case Is < 210 : Return "Cyan"
                Case Is < 270 : Return "Blue"
                Case Is < 320 : Return "Magenta"
                Case Else : Return "Red"
            End Select
        End If
    End Function

    ' ================================================================================
    '  RGB-COLOR TO NEAREST COLOR NAME
    ' ================================================================================
    Private Structure TStoredColours
        Dim Name As String
        Dim R As Short
        Dim G As Short
        Dim B As Short
    End Structure
    Private SysColorsList As List(Of TStoredColours) = New List(Of TStoredColours)

    Private Sub GenerateSystemColorList()
        ' -----------------------------------------------------------------------------
        Dim ColourType As Type = GetType(System.Drawing.Color)
        Dim PropertyInfo As System.Reflection.PropertyInfo()
        PropertyInfo = ColourType.GetProperties(System.Reflection.BindingFlags.[Static] Or _
                                                System.Reflection.BindingFlags.DeclaredOnly Or _
                                                System.Reflection.BindingFlags.[Public])
        For Each c As System.Reflection.PropertyInfo In PropertyInfo
            If Not c.Name = "Transparent" Then
                AddColorToList(c.Name, SysColorsList)
            End If
        Next
        Dim SortOrder As System.Collections.Generic.IEnumerable(Of TStoredColours)
        SortOrder = From c In SysColorsList Order By c.Name Select c
        SysColorsList = SortOrder.ToList
    End Sub

    Private Sub AddColorToList(ByVal ColorName As String, ByVal list As List(Of TStoredColours))
        Dim c As Color = Color.FromName(ColorName)
        list.Add(New TStoredColours With {.Name = ColorName, .R = c.R, .G = c.G, .B = c.B})
    End Sub

    Friend Function Utils_GetNearestColorName(ByVal col As Color) As String
        If SysColorsList.Count = 0 Then GenerateSystemColorList()
        Dim NearestColour As TStoredColours
        NearestColour = New TStoredColours With {.Name = "NULL", .R = 255, .G = 255, .B = 255}
        Dim Diff As Integer
        Dim MinDiff As Integer = Integer.MaxValue
        For Each c As TStoredColours In SysColorsList
            If col = Color.FromArgb(255, c.R, c.G, c.B) Then
                Return c.Name
            Else
                Diff = Math.Abs(col.R - c.R) + Math.Abs(col.G - c.G) + Math.Abs(col.B - c.B)
                If Diff < MinDiff Then
                    MinDiff = Diff
                    NearestColour = c
                End If
            End If
        Next
        Return NearestColour.Name
    End Function


    ' ======================================================================================================
    '  COLORS
    ' ======================================================================================================
    Private Function StringStartingWithLetter(ByVal str As String) As Boolean
        Return Char.IsLetter(str(0))
    End Function

    Friend Function GrayScaleLuminance(ByVal c As Color) As Byte
        Dim lum As Double = 0.299 * c.R + 0.587 * c.G + 0.114 * c.B
        Return CByte(lum)
    End Function

    Friend Function ModifiedColor(ByVal c As Color, ByVal delta As Int32) As Color
        Dim r As Int32 = c.R + delta
        Dim g As Int32 = c.G + delta
        Dim b As Int32 = c.B + delta
        If r < 0 Then r = 0
        If g < 0 Then g = 0
        If b < 0 Then b = 0
        If r > 255 Then r = 255
        If g > 255 Then g = 255
        If b > 255 Then b = 255
        Return Color.FromArgb(r, g, b)
    End Function

    Friend Function GetColorNamesAsArray() As String()
        Dim sa() As String = GetColorNames().Split(" "c)
        Return sa
    End Function

    Friend Function GetColorNames() As String
        GetColorNames = "Black MidnightBlue Navy DarkBlue MediumBlue Blue DarkSlateGray DarkSlateBlue " + _
               "Indigo Maroon DarkRed Purple DarkMagenta BlueViolet DarkOrchid DarkViolet " + _
               "MediumOrchid MediumPurple MediumSlateBlue SlateBlue RoyalBlue DodgerBlue " + _
               "DeepSkyBlue CornflowerBlue SteelBlue CadetBlue LightSeaGreen DarkCyan Teal " + _
               "DarkGreen Green ForestGreen SeaGreen DarkOliveGreen Olive OliveDrab LimeGreen " + _
               "MediumSeaGreen Sienna SaddleBrown Brown Firebrick Crimson Red MediumVioletRed " + _
               "DeepPink Fuchsia Magenta IndianRed OrangeRed Tomato Chocolate DarkGoldenrod " + _
               "DimGray Gray SlateGray LightSlateGray DarkGray Silver LightSteelBlue LightBlue " + _
               "DarkSeaGreen MediumAquamarine DarkTurquoise MediumTurquoise Peru Goldenrod DarkKhaki " + _
               "RosyBrown PaleVioletRed Orchid Violet Plum HotPink LightCoral Salmon DarkSalmon Coral " + _
               "DarkOrange Orange SandyBrown LightSalmon BurlyWood Tan Thistle LightPink Pink " + _
               "Gold Yellow Lime SpringGreen MediumSpringGreen YellowGreen Chartreuse LawnGreen " + _
               "GreenYellow PaleGreen LightGreen LightSkyBlue SkyBlue Turquoise Aqua Cyan " + _
               "Aquamarine PaleTurquoise PowderBlue LightCyan LightGray Gainsboro Lavender Khaki " + _
               "PaleGoldenrod PeachPuff NavajoWhite Wheat Bisque Moccasin MistyRose BlanchedAlmond " + _
               "PapayaWhip AntiqueWhite LavenderBlush SeaShell Linen OldLace Beige " + _
               "Cornsilk LightYellow LightGoldenrodYellow LemonChiffon FloralWhite Ivory Snow " + _
               "Azure AliceBlue Honeydew MintCream WhiteSmoke GhostWhite White"
        GetColorNames += " " + GetColorNames2()
    End Function

    'Friend Function GetColorNames3() As String
    '    Dim s As String = ""
    '    Dim i As Int32
    '    For i = 1 To 9
    '        s += "Red" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Green" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Blue" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Cyan" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Magenta" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Yellow" + i.ToString + " "
    '    Next
    '    For i = 1 To 9
    '        s += "Gray" + i.ToString + " "
    '    Next
    '    Return s.Trim
    'End Function

    Friend Function GetColorNames2() As String
        Dim s As String = ""
        Dim i As Int32
        For i = 1 To 9
            s += "Red" + i.ToString + " "
            s += "Green" + i.ToString + " "
            s += "Blue" + i.ToString + " "
            s += "Cyan" + i.ToString + " "
            s += "Magenta" + i.ToString + " "
            s += "Yellow" + i.ToString + " "
            s += "Gray" + i.ToString + " "
        Next
        Return s.Trim
    End Function


    Friend Function ColorFromColorName(ByVal str As String) As Color
        ' -----------------------------------------------
        Dim n As Double = Val(Right(str, 1))
        If n <> 0 Then
            n = n / 9
            n = n ^ 0.3
            n *= 255
            Dim r, g, b As Double
            Select Case Left(str, str.Length - 1).ToLower
                Case "red" : r = n
                Case "green" : g = n
                Case "blue" : b = n
                Case "magenta" : r = n : b = n
                Case "yellow" : r = n : g = n
                Case "cyan" : g = n : b = n
                Case "gray" : r = n : g = n : b = n
            End Select
            Return Color.FromArgb(CInt(r), CInt(g), CInt(b))
        End If
        ' -----------------------------------------------
        Dim c As Color = Color.FromName(str)
        If c.A <> 0 Then
            Return c
        End If
        Return Color.Empty
    End Function

    ' ======================================================================================================
    '  OPEN HELP FILE
    ' ======================================================================================================
    Friend Sub OpenHelpFile()
        ' ------------------------------------------------------------------------------------------------
        If TryToOpenFile(HelpFile_PathAndName) Then Return
        If TryToOpenHelp(Application.StartupPath + "\Docs") Then Return
        ' ------------------------------------------------------------------------------------------------
        Form_Input.Text = "Message from Automation"
        Form_Input.Prompt = "Help file not found" + vbCrLf + vbCrLf
        Form_Input.Prompt += "Help files are searched in the ""Docs"" folder, in the following languages order:" + vbCrLf
        Form_Input.Prompt += "ENG, FRA, SPA, ITA, POR, JPN, CHI" + vbCrLf + vbCrLf
        Form_Input.Prompt += "Press the 'OPEN DOCS FOLDER' button and copy your preferred Help file." + vbCrLf + vbCrLf
        Form_Input.Prompt += "Otherwise press the 'SELECT' button and select your Help file location."
        Form_Input.txt_Input.Visible = False
        Form_Input.chk_Cancel.Text = "OPEN DOCS FOLDER"
        Form_Input.chk_Ok.Text = "SELECT"
        Dim OldFormInputLocation As Point = Form_Input.Location
        Dim OldFormInputSize As Size = Form_Input.Size
        Dim OldCancelButtonSize As Size = Form_Input.chk_Cancel.Size
        Form_Input.Left = Form1.Left + 120
        Form_Input.Top = Form1.Top + 110
        Form_Input.Width = 580
        Form_Input.Height = 270
        Form_Input.chk_Cancel.Size = New Size(180, Form_Input.chk_Cancel.Height)
        Form_Input.ShowDialog(Form1)
        If Form_Input.ValidInput Then
            LocateHelpFile()
        Else
            Dim s As String = Application.StartupPath + "\Docs"
            My.Computer.FileSystem.CreateDirectory(s)
            Process.Start(s)
        End If
        Form_Input.Location = OldFormInputLocation
        Form_Input.Size = OldFormInputSize
        Form_Input.chk_Cancel.Size = OldCancelButtonSize
    End Sub

    Private Function TryToOpenHelp(ByVal path As String) As Boolean
        path += "\Theremino_Automation_Help"
        If TryToOpenFile(path + "_ENG.pdf") Then Return True
        If TryToOpenFile(path + "_ESP.pdf") Then Return True
        If TryToOpenFile(path + "_FRA.pdf") Then Return True
        If TryToOpenFile(path + "_ITA.pdf") Then Return True
        If TryToOpenFile(path + "_POR.pdf") Then Return True
        If TryToOpenFile(path + "_GER.pdf") Then Return True
        If TryToOpenFile(path + "_CHI.pdf") Then Return True
        If TryToOpenFile(path + "_JPN.pdf") Then Return True
        Return False
    End Function
    Private Function TryToOpenFile(ByVal file As String) As Boolean
        If My.Computer.FileSystem.FileExists(file) Then
            Process.Start(file)
            Return True
        Else
            Return False
        End If
    End Function

    Friend Sub LocateHelpFile()
        Dim ofd As OpenFileDialog = New OpenFileDialog()
        'ofd.InitialDirectory = PrepareSaveLoadFolder()
        ' ---------------------------------- left part of filename not visible
        ofd.FileName = IO.Path.GetFileName(HelpFile_PathAndName)
        ofd.DefaultExt = ".pdf"
        ' ------------------------ filename completely visible but no extension
        'ofd.FileName = IO.Path.GetFileNameWithoutExtension(FilePathAndName)
        'ofd.DefaultExt = IO.Path.GetExtension(FilePathAndName)
        ' ------------------------ filename completely visible and better dialog
        ofd.ShowHelp = True
        ' ------------------------------------------------------------
        'ofd.RestoreDirectory = True
        ofd.AutoUpgradeEnabled = True
        ofd.Title = "Select help file"
        ofd.ValidateNames = False
        ofd.Multiselect = True
        ofd.Filter = "All files (*.*)|*.*"
        If ofd.ShowDialog() = Windows.Forms.DialogResult.OK Then
            HelpFile_PathAndName = ofd.FileName
            TryToOpenFile(HelpFile_PathAndName)
        End If
    End Sub

    ' ==============================================================================================================
    '   ToolTips
    ' ==============================================================================================================
    Private WithEvents ttp1 As ToolTip = New ToolTip
    Friend Sub ToolTips_Init()
        ttp1.ShowAlways = True
        ttp1.UseAnimation = True
        ttp1.UseFading = True
        ttp1.IsBalloon = True
        ttp1.AutoPopDelay = 32700
        ttp1.InitialDelay = 500
        ttp1.ReshowDelay = 500
        ' --------------------------------------------------------------- Add here Controls and ToolTip-Text
        ttp1.SetToolTip(Form1.chk_ControlsBack, "History back")
        ttp1.SetToolTip(Form1.chk_ControlsForward, "History forward")
        ttp1.SetToolTip(Form1.chk_ControlsGoToNext, "Find next occurrence")
        ttp1.SetToolTip(Form1.chk_ControlsGoToDeclaration, "Goto declaration")

        ttp1.SetToolTip(Form1.chk_ControlsLoad, "Open a dialog to load a program." + vbCrLf + "Press the right button to reload this program and test it.")
        ttp1.SetToolTip(Form1.chk_ControlsSave, "Open a dialog to save the program with another name." + vbCrLf + "Press the right button to save the program with a single click.")

        ttp1.SetToolTip(Form1.chk_Comment, "Comment the selected text")
        ttp1.SetToolTip(Form1.chk_Uncomment, "Uncomment the selected text")
        ttp1.SetToolTip(Form1.chk_Undo, "Undo the last changes")
        ttp1.SetToolTip(Form1.chk_Redo, "Redo the last changes")
        ttp1.SetToolTip(Form1.chk_HistoryBack, "History back")
        ttp1.SetToolTip(Form1.chk_HistoryForward, "History forward")
        ttp1.SetToolTip(Form1.chk_GoToNextOccurrence, "Find next occurrence")
        ttp1.SetToolTip(Form1.chk_GotoDeclaration, "Goto declaration")
        ttp1.SetToolTip(Form1.chk_Debug, "Open the debug window")
        ttp1.SetToolTip(Form1.chk_Help, "Open the HELP document." + vbCrLf + "Press the right mouse button to select the Help file")

        'ToolStripButton_Validate.ToolTipText = Msg_Validate1
        'ToolStripButton_BeepOnErrors.ToolTipText = Msg_BeepOnErrors1
        'ToolStripButton_Lock.ToolTipText = Msg_Lock1 & vbCrLf & Msg_Lock2
        'ToolStripButton_Disconnect.ToolTipText = Msg_Disconnect1
    End Sub

    ' --------------------- this corrects the XP bug that causes tooltip disappearing for ever 
    'Private Sub controls_MouseEnter(ByVal sender As Object, _
    '                               ByVal e As System.EventArgs) Handles txt_CommSpeed.MouseEnter, _
    '                                                                    txt_MinChange.MouseEnter, _
    '                                                                    txt_CapArea.MouseEnter
    '    ttp1.Active = False
    '    ttp1.Active = True
    'End Sub


    Friend Function StringToInt32(ByVal s As String) As Int32
        Dim n As Double = Val(s)
        If n < Int32.MinValue Or n > Int32.MaxValue Then Return 0
        Return CInt(n)
    End Function

    ' =============================================================================================
    '  This "BrowseForFile" function is a substitute for the "OpenFileDialog" 
    ' ---------------------------------------------------------------------------------------------
    '  OpenFileDialog generates an error when called from the THREAD of the RUNNING PROGRAM 
    '  A dangerous solution was to set the THREAD in STA MODE
    '  But using the GetOpenFileName API the STA MODE is not required
    ' ============================================================================================= 
    Private Structure OPENFILENAME
        Public lStructSize As Integer
        Public hwndOwner As Integer
        Public hInstance As Integer
        Public lpstrFilter As String
        Public lpstrCustomFilter As String
        Public nMaxCustFilter As Integer
        Public nFilterIndex As Integer
        Public lpstrFile As String
        Public nMaxFile As Integer
        Public lpstrFileTitle As String
        Public nMaxFileTitle As Integer
        Public lpstrInitialDir As String
        Public lpstrTitle As String
        Public flags As Integer
        Public nFileOffset As Short
        Public nFileExtension As Short
        Public lpstrDefExt As String
        Public lCustData As Integer
        Public lpfnHook As Integer
        Public lpTemplateName As String
    End Structure

    Private Declare Function GetOpenFileName Lib "comdlg32.dll" Alias "GetOpenFileNameA" (ByRef pOpenfilename As OPENFILENAME) As Integer

    Public Function BrowseForFile(ByVal InitialDir As String, _
                                  Optional ByVal Title As String = "Select File", _
                                  Optional ByVal Filters As String = "", _
                                  Optional ByVal ParentHwnd As Integer = Nothing) As String
        '
        Const OFN_ALLOWMULTISELECT As Integer = &H200
        Const OFN_EXPLORER As Integer = &H80000
        'Const OFN_SHOWHELP As Int32 = &H10
        Const MAXLEN As Integer = 254
        Dim OFN As OPENFILENAME = New OPENFILENAME
        OFN.lStructSize = Runtime.InteropServices.Marshal.SizeOf(OFN)
        OFN.hwndOwner = ParentHwnd
        ' -------------------------------------------- Replace friendly delimitators with nulls
        Filters = Replace(Filters, "|", vbNullChar)
        Filters = Replace(Filters, ";", vbNullChar)
        If Right$(Filters, 1) <> vbNullChar Then
            Filters = Filters & vbNullChar
        End If
        OFN.lpstrFilter = Filters & "All Files (*.*)" & vbNullChar & "*.*" & vbNullChar
        ' ---------------------------------------------
        OFN.lpstrFile = StrDup(MAXLEN - 1, " ")
        OFN.nMaxFile = MAXLEN
        OFN.lpstrFileTitle = StrDup(MAXLEN - 1, " ")
        OFN.nMaxFileTitle = MAXLEN
        ' --------------------------------------------- INITIAL DIR
        If InitialDir.EndsWith("\") Then InitialDir = InitialDir.Substring(0, InitialDir.Length - 1)
        OFN.lpstrInitialDir = InitialDir
        OFN.lpstrFile = InitialDir + "\." ' <<--- this ensures to open the correct dir
        ' ---------------------------------------------
        OFN.lpstrTitle = Title
        OFN.flags = OFN_ALLOWMULTISELECT Or OFN_EXPLORER 'Or OFN_SHOWHELP
        ' ---------------------------------------------
        If GetOpenFileName(OFN) = 1 Then
            BrowseForFile = OFN.lpstrFile
        Else
            BrowseForFile = ""
        End If
        ' --------------------------------------------- restore the default current directory
        SetAppPathAsDefaultCurrentDir()
    End Function

    ' =============================================================================================
    '  EscapeSpecialCharacters for SendKeys
    ' ============================================================================================= 
    Friend Function EscapeSpecialCharacters(ByVal txt As String) As String
        txt = txt.Replace("+", "{+}")
        txt = txt.Replace("^", "{^}")
        txt = txt.Replace("%", "{%}")
        txt = txt.Replace("~", "{~}")
        txt = txt.Replace("(", "{(}")
        txt = txt.Replace(")", "{)}")
        Return txt
    End Function


    ' =============================================================================================
    '  Set Application.Path As Default Current Dir
    ' =============================================================================================
    Friend Sub SetAppPathAsDefaultCurrentDir()
        FileIO.FileSystem.CurrentDirectory = Application.StartupPath
    End Sub

    ' =============================================================================================
    '  InitProgramIdentifiers
    ' =============================================================================================
    Friend Sub InitProgramIdentifiers()
        ' ----------------------------------------------- read Identifiers file
        Dim FileLines(-1) As String
        Dim fname As String = Application.StartupPath + "\Files\Identifiers.txt"
        If My.Computer.FileSystem.FileExists(fname) Then
            FileLines = IO.File.ReadAllLines(fname)
        End If
        ' ----------------------------------------------- init identifiers
        For j As Int32 = 0 To FileLines.Length - 1
            Dim sa() As String = FileLines(j).Trim.Split("="c)
            If sa.Length = 2 Then
                Select Case sa(0).ToLower
                    Case "programguid" : ProgGUID = sa(1).Trim
                    Case "programid1" : ProgID1 = ExtractID(sa(1).Trim)
                    Case "programid2" : ProgID2 = ExtractID(sa(1).Trim)
                    Case "programid3" : ProgID3 = ExtractID(sa(1).Trim)
                    Case "sew" : SewID = sa(1)
                End Select
            End If
        Next
        ' ----------------------------------------------- init function enablers
        SewFunctionsEnabled = EnableTest(SewID)
    End Sub

    Private ProgGUID As String = ""
    Private ProgID1 As String = ""
    Private ProgID2 As String = ""
    Private ProgID3 As String = ""
    Private SewID As String = ""

    Private Function EnableTest(ByVal p As String) As Boolean
        Dim ID As String = ""
        If ProgGUID = "741f40ef-4c13-4a53-8f0e-e6dd0f33654a" Then
            ID = ProgID1 + ProgID2 + ProgID3
        End If
        Return ID.Contains(PassCodify(p.Trim))
    End Function

    Private Function PassCodify(ByVal p As String) As String
        Dim s As String = ""
        For i As Int32 = 0 To p.Length - 1
            Select Case p(i)
                Case "Z"c : s += "A"
                Case "z"c : s += "a"
                Case Else : s += ChrW(AscW(p(i)) + 1)
            End Select
        Next
        Return s
    End Function

    Private Function ExtractID(ByVal id As String) As String
        Dim s As String = ""
        For i As Int32 = 0 To id.Length - 1
            s = id(i) + s
            i += 2
        Next
        Return s
    End Function

    
    ' =============================================================================================
    '   FUNCTION PARAMS - LABEL WITH PARAMS - GOSUB WITH PARAMS 
    '   This function removes params from str (passed by ref)
    '   and returns an array with the single parameters
    ' =============================================================================================
    Friend Function ExtractFunctionParams(ByRef str As String) As String()
        If str.Contains("(") And str.EndsWith(")") Then
            If FunctionParamsAreValid(str) Then
                Return ExtractFunctionParams_Core(str)
            End If
        End If
        ReDim ExtractFunctionParams(-1)
    End Function

    Private Function ExtractFunctionParams_Core(ByRef str As String) As String()
        Dim n As Int32 = str.IndexOf("("c)
        Dim params As String = str.Substring(n + 1)
        str = str.Remove(n).Trim
        'If params.EndsWith(")") Then

        ' --------------------------------------------- remove the last ")"
        params = params.Substring(0, params.Length - 1)
        ' -------------------------------------------- compact separators with spaces or tabs to ")("
        params = params.Replace(vbTab, " ")
        While params.Contains(" (")
            params = params.Replace(" (", "(")
        End While
        ' -------------------------------------------- split param array at separators
        Dim sa() As String
        sa = Strings.Split(params, ")(")
        ' -------------------------------------------- trim single params
        For i As Int32 = 0 To sa.Length - 1
            sa(i) = sa(i).Trim
        Next
        Return sa
    End Function

    Friend Sub AssignParamsToFunctionVars(ByVal params() As String, ByVal nline As Int32)
        Dim l As String
        Dim nparam As Int32 = 0
        For i As Int32 = nline + 1 To RTB_Lines.Length - 1
            If nparam > params.Length - 1 Then Exit For
            l = RTB_Lines(i)
            RemoveComments(l)
            l = ReplaceMultipleSpacesAndTabAndTrim(l).ToLower
            If l <> "" Then
                Dim sa() As String = l.Split(" "c)
                If sa(0) = "variable" Then
                    If params(nparam) <> "Null" Then
                        Select Case sa(1).ToLower
                            Case "numeric"
                                If IsNumericUserVariable(sa(2)) Then SetNumericUserVariable(sa(2), EvalDouble(params(nparam)))
                            Case "string"
                                If IsStringUserVariable(sa(2)) Then SetStringUserVariable(sa(2), EvalString_AutoQuoting(params(nparam)))
                        End Select
                    End If
                    nparam += 1
                Else
                    Exit For
                End If
            End If
        Next
    End Sub

    Friend Function GetFunctionNameWithoutParams(ByRef par As String) As String
        If par.Contains("(") Then
            Dim i As Int32 = par.IndexOf("("c)
            par = par.Remove(i).Trim
        End If
        Return par
    End Function

    Friend Function FunctionParamsAreValid(ByVal params As String) As Boolean
        Dim n As Int32 = 0
        For i As Int32 = 0 To params.Length - 1
            If params(i) = "(" Then n += 1
            If params(i) = ")" Then n -= 1
        Next
        If n <> 0 Then Return False
        If params.Contains("(") And params.EndsWith(")") Then
            Dim p As String() = ExtractFunctionParams_Core(params)
            For i As Int32 = 0 To p.Length - 1
                If Not IsValidBooleanStringOrDouble(p(i)) Then Return False
            Next
        End If
        Return True
    End Function

    ' =============================================================================================
    '   GET FILES RECURSIVE
    ' =============================================================================================
    Private RecursiveLevel As Int32
    Friend Function GetFilesRecursive(ByVal folder As String, ByVal wildcard As String) As String()
        Dim files As New List(Of String)
        Dim subfolders As String()
        On Error Resume Next
        subfolders = System.IO.Directory.GetDirectories(folder)
        On Error GoTo 0
        If subfolders IsNot Nothing Then
            For i As Int32 = 0 To subfolders.Length - 1
                Dim subfolder As String = subfolders(i)
                RecursiveLevel += 1
                files.AddRange(GetFilesRecursive(subfolder, wildcard))
            Next
            On Error Resume Next
            files.AddRange(System.IO.Directory.GetFiles(folder, wildcard))
            On Error GoTo 0
        End If
        RecursiveLevel -= 1
        Return files.ToArray
    End Function

End Module
