﻿Imports System.Text
Imports System.IO

Module Module_WavFileCreator

    Friend Sub WriteWaveFile(ByVal filePath As String, _
                             ByVal audioData As Byte(), _
                             ByVal sampleRate As Integer, _
                             ByVal bitDepth As Integer, _
                             ByVal numChannels As Integer)
        ' -----------------------------------------------------------------
        Directory.CreateDirectory(Path.GetDirectoryName(filePath))
        ' -----------------------------------------------------------------
        Using stream As New FileStream(filePath, FileMode.Create)
            WriteWaveStream(stream, audioData, sampleRate, bitDepth, numChannels)
        End Using
    End Sub

    Friend WaveBytes() As Byte

    Friend Sub WriteWaveArray(ByVal audioData As Byte(), _
                              ByVal sampleRate As Integer, _
                              ByVal bitDepth As Integer, _
                              ByVal numChannels As Integer)
        '
        Using stream As New MemoryStream
            WriteWaveStream(stream, audioData, sampleRate, bitDepth, numChannels)
            WaveBytes = stream.ToArray()
        End Using
    End Sub

    ' about 1 ms for 10 seconds of audio
    Friend Sub WriteWaveStream(ByRef stream As IO.Stream, _
                               ByVal audioData As Byte(), _
                               ByVal sampleRate As Integer, _
                               ByVal bitDepth As Integer, _
                               ByVal numChannels As Integer)
        '
        ' --------------------------------------------------------- Calculate values based on sample rate, bit depth, and number of channels
        Dim byteRate As Integer = sampleRate * bitDepth \ 8 * numChannels
        Dim blockAlign As Integer = bitDepth \ 8 * numChannels
        '
        ' --------------------------------------------------------- Write WAV file header
        Using writer As New IO.BinaryWriter(stream)
            '
            writer.Write(Encoding.ASCII.GetBytes("RIFF"))   ' Chunk ID
            writer.Write(0)                                 ' Chunk Size (placeholder)
            writer.Write(Encoding.ASCII.GetBytes("WAVE"))   ' Format
            writer.Write(Encoding.ASCII.GetBytes("fmt "))   ' Subchunk 1 ID
            writer.Write(16)                                ' Subchunk 1 Size
            writer.Write(CShort(1))                         ' Audio Format - PCM format
            writer.Write(CShort(numChannels))               ' Number of Channels
            writer.Write(sampleRate)                        ' Sample Rate
            writer.Write(byteRate)                          ' Byte Rate
            writer.Write(CShort(blockAlign))                ' Block Align
            writer.Write(CShort(bitDepth))                  ' Bits Per Sample
            writer.Write(Encoding.ASCII.GetBytes("data"))   ' Subchunk 2 ID
            writer.Write(0)                                 ' Subchunk 2 Size (placeholder)
            '
            ' ---------------------------------------------- Write audio data (must be before the chunk updates)
            writer.Write(audioData)
            ' ---------------------------------------------- Update Chunk Size field in WAV file header
            writer.Seek(4, SeekOrigin.Begin)
            writer.Write(CInt(stream.Length - 8))
            ' ---------------------------------------------- Update Subchunk 2 Size field in WAV file header
            writer.Seek(40, SeekOrigin.Begin)
            writer.Write(CInt(stream.Length - 44))
        End Using
        '
    End Sub

End Module
