﻿Module FastTimer

    ' ===========================================================================================================
    '   FAST TIMER
    ' ===========================================================================================================
    Private FastTimerThread As Threading.Thread
    Friend StopThreadFlag As Boolean

    Friend Sub FastTimer_Start()
        FastTimer_Stop()
        ' --------------------------------------------------------- 
        SchedulerMaxPrecision()
        ' --------------------------------------------------------- start the Thread 
        FastTimerThread = New Threading.Thread(AddressOf FastThread_Execute)
        ' --------------------------------------------------------- make the thread as background thread.
        FastTimerThread.IsBackground = True
        ' --------------------------------------------------------- set the Priority of the thread.
        FastTimerThread.Priority = Threading.ThreadPriority.AboveNormal
        ' --------------------------------------------------------- start the thread
        StopThreadFlag = False
        FastTimerThread.Start()
    End Sub

    Friend Sub FastTimer_Stop()
        If FastTimerThread Is Nothing Then Return
        If Not FastTimerThread.IsAlive Then Return
        ' --------------------------------------------- try to join the thread
        StopThreadFlag = True
        FastTimerThread.Join(10)
        ' --------------------------------------------- catch because sometimes the thread is already closed
        Try
            While FastTimerThread.IsAlive
                Threading.Thread.Sleep(10)
                Application.DoEvents()
            End While
            ' ----------------------------------------- eventually abort
            If FastTimerThread.IsAlive Then
                MessageBox.Show("ExecutionThread not closed correctly.", _
                                "Message from Automation", _
                                MessageBoxButtons.OK, MessageBoxIcon.Warning)
                FastTimerThread.Abort()
            End If
        Catch
            ' ----------------------------------------- try catch needed to be sure to close the app
            Try
                FastTimerThread.Abort()
            Catch
            End Try
        End Try
        ' ---------------------------------------------
        FastTimerThread = Nothing
        ' ---------------------------------------------
        SchedulerDefaultPrecision()
    End Sub

    Private Sub FastThread_Execute()
        Do
            ReadSlots()
            Threading.Thread.Sleep(1)
        Loop Until StopThreadFlag
    End Sub

    ' =========================================================================
    '  WINDOWS SCHEDULER PRECISION
    ' ========================================================================= 
    Private Declare Function timeBeginPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Declare Function timeEndPeriod Lib "winmm.dll" (ByVal uPeriod As Int32) As Int32
    Private Sub SchedulerMaxPrecision()
        If OperatingSystemIsWindows Then
            timeBeginPeriod(1)
        End If
    End Sub
    Private Sub SchedulerDefaultPrecision()
        If OperatingSystemIsWindows Then
            timeEndPeriod(1)
        End If
    End Sub

End Module
