﻿Imports System.Drawing.Imaging


Module Module_SaveLoad

    Friend EventsAreEnabled As Boolean = False

    Friend GCI As Globalization.CultureInfo = Globalization.CultureInfo.InvariantCulture

    ' =======================================================================================================
    '   APP TITLE AND VERSION (ALL FIELDS)
    ' =======================================================================================================
    Friend Function AppTitleAndVersion(Optional ByVal RemoveThereminoPrefix As Boolean = False) As String
        Dim appname As String = IO.Path.GetFileNameWithoutExtension(Application.ExecutablePath)
        Dim Title As String = Replace(appname, "_", " ")
        If RemoveThereminoPrefix Then Title = Replace(Title, "Theremino ", "")
        Return Title + " - V" + GetVersionString()
    End Function

    ' =======================================================================================================
    '   GET VERSION (ALSO TRIPLE OR QUADRUPLE FIELDS)
    ' =======================================================================================================
    Friend Function GetVersionString() As String
        Dim s() As String = Split(My.Application.Info.Version.ToString, ".")
        GetVersionString = s(0) + "." + s(1)
        If s(2) <> "0" Then
            GetVersionString += "." + s(2)
            If s(3) <> "0" Then
                GetVersionString += "." + s(3)
            End If
        End If
    End Function


    ' =======================================================================================
    '  FORM FUNCTIONS
    ' =======================================================================================
    Friend Sub LimitFormPosition(ByVal f As System.Windows.Forms.Form)
        If f.WindowState <> FormWindowState.Normal Then Return
        GetMaxScreenBounds()
        EnsureFormVisible(f)
        'EnsureFormCompletelyVisible(f)
    End Sub

    Private SB As Rectangle = New Rectangle(Integer.MaxValue, Integer.MaxValue, Integer.MinValue, Integer.MinValue)

    Private Sub GetMaxScreenBounds()
        For Each s As Screen In System.Windows.Forms.Screen.AllScreens
            SB = Rectangle.Union(SB, s.WorkingArea)
        Next
    End Sub

    'Private Sub EnsureFormCompletelyVisible(ByVal frm As Form)
    '    With frm
    '        .Width = Math.Min(.Width, SB.Width)         ' not more than a maximized window
    '        .Height = Math.Min(.Height, SB.Height)      ' not more than a maximized window
    '        .Width = Math.Max(.Width, 32)               ' at least 32x24
    '        .Height = Math.Max(.Height, 24)             ' at least 32x24
    '        .Left = Math.Min(.Left, SB.Right - .Width)  ' not beyond the right border
    '        .Top = Math.Min(.Top, SB.Bottom - .Height)  ' not beyond the bottom border
    '        .Left = Math.Max(.Left, SB.Left)            ' at least at the left border
    '        .Top = Math.Max(.Top, SB.Top)               ' at least at the top border
    '    End With
    'End Sub

    Private Sub EnsureFormVisible(ByVal frm As Form)
        With frm
            .Width = Math.Min(.Width, SB.Width)             ' not more than VIRTUALSCREEN dimensions
            .Height = Math.Min(.Height, SB.Height)          ' not more than VIRTUALSCREEN dimensions 
            .Width = Math.Max(.Width, 32)                   ' at least 32x24
            .Height = Math.Max(.Height, 24)                 ' at least 32x24
            .Left = Math.Min(.Left, SB.Right - 50)          ' not beyond right border - 50 pixels
            .Top = Math.Min(.Top, SB.Bottom - 100)          ' not beyond bottom border - 50 pixels
            .Left = Math.Max(.Left, SB.Left + 100 - .Width) ' at least at left border + 50 pixels
            .Top = Math.Max(.Top, SB.Top - 10)              ' at least at top border
        End With
    End Sub

    ' (The value of the RestoreBounds property is valid only 
    '   when the WindowState property of the Form class is not equal to Normal)
    Friend Function GetFormRectangle(ByVal frm As Form) As Rectangle
        Dim r As Rectangle
        If frm.WindowState = FormWindowState.Normal Then
            r = frm.Bounds
        Else
            r = frm.RestoreBounds
        End If
        Return r
    End Function


    ' ================================================================================================
    '  Private Read-Write functions
    ' ================================================================================================
    Private Function TabString(ByVal Name As String, _
                               Optional ByVal Value As Double = Double.NaN, _
                               Optional ByVal fmt As String = "") As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)
        If Double.IsNaN(Value) Then
            Return Name
        Else
            Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString(fmt, GCI)
        End If
    End Function
    Private Function TabString(ByVal Name As String, _
                                  ByVal Value As Boolean) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value.ToString
    End Function
    Private Function TabString(ByVal Name As String, _
                                ByVal Value As String) As String

        Dim nTab As Int32 = Math.Max(0, 22 - Name.Length)

        Return Name & "=" & Strings.StrDup(nTab, " ") & Value
    End Function
    Private Function Val_Single(ByVal l As String) As Single
        Return CSng(Val(l.Replace(",", ".")))
    End Function
    Private Function Val_Double(ByVal l As String) As Double
        Return Val(l.Replace(",", "."))
    End Function
    Private Function Val_Int(ByVal l As String) As Int32
        Return CInt(Val(l))
    End Function
    Private Function ExtractParamName(ByRef s As String) As String
        ' ------------------------- Returns the first field from begin to the first "=" symbol
        ' -------------------------  and removes it from the string
        Dim i As Integer
        i = InStr(s, "=")
        If i > 0 Then
            ExtractParamName = Trim(Strings.Left(s, i - 1))
            s = Trim(Mid(s, i + 1))
        Else
            ExtractParamName = Trim(s)
            s = ""
        End If
    End Function
    Private Function AssemblyName() As String
        Return System.Reflection.Assembly.GetExecutingAssembly.GetName.Name
    End Function


    ' ==================================================================================================
    '  SAVE LOAD -- Program INI
    ' ==================================================================================================
    Friend Sub Save_INI(Optional ByVal iniFileName As String = "")
        '
        Dim SaveAll As Boolean = False
        If iniFileName = "" Then
            iniFileName = Application.StartupPath & "\" & AssemblyName() & "_INI.txt"
            SaveAll = True
        End If
        Dim f As System.IO.StreamWriter = Nothing
        Try
            f = IO.File.CreateText(iniFileName)
            '
            If SaveAll Then
                f.WriteLine("")
                f.WriteLine("===========================================")
                f.WriteLine(" Program Params")
                f.WriteLine("===========================================")
                ' ------------------------------------------------------------------------------ FORM BOUNDS
                Dim r As Rectangle
                r = GetFormRectangle(Form1)
                f.WriteLine(TabString("Form1_Top", r.Top))
                f.WriteLine(TabString("Form1_Left", r.Left))
                f.WriteLine(TabString("Form1_Width", r.Width))
                f.WriteLine(TabString("Form1_Height", r.Height))
                f.WriteLine(TabString("Form1_WindowState", Form1.WindowState))
                '
                f.WriteLine(TabString(""))
                f.WriteLine("===========================================")
                f.WriteLine(TabString(" Vectorization and GCode"))
                f.WriteLine("===========================================")
            End If
            '
            f.WriteLine(TabString("GroupBox1_Height", Form1.GroupBox1.Height))
            f.WriteLine(TabString("GroupBox2_Height", Form1.GroupBox2.Height))
            f.WriteLine(TabString("GroupBox3_Height", Form1.GroupBox3.Height))
            '
            f.WriteLine(TabString("PixelPerMM", PixelPerMM))
            f.WriteLine(TabString("StartX", StartX))
            f.WriteLine(TabString("StartY", StartY))
            '
            f.WriteLine(TabString("WireDiam", Form1.txt_WireDiam.NumericValue))
            f.WriteLine(TabString("NeedleDiam", Form1.txt_NeedleDiam.NumericValue))
            '
            f.WriteLine(TabString("SewingAreaMaxY", Form1.txt_SewingAreaMaxY.NumericValueInteger))
            f.WriteLine(TabString("SewingAreaMinY", Form1.txt_SewingAreaMinY.NumericValueInteger))
            f.WriteLine(TabString("SewingAreaMaxX", Form1.txt_SewingAreaMaxX.NumericValueInteger))
            f.WriteLine(TabString("SewingAreaMinX", Form1.txt_SewingAreaMinX.NumericValueInteger))
            '
            f.WriteLine(TabString("ScaleText", Form1.chk_ScaleText.Checked))
            f.WriteLine(TabString("TrueXYZ", Form1.chk_TrueXYZ.Checked))
            '
            f.WriteLine(TabString("SlotX", Form1.txt_SlotX.NumericValueInteger))
            f.WriteLine(TabString("SlotY", Form1.txt_SlotY.NumericValueInteger))
            f.WriteLine(TabString("SlotZ", Form1.txt_SlotZ.NumericValueInteger))
            f.WriteLine(TabString("SlotRestart", Form1.txt_SlotRestart.NumericValueInteger))
        Catch
        End Try
        Try
            f.Close()
        Catch
        End Try
    End Sub


    Friend Sub Load_INI(Optional ByVal iniFileName As String = "")
        ' ------------------------------------------------------------- defaults
        '
        ' -------------------------------------------------------------
        ' ------------------------------------------------------------------------------- 
        ' With "Resume Next" subsequent parameters are loaded and f.Close() is executed
        ' -------------------------------------------------------------------------------
        On Error Resume Next  ' use Resume-Next instead of Try-Catch
        ' ------------------------------------------------------------------------------- 
        Dim l As String
        Dim LoadAll As Boolean = False
        If iniFileName = "" Then
            iniFileName = Application.StartupPath & "\" & AssemblyName() & "_INI.txt"
            LoadAll = True
        End If
        '
        If My.Computer.FileSystem.FileExists(iniFileName) Then

            Dim f As System.IO.StreamReader
            f = IO.File.OpenText(iniFileName)

            Do While Not f.EndOfStream
                l = f.ReadLine()
                Select Case ExtractParamName(l)
                    Case "Form1_Top" : If LoadAll Then Form1.Top = Val_Int(l)
                    Case "Form1_Left" : If LoadAll Then Form1.Left = Val_Int(l)
                    Case "Form1_Width" : If LoadAll Then Form1.Width = Val_Int(l)
                    Case "Form1_Height" : If LoadAll Then Form1.Height = Val_Int(l)
                    Case "Form1_WindowState" : If LoadAll Then Form1.WindowState = CType((Val(l)), FormWindowState)
                        ' ------------------------------------------------------------------------------
                    Case "GroupBox1_Height" : Form1.GroupBox1.Height = Val_Int(l)
                    Case "GroupBox2_Height" : Form1.GroupBox2.Height = Val_Int(l)
                    Case "GroupBox3_Height" : Form1.GroupBox3.Height = Val_Int(l)
                        ' ------------------------------------------------------------------------------
                    Case "PixelPerMM" : PixelPerMM = Val_Single(l)
                    Case "StartX" : StartX = Val_Single(l)
                    Case "StartY" : StartY = Val_Single(l)
                        ' ------------------------------------------------------------------------------
                    Case "WireDiam" : Form1.txt_WireDiam.NumericValue = Val_Double(l)
                    Case "NeedleDiam" : Form1.txt_NeedleDiam.NumericValue = Val_Double(l)
                        ' ------------------------------------------------------------------------------
                    Case "SewingAreaMaxY" : Form1.txt_SewingAreaMaxY.NumericValueInteger = Val_Int(l)
                    Case "SewingAreaMinY" : Form1.txt_SewingAreaMinY.NumericValueInteger = Val_Int(l)
                    Case "SewingAreaMaxX" : Form1.txt_SewingAreaMaxX.NumericValueInteger = Val_Int(l)
                    Case "SewingAreaMinX" : Form1.txt_SewingAreaMinX.NumericValueInteger = Val_Int(l)
                        ' ------------------------------------------------------------------------------
                    Case "ScaleText" : Form1.chk_ScaleText.Checked = l = "True"
                    Case "TrueXYZ" : Form1.chk_TrueXYZ.Checked = l = "True"
                        ' ------------------------------------------------------------------------------
                    Case "SlotX" : Form1.txt_SlotX.NumericValueInteger = Val_Int(l)
                    Case "SlotY" : Form1.txt_SlotY.NumericValueInteger = Val_Int(l)
                    Case "SlotZ" : Form1.txt_SlotZ.NumericValueInteger = Val_Int(l)
                    Case "SlotRestart" : Form1.txt_SlotRestart.NumericValueInteger = Val_Int(l)
                        ' ------------------------------------------------------------------------------
                End Select
            Loop
            f.Close()
        End If
        LimitFormPosition(Form1)
    End Sub


    ' ==================================================================================================
    '  SAVE IMAGE
    ' ==================================================================================================
    Public Sub SaveImage(ByVal img As Image, _
                         ByVal filename As String, _
                         ByVal extension As String, _
                         ByVal Quality As Int32)

        extension = LCase(extension)
        filename = RemoveExtension(filename)
        filename += "." & extension
        '
        If img Is Nothing Then Exit Sub
        Try
            File_Kill(filename)
            If extension = "jpg" Then
                Dim ImageEncoders() As ImageCodecInfo = ImageCodecInfo.GetImageEncoders()
                Dim myEncoder As System.Drawing.Imaging.Encoder = System.Drawing.Imaging.Encoder.Quality
                Dim myEncoderParameters As New EncoderParameters(1)
                Dim myEncoderParameter As New EncoderParameter(myEncoder, Quality)
                myEncoderParameters.Param(0) = myEncoderParameter
                img.Save(filename, ImageEncoders(1), myEncoderParameters)
            Else
                img.Save(filename, ImageFormatFromFileExtension(extension))
            End If
        Catch
            MsgBox("Image save error", MsgBoxStyle.Exclamation)
        End Try
    End Sub

    Private Function ImageFormatFromFileExtension(ByVal extension As String) As ImageFormat
        Select Case LCase(extension)
            Case "jpg" : Return ImageFormat.Jpeg
            Case "png" : Return ImageFormat.Png
            Case "tiff" : Return ImageFormat.Tiff
            Case "exif" : Return ImageFormat.Exif
            Case "emf" : Return ImageFormat.Emf
            Case "wmf" : Return ImageFormat.Wmf
            Case "gif" : Return ImageFormat.Gif
            Case "bmp" : Return ImageFormat.Bmp
                'Case "ico" : Return ImageFormat.Icon
            Case Else : Return ImageFormat.Jpeg
        End Select
    End Function

    Friend Sub File_Kill(ByVal filename As String)
        If My.Computer.FileSystem.FileExists(filename) Then
            My.Computer.FileSystem.DeleteFile(filename, FileIO.UIOption.OnlyErrorDialogs, FileIO.RecycleOption.DeletePermanently)
        End If
    End Sub

    Friend Function RemoveExtension(ByVal str As String) As String
        On Error Resume Next
        Return IO.Path.GetDirectoryName(str) & "\" & IO.Path.GetFileNameWithoutExtension(str)
    End Function


End Module
