﻿Imports System.Drawing.Imaging
Imports System.Drawing.Drawing2D

Module Utils

    ' ==================================================================================================
    '  LineToPointDistance2D - Minimal distance POINT to SEGMENT and POINT to LINE
    ' ==================================================================================================
    Private Function Dot2D(ByVal x0 As Double, ByVal y0 As Double, _
                           ByVal x1 As Double, ByVal y1 As Double, _
                           ByVal x2 As Double, ByVal y2 As Double) As Double
        Return (x1 - x0) * (x2 - x1) + (y1 - y0) * (y2 - y1)
    End Function

    Private Function Cross2D(ByVal x0 As Double, ByVal y0 As Double, _
                             ByVal x1 As Double, ByVal y1 As Double, _
                             ByVal x2 As Double, ByVal y2 As Double) As Double
        Return (x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)
    End Function

    Friend Function Dist2D(ByVal x1 As Double, ByVal y1 As Double, _
                            ByVal x2 As Double, ByVal y2 As Double) As Double
        Return Math.Sqrt((x1 - x2) ^ 2 + (y1 - y2) ^ 2)
    End Function

    Friend Function Distance_PointToSegment_2D(ByVal x0 As Double, ByVal y0 As Double, _
                                               ByVal x1 As Double, ByVal y1 As Double, _
                                               ByVal x2 As Double, ByVal y2 As Double) As Double
        If (Dot2D(x1, y1, x2, y2, x0, y0) > 0) Then Return Dist2D(x2, y2, x0, y0)
        If (Dot2D(x2, y2, x1, y1, x0, y0) > 0) Then Return Dist2D(x1, y1, x0, y0)
        Return Math.Abs(Cross2D(x1, y1, x2, y2, x0, y0) / Dist2D(x1, y1, x2, y2))
    End Function

    Friend Function Distance_PointToLine_2D(ByVal x0 As Double, ByVal y0 As Double, _
                                            ByVal x1 As Double, ByVal y1 As Double, _
                                            ByVal x2 As Double, ByVal y2 As Double) As Double
        Return Math.Abs(Cross2D(x1, y1, x2, y2, x0, y0) / Dist2D(x1, y1, x2, y2))
    End Function

    ' ================================================================================
    '  QUALITY
    ' ================================================================================
    Private Sub SetQuality(ByVal g As Graphics, ByVal nQuality As Int32)
        Select Case nQuality
            Case 1
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.NearestNeighbor
            Case 2
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.Low
            Case 3
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.Bilinear
            Case 4
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.HighQualityBilinear
            Case 5
                g.CompositingQuality = CompositingQuality.HighQuality
                g.SmoothingMode = SmoothingMode.HighQuality
                g.InterpolationMode = InterpolationMode.NearestNeighbor
            Case 6
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.Bicubic
            Case 7
                g.CompositingQuality = CompositingQuality.HighQuality
                g.SmoothingMode = SmoothingMode.HighQuality
                g.InterpolationMode = InterpolationMode.Bicubic
            Case 8
                g.CompositingQuality = CompositingQuality.HighQuality
                g.SmoothingMode = SmoothingMode.HighQuality
                g.InterpolationMode = InterpolationMode.HighQualityBilinear
            Case 9
                g.CompositingQuality = CompositingQuality.HighQuality
                g.SmoothingMode = SmoothingMode.HighQuality
                g.InterpolationMode = InterpolationMode.HighQualityBicubic
                ' --------------------------------------------------------------- ( zero or > 9 ) 
            Case Else
                g.CompositingQuality = CompositingQuality.HighSpeed
                g.SmoothingMode = SmoothingMode.HighSpeed
                g.InterpolationMode = InterpolationMode.Low
        End Select
    End Sub


End Module
